# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 30.0 `numdiff` Chapter.

``numdiff`` - Numerical Differentiation

This module is concerned with calculating approximations to derivatives
of a function f.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_30/flhtml/d04/d04intro.html
"""

# NAG Copyright 2017-2024.

import collections as _collections

from ..base import numdiff as _b_numdiff
from ..base import utils as _b_utils
from ..library import _utils

FwdReturnData = _collections.namedtuple(
    'FwdReturnData',
    ['der', 'erest'],
)

def fwd(xval, nder, hbase, f, data=None):
    u"""
    Numerical differentiation, derivatives up to order 14, function of
    one real variable.

    ``fwd`` calculates a set of derivatives (up to order 14) of a
    function of one real variable at a point, together with a
    corresponding set of error estimates, using an extension of the
    Neville algorithm.

    For full information please refer to the NAG Library document for
    d04aa

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/d04/d04aaf.html

    Parameters
    ----------
    xval : float
        The point at which the derivatives are required, x_0.

    nder : int
        Must be set so that its absolute value is the highest order
        derivative required.

        `nder` > 0

            All derivatives up to order min(`nder`, 14) are calculated.

        `nder` < 0 and `nder` is even

            Only even order derivatives up to order min(-`nder`, 14) are
            calculated.

        `nder` < 0 and `nder` is odd

            Only odd order derivatives up to order min(-`nder`, 13) are
            calculated.

    hbase : float
        The initial step length which may be positive or negative. For
        advice on the choice of `hbase` see [Further Comments].

    f : callable retval = f(x, data=None)
        `f` must evaluate the function f(x) at a specified point.

        Parameters
        ~~~~~~~~~~
        x : float
            The value of the argument x.

            If you have equally spaced tabular data, the following
            information may be useful:

            (i) in any call of ``fwd`` the only values of x for which
                f(x) will be required are x = `xval` and x =
                `xval`+/-(2*j-1)`hbase`, for j = 1,2,...,10; and

            (#) f(x_0) is always computed, but it is disregarded when
                only odd order derivatives are required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of f(x) at the specified point.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    der : float, ndarray, shape (14)
        `der`[j-1] contains an approximation to the jth derivative of
        f(x) at x = `xval`, so long as the jth derivative is one of
        those requested by you when specifying `nder`. For other values
        of j, `der`[j-1] is unused.

    erest : float, ndarray, shape (14)
        An estimate of the absolute error in the corresponding result
        `der`[j-1] so long as the jth derivative is one of those
        requested by you when specifying `nder`. The sign of
        `erest`[j-1] is positive unless the result `der`[j-1] is
        questionable. It is set negative when |`der`[j-1]| <
        |`erest`[j-1]| or when for some other reason there is doubt
        about the validity of the result `der`[j-1] (see
        [Raises/Warns]). For other values of j, `erest`[j-1] is unused.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nder` = 0.

            Constraint: `nder` != 0.

        (`errno` 1)
            On entry, `hbase` = 0.0.

            Constraint: `hbase` != 0.0.

    Notes
    -----
    ``fwd`` provides a set of approximations:

        `der`[j-1], j = 1,2,...,n

    to the derivatives:

        f^(j)(x_0), j = 1,2,...,n

    of a real valued function f(x) at a real abscissa x_0, together with
    a set of error estimates:

        `erest`[j-1], j = 1,2,...,n

    which hopefully satisfy:

        |`der`[j-1]-f^(j)(x_0)| < `erest`[j-1], j = 1,2,...,n.

    You must provide the value of x_0, a value of n (which is reduced to
    14 should it exceed 14), a function which evaluates f(x) for all
    real x, and a step length h.
    The results `der`[j-1] and `erest`[j-1] are based on 21 function
    values:

        f(x_0),f(x_0+/-(2i-1)h), i = 1,2,...,10.

    Internally ``fwd`` calculates the odd order derivatives and the even
    order derivatives separately.
    There is an option you can use for restricting the calculation to
    only odd (or even) order derivatives.
    For each derivative the function employs an extension of the Neville
    Algorithm (see Lyness and Moler (1969)) to obtain a selection of
    approximations.

    For example, for odd derivatives, based on 20 function values,
    ``fwd`` calculates a set of numbers:

        T_{k,p,s}, p = s,s+1,...,6, k = 0,1,...,9-p

    each of which is an approximation to f^(2s+1)(x_0)/(2s+1)!.
    A specific approximation T_{k,p,s} is of polynomial degree 2p+2 and
    is based on polynomial interpolation using function values
    f(x_0+/-(2i-1)h), for k = k,...,k+p.
    In the absence of round-off error, the better approximations would
    be associated with the larger values of p and of k.
    However, round-off error in function values has an increasingly
    contaminating effect for successively larger values of p.
    This function proceeds to make a judicious choice between all the
    approximations in the following way.

    For a specified value of s, let:

        R_p = U_p-L_p, p = s,s+1,...,6

    where U_p = max_k(T_{k,p,s}) and L_p = min_k(T_{k,p,s}), for k =
    0,1,...,9-p, and let p_ be such that R_p_= min_p(R_p), for p =
    s,...,6.

    The function returns:

        `der`[2s] = 1/(8-p_)*{sum_k =
        0^9-p_T_{k,p_,s}-U_p_-L_p_}*(2s+1)!

    and

        `erest`[2s] = R_p_*(2s+1)!*K_{2*s+1}

    where K_j is a safety factor which has been assigned the values:

    [table omitted]

    on the basis of performance statistics.

    The even order derivatives are calculated in a precisely analogous
    manner.

    References
    ----------
    Lyness, J N and Moler, C B, 1966, `van der Monde systems and
    numerical differentiation`, Numer. Math. (8), 458--464

    Lyness, J N and Moler, C B, 1969, `Generalised Romberg methods for
    integrals of derivatives`, Numer. Math. (14), 1--14
    """

    def cb_f(x, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _retval_f = f(
                x,
                data,
            )
        else:
            _retval_f = f(
                x,
            )
        return _retval_f

    _der = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (14,),
    )
    _erest = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (14,),
    )
    fun_name = 'naginterfaces.library.numdiff.fwd'

    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    _b_numdiff.fwd(
        xval,
        nder,
        hbase,
        _der,
        _erest,
        _f,
        data,
    )
    _der_py = _der
    _erest_py = _erest
    _return_data = FwdReturnData(
        _der_py,
        _erest_py,
    )
    return _return_data

RcommReturnData = _collections.namedtuple(
    'RcommReturnData',
    ['der', 'erest'],
)

def rcomm(xval, fval):
    u"""
    Numerical differentiation, user-supplied function values,
    derivatives up to order 14, derivatives with respect to one real
    variable.

    ``rcomm`` calculates a set of derivatives (up to order 14) of a
    function at a point with respect to a single variable.
    A corresponding set of error estimates is also returned.
    Derivatives are calculated using an extension of the Neville
    algorithm.
    This function differs from ``fwd``, in that the abscissae and
    corresponding function values must be calculated before this
    function is called.
    The abscissae may be generated using ``sample``.

    For full information please refer to the NAG Library document for
    d04ba

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/d04/d04baf.html

    Parameters
    ----------
    xval : float, array-like, shape (21)
        The abscissae at which the function has been evaluated, as
        described in [Notes]. These can be generated by calling
        ``sample``. The order of the abscissae is irrelevant.

    fval : float, array-like, shape (21)
        `fval`[j-1] must contain the function value at `xval`[j-1], for
        j = 1,2,...,21.

    Returns
    -------
    der : float, ndarray, shape (14)
        The 14 derivative estimates.

    erest : float, ndarray, shape (14)
        The 14 error estimates for the derivatives.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The derived h is below tolerance.

            Derived h > *<value>* is required. Derived h = *<value>*.

        (`errno` 1)
            On entry, the values of `xval` are not correctly spaced.

            Derived h = *<value>*.

    Notes
    -----
    ``rcomm`` provides a set of approximations:

        `der`[j-1], j = 1,2,...,14

    to the derivatives:

        f^(j)(x_0), j = 1,2,...,14

    of a real valued function f(x) at a real abscissa x_0, together with
    a set of error estimates:

        `erest`[j-1], j = 1,2,...,14

    which hopefully satisfy:

        |`der`[j-1]-f^(j)(x_0)| < `erest`[j-1], j = 1,2,...,14.

    The results `der`[j-1] and `erest`[j-1] are based on 21 function
    values:

        f(x_0),f(x_0+/-(2i-1)h), i = 1,2,...,10.

    The abscissae x and the corresponding function values f(x) should be
    passed into ``rcomm`` as the vectors `xval` and `fval` respectively.
    The step size h is derived from the abscissae in `xval`.
    See [Further Comments] for a discussion of how the derived value of
    h may affect the results of ``rcomm``.
    The order in which the abscissae and function values are stored in
    `xval` and `fval` is irrelevant, provided that the function value at
    any given index corresponds to the value of the abscissa at the same
    index.
    Abscissae may be automatically generated using ``sample`` if
    desired.
    For each derivative ``rcomm`` employs an extension of the Neville
    Algorithm (see Lyness and Moler (1969)) to obtain a selection of
    approximations.

    For example, for odd derivatives, this function calculates a set of
    numbers:

        T_{k,p,s}, p = s,s+1,...,6, k = 0,1,...,9-p

    each of which is an approximation to f^(2s+1)(x_0)/(2s+1)!.
    A specific approximation T_{k,p,s} is of polynomial degree 2p+2 and
    is based on polynomial interpolation using function values
    f(x_0+/-(2i-1)h), for k = k,...,k+p.
    In the absence of round-off error, the better approximations would
    be associated with the larger values of p and of k.
    However, round-off error in function values has an increasingly
    contaminating effect for successively larger values of p.
    This function proceeds to make a judicious choice between all the
    approximations in the following way.

    For a specified value of s, let:

        R_p = U_p-L_p, p = s,s+1,...,6

    where U_p = max_k(T_{k,p,s}) and L_p = min_k(T_{k,p,s}), for k =
    0,1,...,9-p, and let p_ be such that R_p_= min_p(R_p), for p =
    s,...,6.

    This function returns:

        `der`[2s] = 1/(8-p_)*{sum_k =
        0^9-p_T_{k,p_,s}-U_p_-L_p_}*(2s+1)!

    and

        `erest`[2s] = R_p_*(2s+1)!*K_{2s+1}

    where K_j is a safety factor which has been assigned the values:

    [table omitted]

    on the basis of performance statistics.

    The even order derivatives are calculated in a precisely analogous
    manner.

    References
    ----------
    Lyness, J N and Moler, C B, 1969, `Generalised Romberg methods for
    integrals of derivatives`, Numer. Math. (14), 1--14
    """

    fun_name = 'naginterfaces.library.numdiff.rcomm'

    _xval = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xval'},
        xval,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _fval = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'fval'},
        fval,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _der = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (14,),
    )
    _erest = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (14,),
    )
    _b_numdiff.rcomm(
        _xval,
        _fval,
        _der,
        _erest,
    )
    _der_py = _der
    _erest_py = _erest
    _return_data = RcommReturnData(
        _der_py,
        _erest_py,
    )
    return _return_data

def sample(x_0, hbase):
    u"""
    Generates sample points for function evaluations by ``rcomm``.

    ``sample`` generates abscissae about a target abscissa x_0 for use
    in a subsequent call to ``rcomm``.

    For full information please refer to the NAG Library document for
    d04bb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/d04/d04bbf.html

    Parameters
    ----------
    x_0 : float
        The abscissa x_0 at which derivatives are required.

    hbase : float
        The chosen step size h. If h < 10 epsilon, where epsilon =
        machine.precision(), the default h = epsilon^(1/4) will be used.

    Returns
    -------
    xval : float, ndarray, shape (21)
        The abscissae for passing to ``rcomm``.

    Notes
    -----
    ``sample`` may be used to generate the necessary abscissae about a
    target abscissa x_0 for the calculation of derivatives using
    ``rcomm``.

    For a given x_0 and h, the abscissae correspond to the set
    {x_0,x_0+/-(2*j-1)h}, for j = 1,2,...,10.
    These 21 points will be returned in ascending order in `xval`.
    In particular, `xval`[10] will be equal to x_0.

    References
    ----------
    Lyness, J N and Moler, C B, 1969, `Generalised Romberg methods for
    integrals of derivatives`, Numer. Math. (14), 1--14
    """

    _xval = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (21,),
    )
    _b_numdiff.sample(
        x_0,
        hbase,
        _xval,
    )
    _xval_py = _xval
    return _xval_py
