# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
# pylint: disable=simplifiable-if-statement
u"""
Interfaces for the NAG Mark 30.0 `stat` Chapter.

``stat`` - Simple Calculations on Statistical Data

This module covers the following topics:

    descriptive statistics and exploratory data analysis;

    statistical distribution functions and their inverses;

    testing for Normality and other distributions.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01intro.html

See Also
--------
``naginterfaces.library.examples.stat`` :
    This subpackage contains examples for the ``stat`` module.
    See also the :ref:`library_stat_ex` subsection.
"""

# NAG Copyright 2017-2024.

import collections as _collections

from ..base import stat as _b_stat
from ..base import utils as _b_utils
from ..library import _utils

Summary2varReturnData = _collections.namedtuple(
    'Summary2varReturnData',
    ['nval', 'res'],
)

def summary_2var(x1, x2, wt=None):
    u"""
    Means, corrected sums of squares and cross-products, etc., two
    variables, from raw data.

    ``summary_2var`` computes the means, standard deviations, corrected
    sums of squares and products, maximum and minimum values, and the
    product-moment correlation coefficient for two variables.
    Unequal weighting may be given.

    For full information please refer to the NAG Library document for
    g01ab

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01abf.html

    Parameters
    ----------
    x1 : float, array-like, shape (n)
        The observations from the first sample, x_i, for i = 1,2,...,n.

    x2 : float, array-like, shape (n)
        The observations from the second sample, y_i, for i = 1,2,...,n.

    wt : None or float, array-like, shape (n), optional
        If weights are being supplied then the elements of `wt` must
        contain the weights associated with the observations, w_i, for i
        = 1,2,...,n.

        If weights are not supplied then `wt` may be None.

    Returns
    -------
    nval : int
        Is used to indicate the number of valid observations, m; see
        [Notes] (g).

    res : float, ndarray, shape (13)
        The elements of `res` contain the following results:

        [table omitted]

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 3)
            The number of valid cases, m, is 0.

        (`errno` 3)
            On entry, `wt`[*<value>*] = *<value>*.

            Constraint: `wt`[i-1] >= 0.0, for i = 1,2,...,n

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The number of valid cases, m, is 1. In this case standard
            deviation and product-moment correlation coefficient cannot
            be calculated.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    The data consist of two samples of n observations, denoted by x_i,
    and y_i, for i = 1,2,...,n, with corresponding weights w_i, for i =
    1,2,...,n.

    If no specific weighting is given, then each w_i is set to 1.0 in
    ``summary_2var``.

    The quantities calculated are:

    (a) The sum of weights,

            W = sum_i = 1^nw_i.

    (#) The means,

            x_= sum_i = 1^nw_i x_i/W, y_= sum_i = 1^nw_i y_i/W.

    (#) The corrected sums of squares and products

            c_11 = sum_i = 1^n w_i (x_i-x_)^2 ;; c_21 = c_12 = sum_i =
            1^n w_i (x_i-x_) (y_i-y_);; c_22 = sum_i = 1^n w_i
            (y_i-y_)^2 .

    (#) The standard deviations

            s_j = sqrt((c_{jj})/d), where j = 1,2 and d = W-sum_i =
            1^nw_i^2/W.

    (#) The product-moment correlation coefficient

            R = (c_12)/sqrt(c_11 c_22).

    (#) The minimum and maximum elements in each of the two samples.

    (#) The number of pairs of observations, m, for which w_i > 0, i.e.,
        the number of **valid** observations. The quantities in (d) and
        (e) above will only be computed if m >= 2. All other items are
        computed if m >= 1.
    """

    fun_name = 'naginterfaces.library.stat.summary_2var'

    _x1 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x1'},
        x1,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x1, {'fun_name': fun_name, 'entity': 'x1'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x2'},
        x2,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _wt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'wt'},
        wt,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _res = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (13,),
    )
    try:
        _return_scalar = _b_stat.summary_2var(
            _n,
            _x1,
            _x2,
            _wt,
            _res,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _nval_py = _return_scalar
    _res_py = _res
    _return_data = Summary2varReturnData(
        _nval_py,
        _res_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

SummaryFreqReturnData = _collections.namedtuple(
    'SummaryFreqReturnData',
    ['xmean', 's2', 's3', 's4', 'n'],
)

def summary_freq(x, ifreq):
    u"""
    Mean, variance, skewness, kurtosis, etc., one variable, from
    frequency table.

    ``summary_freq`` calculates the mean, standard deviation and
    coefficients of skewness and kurtosis for data grouped in a
    frequency distribution.

    For full information please refer to the NAG Library document for
    g01ad

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01adf.html

    Parameters
    ----------
    x : float, array-like, shape (k)
        The elements of `x` must contain the boundary values of the
        classes in ascending order, so that class i is bounded by the
        values in `x`[i-1] and `x`[i], for i = 1,2,...,k-1.

    ifreq : int, array-like, shape (k)
        The ith element of `ifreq` must contain the frequency associated
        with the ith class, for i = 1,2,...,k-1. `ifreq`[k-1] is not
        used by the function.

    Returns
    -------
    xmean : float
        The mean value, y_.

    s2 : float
        The standard deviation, s_2.

    s3 : float
        The coefficient of skewness, s_3.

    s4 : float
        The coefficient of kurtosis, s_4.

    n : int
        The total frequency, n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k > 1.

        (`errno` 2)
            On entry, I = *<value>*, `x`[I-2] = *<value>* and `x`[I-1] =
            *<value>*.

            Constraint: `x`[I-2] <= `x`[I-1].

        (`errno` 3)
            Either `ifreq`[i] < 0 for some i, or the sum of frequencies
            is zero.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The total frequency, n, is less than 2, hence the quantities
            s_2, s_3 and s_4 cannot be calculated.

    Notes
    -----
    The input data consist of a univariate frequency distribution,
    denoted by f_i, for i = 1,2,...,k-1, and the boundary values of the
    classes x_i, for i = 1,2,...,k.
    Thus the frequency associated with the interval (x_i,x_{i+1}) is
    f_i, and ``summary_freq`` assumes that all the values in this
    interval are concentrated at the point

        y_i = (x_{i+1}+x_i)/2, i = 1,2,...,k-1.

    The following quantities are calculated:

    (a) total frequency,

            n = sum_i = 1^k-1f_i.

    (#) mean,

            y_= sum_i = 1^k-1f_i y_i/n.

    (#) standard deviation,

            s_2 = sqrt(sum_i = 1^k-1f_i(y_i-y_)^2/((n-1))), n >= 2.

    (#) coefficient of skewness,

            s_3 = sum_i = 1^k-1f_i(y_i-y_)^3/((n-1)*s_2^3), n >= 2.

    (#) coefficient of kurtosis,

            s_4 = sum_i = 1^k-1f_i(y_i-y_)^4/((n-1)*s_2^4)-3, n >= 2.

    The function has been developed primarily for groupings of a
    continuous variable.
    If, however, the function is to be used on the frequency
    distribution of a discrete variable, taking the values
    y_1,...,y_{k-1}, then the boundary values for the classes may be
    defined as follows:

    (i) for k > 2,

            [table omitted]

    (#) for k = 2,

            x_1 = y_1-a and x_2 = y_1+a for any a > 0.
    """

    fun_name = 'naginterfaces.library.stat.summary_freq'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _ifreq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ifreq'},
        ifreq,
        copy=False,
    )
    try:
        _return_scalars = _b_stat.summary_freq(
            _k,
            _x,
            _ifreq,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _xmean_py = _return_scalars.xmean
    _s2_py = _return_scalars.s2
    _s3_py = _return_scalars.s3
    _s4_py = _return_scalars.s4
    _n_py = _return_scalars.n
    _return_data = SummaryFreqReturnData(
        _xmean_py,
        _s2_py,
        _s3_py,
        _s4_py,
        _n_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

FrequencyTableReturnData = _collections.namedtuple(
    'FrequencyTableReturnData',
    ['cb', 'ifreq', 'xmin', 'xmax'],
)

def frequency_table(x, cb=None):
    u"""
    Frequency table from raw data.

    ``frequency_table`` constructs a frequency distribution of a
    variable, according to either user-supplied, or function-calculated
    class boundary values.

    For full information please refer to the NAG Library document for
    g01ae

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01aef.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The sample of observations of the variable for which the
        frequency distribution is required, x_i, for i = 1,2,...,n. The
        values may be in any order.

    cb : None or float, array-like, shape (k), optional
        If iclass = 0, the elements of `cb` need not be assigned values,
        as ``frequency_table`` calculates k-1 class boundary values.

        If iclass = 1, the first k-1 elements of `cb` must contain the
        class boundary values you supplied, in ascending order.

        In both cases, the element `cb`[k-1] need not be assigned, as it
        is not used in the function.

    Returns
    -------
    cb : float, ndarray, shape (k)
        The first k-1 elements of `cb` contain the class boundary values
        in ascending order.

    ifreq : int, ndarray, shape (k)
        The elements of `ifreq` contain the frequencies in each class,
        f_i, for i = 1,2,...,k. In particular `ifreq`[0] contains the
        frequency of the class up to `cb`[0], f_1, and `ifreq`[k-1]
        contains the frequency of the class greater than `cb`[k-2], f_k.

    xmin : float
        The smallest value in the sample, a.

    xmax : float
        The largest value in the sample, b.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 2.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 3)
            On entry, `cb`[*<value>*] = *<value>* and `cb`[*<value>*] =
            *<value>*.

            Constraint: `cb`[*<value>*] < `cb`[*<value>*].

    Notes
    -----
    The data consists of a sample of n observations of a continuous
    variable, denoted by x_i, for i = 1,2,...,n.
    Let a = min(x_1, ..., x_n) and b = max(x_1, ..., x_n).

    ``frequency_table`` constructs a frequency distribution with k(> 1)
    classes denoted by f_i, for i = 1,2,...,k.

    The boundary values may be either user-supplied, or
    function-calculated, and are denoted by y_j, for j = 1,2,...,k-1.

    If the boundary values of the classes are to be function-calculated,
    then they are determined in one of the following ways:

    (a) if k > 2, the range of x values is divided into k-2 intervals of
        equal length, and two extreme intervals, defined by the class
        boundary values y_1,y_2,...,y_{k-1};

    (#) if k = 2, y_1 = 1/2(a+b).

    However formed, the values y_1,...,y_{k-1} are assumed to be in
    ascending order.
    The class frequencies are formed with

        f_1 = the number of x values in the interval (- infinity,y_1)

        f_i = the number of x values in the interval [y_{i-1},y_i),  i =
        2,...,k-1

        f_k = the number of x values in the interval [y_{k-1},
        infinity),

    where [ means inclusive, and) means exclusive.
    If the class boundary values are function-calculated and k > 2, then
    f_1 = f_k = 0, and y_1 and y_{k-1} are chosen so that y_1 < a and
    y_{k-1} > b.

    If a frequency distribution is required for a discrete variable,
    then it is suggested that you supply the class boundary values;
    function-calculated boundary values may be slightly imprecise (due
    to the adjustment of y_1 and y_{k-1} outlined above) and cause
    values very close to a class boundary to be assigned to the wrong
    class.
    """

    fun_name = 'naginterfaces.library.stat.frequency_table'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _cb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cb'},
        cb,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _k = _b_utils._EngineIntScalarType(
        2 if cb is None else _b_utils._EngineFloat64ArrayType.get_shape(_cb, {'fun_name': fun_name, 'entity': 'cb'}, exp_rank=1, allow_none=True)[0], {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    if cb is None:
        _iclass = 0
        _iclass_py = _iclass
    else:
        _iclass = 1
        _iclass_py = _iclass
    _k_py = _b_utils._EngineIntScalarType.to_py(_k)
    if cb is None:
        _cb = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_k_py,),
        )
    _ifreq = _b_utils._EngineIntArrayType.empty_ndarray(
        (_k_py,),
    )
    _return_scalars = _b_stat.frequency_table(
        _n,
        _k,
        _x,
        _iclass,
        _cb,
        _ifreq,
    )
    _cb_py = _cb
    _ifreq_py = _ifreq
    _xmin_py = _return_scalars.xmin
    _xmax_py = _return_scalars.xmax
    _return_data = FrequencyTableReturnData(
        _cb_py,
        _ifreq_py,
        _xmin_py,
        _xmax_py,
    )
    return _return_data

ContingencyTableReturnData = _collections.namedtuple(
    'ContingencyTableReturnData',
    ['nobs', 'num', 'pred', 'chis', 'p', 'npos', 'ndf', 'm1', 'n1'],
)

def contingency_table(nobs, num=0):
    u"""
    Two-way contingency table analysis, with chi^2/Fisher's exact test.

    ``contingency_table`` performs the analysis of a two-way r*c
    contingency table or classification.
    If r = c = 2, and the total number of objects classified is 40 or
    fewer, then the probabilities for Fisher's exact test are computed.
    Otherwise, a test statistic is computed (with Yates' correction when
    r = c = 2), which under the assumption of no association between the
    classifications has approximately a chi-square distribution with
    (r-1)*(c-1) degrees of freedom.

    For full information please refer to the NAG Library document for
    g01af

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01aff.html

    Parameters
    ----------
    nobs : int, array-like, shape (m, n)
        The elements `nobs`[i-1,j-1], for j = 1,2,...,n, for i =
        1,2,...,m, must contain the frequencies for the two-way
        classification. The (m+1)th row and the (n+1)th column of `nobs`
        need not be set.

    num : int, optional
        The value assigned to `num` must determine whether automatic
        'shrinkage' is required when any r_{ij} < 1, as outlined in
        [Notes](1).

        If `num` = 1, shrinkage is required, otherwise shrinkage is not
        required.

    Returns
    -------
    nobs : int, ndarray, shape (m, n)
        Contains the following information:

            `nobs`[i-1,j-1], for j = 1,2,...,n_1, for i = 1,2,...,m_1,
            contain the frequencies for the two-way classification after
            'shrinkage' has taken place (see [Notes]).

            `nobs`[i-1,n], for i = 1,2,...,m_1, contain the total
            frequencies in the remaining rows, R_i.

            `nobs`[m,j-1], for j = 1,2,...,n_1, contain the total
            frequencies in the remaining columns, C_j.

            `nobs`[m,n], contains the total frequency, T.

        If any 'shrinkage' has occurred, all other cells contain no
        useful information.

    num : int
        When Fisher's exact test for a 2*2 classification is used then
        `num` contains the number of elements used in the array `p`,
        otherwise `num` is set to zero.

    pred : float, ndarray, shape (m, n)
        The elements `pred`[i-1,j-1], where i = 1,2,...,`m1` and j =
        1,2,...,`n1` contain the expected frequencies, r_{ij}
        corresponding to the observed frequencies `nobs`[i-1,j-1],
        except in the case when Fisher's exact test for a 2*2
        classification is to be used, when `pred` is not used. No other
        elements are utilized.

    chis : float
        The value of the test statistic, chi^2, except when Fisher's
        exact test for a 2*2 classification is used in which case it is
        unspecified.

    p : float, ndarray, shape (21)
        The first `num` elements contain the probabilities associated
        with the various possible frequency tables, P_r, for r =
        0,1,...,R_1, the remainder are unspecified.

    npos : int
        `p`[`npos`-1] holds the probability associated with the given
        table of frequencies.

    ndf : int
        The value of `ndf` gives the number of degrees of freedom for
        the chi-square distribution, (m_1-1)*(n_1-1); when Fisher's
        exact test is used `ndf` = 1.

    m1 : int
        The number of rows of the two-way classification, after any
        'shrinkage', m_1.

    n1 : int
        The number of columns of the two-way classification, after any
        'shrinkage', n_1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The number of rows or columns of `nobs` is less than 2.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 2.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m > 2.

        (`errno` 2)
            At least one frequency is negative, or all frequencies are
            zero.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    The data consist of the frequencies for the two-way classification,
    denoted by n_{ij}, for j = 1,2,...,n, for i = 1,2,...,m with m,n >
    1.

    A check is made to see whether any row or column of the matrix of
    frequencies consists entirely of zeros, and if so, the matrix of
    frequencies is reduced by omitting that row or column.
    Suppose the final size of the matrix is m_1*n_1 (m_1,n_1 > 1), and
    let

        R_i = sum_j = 1^n_1n_{ij}, the total frequency for the ith row,
        for i = 1,2,...,m_1,

        C_j = sum_i = 1^m_1n_{ij}, the total frequency for the jth
        column, for j = 1,2,...,n_1, and

        T = sum_i = 1^m_1R_i = sum_j = 1^n_1C_j, the total frequency.

    There are two situations:

    (i) If m_1 > 2 and/or n_1 > 2, or m_1 = n_1 = 2 and T > 40, then the
        matrix of expected frequencies, denoted by r_{ij}, for i =
        1,2,...,m_1 and j = 1,2,...,n_1, and the test statistic, chi^2,
        are computed, where

            r_{ij} = R_i C_j/T, i = 1,2,...,m_1;j = 1,2,...,n_1

        and

            chi^2 = sum_i = 1^m_1sum_j =
            1^n_1[|r_{ij}-n_{ij}|-Y]^2/r_{ij},

        where

            Y = {1/2 if m_1 = n_1 = 2;0 otherwise

        is Yates' correction for continuity.

        Under the assumption that there is no association between the
        two classifications, chi^2 will have approximately a chi-square
        distribution with (m_1-1)*(n_1-1) degrees of freedom.

        An option exists which allows for further 'shrinkage' of the
        matrix of frequencies in the case where r_{ij} < 1 for the
        (i,j)th cell.
        If this is the case, then row i or column j will be combined
        with the adjacent row or column with smaller total.
        Row i is selected for combination if R_i*m_1 <= C_j*n_1.
        This 'shrinking' process is continued until r_{ij} >= 1 for all
        cells (i,j).

    (#) If m_1 = n_1 = 2 and T <= 40, the probabilities to enable
        Fisher's exact test to be made are computed.

        The matrix of frequencies may be rearranged so that R_1 is the
        smallest marginal (i.e., column and row) total, and C_2 >= C_1.
        Under the assumption of no association between the
        classifications, the probability of obtaining r entries in cell
        (1,1) is computed where

            P_{r+1} =
            (R_1!R_2!C_1!C_2!)/(T!r!(R_1-r)!(C_1-r)!(T-C_1-R_1+r)!), r =
            0,1,...,R_1.

        The probability of obtaining the table of given frequencies is
        returned.
        A test of the assumption against some alternative may then be
        made by summing the relevant values of P_r.
    """

    fun_name = 'naginterfaces.library.stat.contingency_table'

    _nobs = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'nobs'},
        nobs,
        copy=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_nobs, 'nobs'),),
    )
    _m_py = _b_utils._EngineIntArrayType.get_shape(_nobs, {'fun_name': fun_name, 'entity': 'nobs'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntArrayType.get_shape(_nobs, {'fun_name': fun_name, 'entity': 'nobs'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _pred = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _n_py),
        sorder=_m_sorder,
    )
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (21,),
    )
    _return_scalars = _b_stat.contingency_table(
        _m,
        _n,
        _nobs,
        num,
        _pred,
        _p,
    )
    _nobs_py = _nobs
    _num_py = _return_scalars.num
    _pred_py = _pred
    _chis_py = _return_scalars.chis
    _p_py = _p
    _npos_py = _return_scalars.npos
    _ndf_py = _return_scalars.ndf
    _m1_py = _return_scalars.m1
    _n1_py = _return_scalars.n1
    _return_data = ContingencyTableReturnData(
        _nobs_py,
        _num_py,
        _pred_py,
        _chis_py,
        _p_py,
        _npos_py,
        _ndf_py,
        _m1_py,
        _n1_py,
    )
    return _return_data

def five_point_summary(x):
    u"""
    Computes a five-point summary (median, hinges and extremes).

    ``five_point_summary`` calculates a five-point summary for a single
    sample.

    For full information please refer to the NAG Library document for
    g01al

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01alf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The sample observations, x_1,x_2,...,x_n.

    Returns
    -------
    res : float, ndarray, shape (5)
        `res` contains the five-point summary.

        `res`[0]

            The minimum.

        `res`[1]

            The lower hinge.

        `res`[2]

            The median.

        `res`[3]

            The upper hinge.

        `res`[4]

            The maximum.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 5.

    Notes
    -----
    ``five_point_summary`` calculates the minimum, lower hinge, median,
    upper hinge and the maximum of a sample of n observations.

    The data consist of a single sample of n observations denoted by x_i
    and let z_i, for i = 1,2,...,n, represent the sample observations
    sorted into ascending order.

    Let m = n/2 if n is even and ((n+1))/2 if n is odd,

    and k = m/2 if m is even and ((m+1))/2 if m is odd.

    Then we have

    [table omitted]

    References
    ----------
    Erickson, B H and Nosanchuk, T A, 1985, `Understanding Data`, Open
    University Press, Milton Keynes

    Tukey, J W, 1977, `Exploratory Data Analysis`, Addison--Wesley
    """

    fun_name = 'naginterfaces.library.stat.five_point_summary'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _res = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (5,),
    )
    _b_stat.five_point_summary(
        _n,
        _x,
        _res,
    )
    _res_py = _res
    return _res_py

def quantiles(rv, q):
    u"""
    Find quantiles of an unordered vector, real numbers.

    ``quantiles`` finds specified quantiles from a vector of unsorted
    data.

    For full information please refer to the NAG Library document for
    g01am

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01amf.html

    Parameters
    ----------
    rv : float, array-like, shape (n)
        The vector whose quantiles are to be determined.

    q : float, array-like, shape (nq)
        The quantiles to be calculated, in ascending order. Note that
        these must be between 0.0 and 1.0, with 0.0 returning the
        smallest element and 1.0 the largest.

    Returns
    -------
    qv : float, ndarray, shape (nq)
        `qv`[i-1] contains the quantile specified by the value provided
        in `q`[i-1], or an interpolated value if the quantile falls
        between two data values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 2)
            On entry, nq = *<value>*.

            Constraint: nq > 0.

        (`errno` 3)
            On entry, an element of `q` was less than 0.0 or greater
            than 1.0.

        (`errno` 4)
            On entry, `q` was not in ascending order.

    Notes
    -----
    A quantile is a value which divides a frequency distribution such
    that there is a given proportion of data values below the quantile.
    For example, the median of a dataset is the 0.5 quantile because
    half the values are less than or equal to it; and the 0.25 quantile
    is the 25th percentile.

    ``quantiles`` uses a modified version of Singleton's
    'median-of-three' Quicksort algorithm (Singleton (1969)) to
    determine specified quantiles of a vector of real values.
    The input vector is partially sorted, as far as is required to
    compute desired quantiles; for a single quantile, this is much
    faster than sorting the entire vector.
    Where necessary, linear interpolation is also carried out to return
    the values of quantiles which lie between two data points.

    References
    ----------
    Singleton, R C, 1969, `An efficient algorithm for sorting with
    minimal storage: Algorithm 347`, Comm. ACM (12), 185--187
    """

    fun_name = 'naginterfaces.library.stat.quantiles'

    _rv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rv'},
        rv,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_rv, {'fun_name': fun_name, 'entity': 'rv'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _q = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'q'},
        q,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nq_py = _b_utils._EngineFloat64ArrayType.get_shape(_q, {'fun_name': fun_name, 'entity': 'q'}, exp_rank=1)[0]
    _nq = _b_utils._EngineIntScalarType(
        _nq_py, {'fun_name': fun_name, 'entity': '_nq'},
    ).to_ctypes()
    _qv = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nq_py,),
    )
    _b_stat.quantiles(
        _n,
        _rv,
        _nq,
        _q,
        _qv,
    )
    _qv_py = _qv
    return _qv_py

QuantilesStreamFixedReturnData = _collections.namedtuple(
    'QuantilesStreamFixedReturnData',
    ['ind', 'np', 'qv'],
)

def quantiles_stream_fixed(ind, n, rv, nb, eps, q, nq, rcomm, icomm):
    u"""
    Calculates approximate quantiles from a data stream of known size.

    ``quantiles_stream_fixed`` finds approximate quantiles from a data
    stream of known size using an out-of-core algorithm.

    For full information please refer to the NAG Library document for
    g01an

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01anf.html

    Parameters
    ----------
    ind : int
        Indicates the action required in the current call to
        ``quantiles_stream_fixed``.

        `ind` = 0

            Return the required length of `rcomm` and `icomm` in
            `icomm`[0] and `icomm`[1] respectively. `n` and `eps` must
            be set and licomm must be at least 2.

        `ind` = 1

            Initialise the communication arrays and process the first
            `nb` values from the data stream as supplied in `rv`.

        `ind` = 2

            Process the next block of `nb` values from the data stream.
            The calling program must update `rv` and (if required) `nb`,
            and re-enter ``quantiles_stream_fixed`` with all other
            parameters unchanged.

        `ind` = 3

            Calculate the `nq` epsilon-approximate quantiles specified
            in `q`. The calling program must set `q` and `nq` and
            re-enter ``quantiles_stream_fixed`` with all other
            parameters unchanged. This option can be chosen only when
            `np` >= ceiling(exp(1.0)/`eps`).

    n : int
        n, the total number of values in the data stream.

    rv : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `ind` in (1, 2): `nb`; otherwise: 0.

        If `ind` = 1 or 2, the vector containing the current block of
        data, otherwise `rv` is not referenced.

    nb : int
        If `ind` = 1 or 2, the size of the current block of data. The
        size of blocks of data in array `rv` can vary;, therefore, `nb`
        can change between calls to ``quantiles_stream_fixed``.

    eps : float
        Approximation factor epsilon.

    q : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `ind`=3: `nq`; otherwise: 0.

        If `ind` = 3, the quantiles to be calculated, otherwise `q` is
        not referenced. Note that `q`[i] = 0.0, corresponds to the
        minimum value and `q`[i] = 1.0 to the maximum value.

    nq : int
        If `ind` = 3, the number of quantiles requested, otherwise `nq`
        is not referenced.

    rcomm : float, ndarray, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `ind` != 0: the value returned in `icomm`[0] by a
        call to ``quantiles_stream_fixed`` with `ind` = 0; otherwise: 0.

        Communication array.

    icomm : int, ndarray, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `ind` != 0: the value returned in `icomm`[1] by a
        call to ``quantiles_stream_fixed`` with `ind` = 0; otherwise: 2.

        Communication array.

    Returns
    -------
    ind : int
        Indicates output from a successful call.

        `ind` = 1

            Lengths of `rcomm` and `icomm` have been returned in
            `icomm`[0] and `icomm`[1] respectively.

        `ind` = 2

            ``quantiles_stream_fixed`` has processed `np` data points
            and expects to be called again with additional data (i.e.,
            `np` < `n`).

        `ind` = 3

            ``quantiles_stream_fixed`` has returned the requested
            epsilon-approximate quantiles in `qv`. These quantiles are
            based on `np` data points.

        `ind` = 4

            Routine has processed all `n` data points (i.e., `np` =
            `n`).

    np : int
        The number of elements processed so far.

    qv : float, ndarray, shape (:)
        If `ind` = 3, `qv`[i] contains the epsilon-approximate quantiles
        specified by the value provided in `q`[i].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ind` = *<value>*.

            Constraint: `ind` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 3)
            On entry, `eps` = *<value>*.

            Constraint: exp(1.0)/`n` <= `eps` <= 1.0.

        (`errno` 4)
            On entry, `ind` = 1 or 2 and `nb` = *<value>*.

            Constraint: if `ind` = 1 or 2 then `nb` > 0.

        (`errno` 7)
            Number of data elements streamed, *<value>* is not
            sufficient for a quantile query when `eps` = *<value>*.

            Supply more data or reprocess the data with a higher `eps`
            value.

        (`errno` 8)
            On entry, `ind` = 3 and `nq` = *<value>*.

            Constraint: if `ind` = 3 then `nq` > 0.

        (`errno` 9)
            On entry, `ind` = 3 and `q`[*<value>*] = *<value>*.

            Constraint: if `ind` = 3 then 0.0 <= `q`[i] <= 1.0 for all
            i.

    Notes
    -----
    A quantile is a value which divides a frequency distribution such
    that there is a given proportion of data values below the quantile.
    For example, the median of a dataset is the 0.5 quantile because
    half the values are less than or equal to it.

    ``quantiles_stream_fixed`` uses a slightly modified version of an
    algorithm described in a paper by Zhang and Wang (2007) to determine
    epsilon-approximate quantiles of a data stream of n real values,
    where n is known.
    Given any quantile q in [0.0,1.0], an epsilon-approximate quantile
    is defined as an element in the data stream whose rank falls within
    [(q-epsilon)n,(q+epsilon)n].
    In case of more than one epsilon-approximate quantile being
    available, the one closest to qn is returned.

    References
    ----------
    Zhang, Q and Wang, W, 2007, `A fast algorithm for approximate
    quantiles in high speed data streams`, Proceedings of the 19th
    International Conference on Scientific and Statistical Database
    Management, IEEE Computer Society, 29
    """

    fun_name = 'naginterfaces.library.stat.quantiles_stream_fixed'

    _rv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rv'},
        rv,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _q = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'q'},
        q,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ind = _b_utils._EngineIntScalarType(
        ind, {'fun_name': fun_name, 'entity': 'ind'},
    ).to_ctypes()
    _ind_py = _b_utils._EngineIntScalarType.to_py(_ind)
    _nq = _b_utils._EngineIntScalarType(
        nq, {'fun_name': fun_name, 'entity': 'nq'},
    ).to_ctypes()
    _nq_py = _b_utils._EngineIntScalarType.to_py(_nq)
    if _ind_py == 3:
        _qv_dim1 = _nq_py
    else:
        _qv_dim1 = 0
    exp_shape_qv = (_qv_dim1,)
    _qv = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_qv,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'rcomm'},
        rcomm,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'icomm'},
        icomm,
    )
    _return_scalars = _b_stat.quantiles_stream_fixed(
        _ind,
        n,
        _rv,
        nb,
        eps,
        _q,
        _qv,
        _nq,
        rcomm,
        icomm,
    )
    _ind_py = _return_scalars.ind
    _np_py = _return_scalars.np
    if _ind_py == 3:
        _qv_dim1 = _nq_py
    else:
        _qv_dim1 = 0
    _qv_py = _qv[:_qv_dim1]
    _return_data = QuantilesStreamFixedReturnData(
        _ind_py,
        _np_py,
        _qv_py,
    )
    return _return_data

QuantilesStreamArbitraryReturnData = _collections.namedtuple(
    'QuantilesStreamArbitraryReturnData',
    ['ind', 'np', 'qv'],
)

def quantiles_stream_arbitrary(ind, rv, nb, eps, q, nq, rcomm, icomm):
    u"""
    Calculates approximate quantiles from a data stream of unknown size.

    ``quantiles_stream_arbitrary`` finds approximate quantiles from a
    large arbitrary-sized data stream using an out-of-core algorithm.

    For full information please refer to the NAG Library document for
    g01ap

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01apf.html

    Parameters
    ----------
    ind : int
        `On initial entry`: must be set to 0.Indicates the action
        required in the current call to ``quantiles_stream_arbitrary``.

        `ind` = 0

            Initialize the communication arrays and attempt to process
            the first `nb` values from the data stream. `eps`, `rv` and
            `nb` must be set and licomm must be at least 10.

        `ind` = 1

            Attempt to process the next block of `nb` values from the
            data stream. The calling program must update `rv` and (if
            required) `nb`, and re-enter ``quantiles_stream_arbitrary``
            with all other parameters unchanged.

        `ind` = 2

            Continue calculation following the reallocation of either or
            both of the communication arrays `rcomm` and `icomm`.

        `ind` = 3

            Calculate the `nq` epsilon-approximate quantiles specified
            in `q`. The calling program must set `q` and `nq` and
            re-enter ``quantiles_stream_arbitrary`` with all other
            parameters unchanged. This option can be chosen only when
            `np` >= ceiling(exp(1.0)/`eps`).

    rv : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `ind` in (0, 1, 2): `nb`; otherwise: 0.

        If `ind` = 0, 1 or 2, the vector containing the current block of
        data, otherwise `rv` is not referenced.

    nb : int
        If `ind` = 0, 1 or 2, the size of the current block of data. The
        size of blocks of data in array `rv` can vary;, therefore, `nb`
        can change between calls to ``quantiles_stream_arbitrary``.

    eps : float
        Approximation factor epsilon.

    q : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `ind`=3: `nq`; otherwise: 0.

        If `ind` = 3, the quantiles to be calculated, otherwise `q` is
        not referenced. Note that `q`[i] = 0.0, corresponds to the
        minimum value and `q`[i] = 1.0 to the maximum value.

    nq : int
        If `ind` = 3, the number of quantiles requested, otherwise `nq`
        is not referenced.

    rcomm : float, ndarray, shape (lrcomm), modified in place
        `On entry`: if `ind` = 1 or 2 then the first l elements of
        `rcomm` as supplied to ``quantiles_stream_arbitrary`` must be
        identical to the first l elements of `rcomm` returned from the
        last call to ``quantiles_stream_arbitrary``, where l is the
        value of lrcomm used in the last call. In other words, the
        contents of `rcomm` must not be altered between calls to this
        function. If `rcomm` needs to be reallocated then its contents
        must be preserved. If `ind` = 0 then `rcomm` need not be set.

        `On exit`: `rcomm` holds information required by subsequent
        calls to ``quantiles_stream_arbitrary``.

    icomm : int, ndarray, shape (licomm), modified in place
        `On entry`: if `ind` = 1 or 2 then the first l elements of
        `icomm` as supplied to ``quantiles_stream_arbitrary`` must be
        identical to the first l elements of `icomm` returned from the
        last call to ``quantiles_stream_arbitrary``, where l is the
        value of licomm used in the last call. In other words, the
        contents of `icomm` must not be altered between calls to this
        function. If `icomm` needs to be reallocated then its contents
        must be preserved. If `ind` = 0 then `icomm` need not be set.

        `On exit`: `icomm`[0] holds the minimum required length for
        `rcomm` and `icomm`[1] holds the minimum required length for
        `icomm`. The remaining elements of `icomm` are used for
        communication between subsequent calls to
        ``quantiles_stream_arbitrary``.

    Returns
    -------
    ind : int
        Indicates output from the call.

        `ind` = 1

            ``quantiles_stream_arbitrary`` has processed `np` data
            points and expects to be called again with additional data.

        `ind` = 2

            Either one or more of the communication arrays `rcomm` and
            `icomm` is too small. The new minimum lengths of `rcomm` and
            `icomm` have been returned in `icomm`[0] and `icomm`[1]
            respectively. If the new minimum length is greater than the
            current length then the corresponding communication array
            needs to be reallocated, its contents preserved and
            ``quantiles_stream_arbitrary`` called again with all other
            parameters unchanged.

            If there is more data to be processed, it is recommended
            that lrcomm and licomm are made significantly bigger than
            the minimum to limit the number of reallocations.

        `ind` = 3

            ``quantiles_stream_arbitrary`` has returned the requested
            epsilon-approximate quantiles in `qv`. These quantiles are
            based on `np` data points.

    np : int
        m, the number of elements processed so far.

    qv : float, ndarray, shape (:)
        If `ind` = 3, `qv`[i] contains the epsilon-approximate quantiles
        specified by the value provided in `q`[i].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ind` = *<value>*.

            Constraint: `ind` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `eps` = *<value>*.

            Constraint: 0.0 < `eps` <= 1.0.

        (`errno` 3)
            On entry, `ind` = 0, 1 or 2 and `nb` = *<value>*.

            Constraint: if `ind` = 0, 1 or 2 then `nb` > 0.

        (`errno` 6)
            The contents of `icomm` have been altered between calls to
            this function.

        (`errno` 7)
            The contents of `rcomm` have been altered between calls to
            this function.

        (`errno` 8)
            Number of data elements streamed, *<value>* is not
            sufficient for a quantile query when `eps` = *<value>*.

            Supply more data or reprocess the data with a higher `eps`
            value.

        (`errno` 9)
            On entry, `ind` = 3 and `nq` = *<value>*.

            Constraint: if `ind` = 3 then `nq` > 0.

        (`errno` 10)
            On entry, `ind` = 3 and `q`[*<value>*] = *<value>*.

            Constraint: if `ind` = 3 then 0.0 <= `q`[i] <= 1.0 for all
            i.

    Notes
    -----
    A quantile is a value which divides a frequency distribution such
    that there is a given proportion of data values below the quantile.
    For example, the median of a dataset is the 0.5 quantile because
    half the values are less than or equal to it.

    ``quantiles_stream_arbitrary`` uses a slightly modified version of
    an algorithm described in a paper by Zhang and Wang (2007) to
    determine epsilon-approximate quantiles of a large arbitrary-sized
    data stream of real values, where epsilon is a user-defined
    approximation factor.
    Let m denote the number of data elements processed so far then,
    given any quantile q in [0.0,1.0], an epsilon-approximate quantile
    is defined as an element in the data stream whose rank falls within
    [(q-epsilon)m,(q+epsilon)m].
    In case of more than one epsilon-approximate quantile being
    available, the one closest to qm is used.

    References
    ----------
    Zhang, Q and Wang, W, 2007, `A fast algorithm for approximate
    quantiles in high speed data streams`, Proceedings of the 19th
    International Conference on Scientific and Statistical Database
    Management, IEEE Computer Society, 29
    """

    fun_name = 'naginterfaces.library.stat.quantiles_stream_arbitrary'

    _rv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rv'},
        rv,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _q = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'q'},
        q,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ind = _b_utils._EngineIntScalarType(
        ind, {'fun_name': fun_name, 'entity': 'ind'},
    ).to_ctypes()
    _ind_py = _b_utils._EngineIntScalarType.to_py(_ind)
    _nq = _b_utils._EngineIntScalarType(
        nq, {'fun_name': fun_name, 'entity': 'nq'},
    ).to_ctypes()
    _nq_py = _b_utils._EngineIntScalarType.to_py(_nq)
    if _ind_py == 3:
        _qv_dim1 = _nq_py
    else:
        _qv_dim1 = 0
    exp_shape_qv = (_qv_dim1,)
    _qv = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_qv,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'rcomm'},
        rcomm,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'icomm'},
        icomm,
    )
    _return_scalars = _b_stat.quantiles_stream_arbitrary(
        _ind,
        _rv,
        nb,
        eps,
        _q,
        _qv,
        _nq,
        rcomm,
        icomm,
    )
    _ind_py = _return_scalars.ind
    _np_py = _return_scalars.np
    if _ind_py == 3:
        _qv_dim1 = _nq_py
    else:
        _qv_dim1 = 0
    _qv_py = _qv[:_qv_dim1]
    _return_data = QuantilesStreamArbitraryReturnData(
        _ind_py,
        _np_py,
        _qv_py,
    )
    return _return_data

SummaryOnevarReturnData = _collections.namedtuple(
    'SummaryOnevarReturnData',
    ['pn', 'xmean', 'xsd', 'xskew', 'xkurt', 'xmin', 'xmax'],
)

def summary_onevar(x, wt=None, pn=0, rcomm=None):
    u"""
    Computes univariate summary information: mean, variance, skewness,
    kurtosis.

    ``summary_onevar`` calculates the mean, standard deviation,
    coefficients of skewness and kurtosis, and the maximum and minimum
    values for a set of (optionally weighted) data.
    The input data can be split into arbitrary sized blocks, allowing
    large datasets to be summarised.

    For full information please refer to the NAG Library document for
    g01at

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01atf.html

    Parameters
    ----------
    x : float, array-like, shape (nb)
        The current block of observations, corresponding to x_i, for i =
        k+1,...,k+b, where k is the number of observations processed so
        far and b is the size of the current block of data.

    wt : None or float, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `wt` is not None: nb; otherwise: 0.

        If `wt` is not None, `wt` must contain the user-supplied weights
        corresponding to the block of data supplied in `x`, that is w_i,
        for i = k+1,...,k+b.

        If `wt` is None, w_i = 1 for all i.

    pn : int, optional
        The number of valid observations processed so far, that is the
        number of observations with w_i > 0, for i = 1,2,...,k. On the
        first call to ``summary_onevar``, or when starting to summarise
        a new dataset, `pn` must be set to 0.

        If `pn` != 0, it must be the same value as returned by the last
        call to ``summary_onevar``.

    rcomm : None or float, ndarray, shape (:), optional, modified in
    place
        Note: the required length for this argument is determined as
        follows: if `rcomm` is not None: 20; otherwise: 0.

        `Optionally, on entry`: communication array, used to store
        information between calls to ``summary_onevar``. If `pn` = 0,
        `rcomm` need not be initialized, otherwise it must be unchanged
        since the last call to this function.

        If `rcomm` is None, `rcomm` is not referenced and all the data
        must be supplied in one go.

        `On exit`, if not None on entry: the updated communication
        array. The first five elements of `rcomm` hold information that
        may be of interest with

            [table omitted]

        the remaining elements of `rcomm` are used for workspace and so
        are undefined.

    Returns
    -------
    pn : int
        The updated number of valid observations processed, that is the
        number of observations with w_i > 0, for i = 1,2,...,k+b.

    xmean : float
        x_, the mean of the first k+b observations.

    xsd : float
        s_2, the standard deviation of the first k+b observations.

    xskew : float
        s_3, the coefficient of skewness for the first k+b observations.

    xkurt : float
        s_4, the coefficient of kurtosis for the first k+b observations.

    xmin : float
        The smallest value in the first k+b observations.

    xmax : float
        The largest value in the first k+b observations.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, nb = *<value>*.

            Constraint: nb >= 0.

        (`errno` 41)
            On entry, `wt`[*<value>*] = *<value>*.

            Constraint: if `wt` is not None then `wt`[i-1] >= 0, for i =
            1,2,...,nb.

        (`errno` 51)
            On entry, `pn` = *<value>*.

            Constraint: `pn` >= 0.

        (`errno` 52)
            On entry, `pn` = *<value>*.

            On exit from previous call, `pn` = *<value>*.

            Constraint: if `pn` > 0, `pn` must be unchanged since
            previous call.

        (`errno` 54)
            On entry, `pn` = *<value>*.

            Constraint: if `rcomm` is None then `pn` = 0.

        (`errno` 121)
            `rcomm` has been corrupted between calls.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 53)
            On entry, the number of valid observations is zero.

        (`errno` 71)
            On exit we were unable to calculate `xskew` or `xkurt`. A
            value of 0 has been returned.

        (`errno` 72)
            On exit we were unable to calculate `xsd`, `xskew` or
            `xkurt`. A value of 0 has been returned.

    Notes
    -----
    Given a sample of n observations, denoted by x = {x_i:i = 1,2,...,n}
    and a set of non-negative weights, w = {w_i:i = 1,2,...,n},
    ``summary_onevar`` calculates a number of quantities:

    (a) Mean

            x_= sum(w_i x_i)/W, where W = sum(w_i).

    (#) Standard deviation

            s_2 = sqrt(sum(w_i)(x_i-x_)^2/d), where d = W-sum(w_i^2)/W.

    (#) Coefficient of skewness

            s_3 = sum(w_i)(x_i-x_)^3/(d*s_2^3).

    (#) Coefficient of kurtosis

            s_4 = sum(w_i(x_i-x_)^4)/(d*s_2^4)-3.

    (#) Maximum and minimum elements, with w_i != 0.

    These quantities are calculated using the one pass algorithm of West
    (1979).

    For large datasets, or where all the data is not available at the
    same time, x and w can be split into arbitrary sized blocks and
    ``summary_onevar`` called multiple times.

    References
    ----------
    West, D H D, 1979, `Updating mean and variance estimates: An
    improved method`, Comm. ACM (22), 532--555
    """

    fun_name = 'naginterfaces.library.stat.summary_onevar'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nb_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nb = _b_utils._EngineIntScalarType(
        _nb_py, {'fun_name': fun_name, 'entity': '_nb'},
    ).to_ctypes()
    _wt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'wt'},
        wt,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'rcomm'},
        rcomm,
        allow_none=True,
    )
    try:
        _return_scalars = _b_stat.summary_onevar(
            _nb,
            _x,
            _wt,
            pn,
            rcomm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _pn_py = _return_scalars.pn
    _xmean_py = _return_scalars.xmean
    _xsd_py = _return_scalars.xsd
    _xskew_py = _return_scalars.xskew
    _xkurt_py = _return_scalars.xkurt
    _xmin_py = _return_scalars.xmin
    _xmax_py = _return_scalars.xmax
    _return_data = SummaryOnevarReturnData(
        _pn_py,
        _xmean_py,
        _xsd_py,
        _xskew_py,
        _xkurt_py,
        _xmin_py,
        _xmax_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

SummaryOnevarCombineReturnData = _collections.namedtuple(
    'SummaryOnevarCombineReturnData',
    ['pn', 'xmean', 'xsd', 'xskew', 'xkurt', 'xmin', 'xmax', 'rcomm'],
)

def summary_onevar_combine(mrcomm):
    u"""
    Combines multiple sets of summary information, for use after
    ``summary_onevar``.

    ``summary_onevar_combine`` combines sets of summaries produced by
    ``summary_onevar``.

    For full information please refer to the NAG Library document for
    g01au

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01auf.html

    Parameters
    ----------
    mrcomm : float, array-like, shape (20, b)
        The jth column of `mrcomm` must contain the information returned
        in rcomm from one of the runs of ``summary_onevar``.

    Returns
    -------
    pn : int
        The number of valid observations, that is the number of
        observations with w_i > 0, for i = 1,2,...,n.

    xmean : float
        x_, the mean.

    xsd : float
        s_2, the standard deviation.

    xskew : float
        s_3, the coefficient of skewness.

    xkurt : float
        s_4, the coefficient of kurtosis.

    xmin : float
        The smallest value.

    xmax : float
        The largest value.

    rcomm : float, ndarray, shape (20)
        An amalgamation of the information held in `mrcomm`. This is in
        the same format as rcomm from ``summary_onevar``.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, b = *<value>*.

            Constraint: b >= 1.

        (`errno` 21)
            On entry, `mrcomm` is not in the expected format.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 31)
            On entry, the number of valid observations is zero.

        (`errno` 51)
            On exit we were unable to calculate `xskew` or `xkurt`. A
            value of 0 has been returned.

        (`errno` 52)
            On exit we were unable to calculate `xsd`, `xskew` or
            `xkurt`. A value of 0 has been returned.

    Notes
    -----
    Assume a dataset containing n observations, denoted by x = {x_i:i =
    1,2,...,n} and a set of weights, w = {w_i:i = 1,2,...,n}, has been
    split into b blocks, and each block summarised via a call to
    ``summary_onevar``.
    Then ``summary_onevar_combine`` takes the b communication arrays
    returned by ``summary_onevar`` and returns the mean (x_), standard
    deviation (s_2), coefficients of skewness (s_3) and kurtosis (s_4),
    and the maximum and minimum values for the whole dataset.

    For a definition of x_,s_2,s_3 and s_4 see [Notes for
    summary_onevar].

    References
    ----------
    West, D H D, 1979, `Updating mean and variance estimates: An
    improved method`, Comm. ACM (22), 532--555
    """

    fun_name = 'naginterfaces.library.stat.summary_onevar_combine'

    _mrcomm = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mrcomm'},
        mrcomm,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_py = _b_utils._EngineFloat64ArrayType.get_shape(_mrcomm, {'fun_name': fun_name, 'entity': 'mrcomm'}, exp_rank=2)[1]
    _b = _b_utils._EngineIntScalarType(
        _b_py, {'fun_name': fun_name, 'entity': '_b'},
    ).to_ctypes()
    _rcomm = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (20,),
    )
    try:
        _return_scalars = _b_stat.summary_onevar_combine(
            _b,
            _mrcomm,
            _rcomm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _pn_py = _return_scalars.pn
    _xmean_py = _return_scalars.xmean
    _xsd_py = _return_scalars.xsd
    _xskew_py = _return_scalars.xskew
    _xkurt_py = _return_scalars.xkurt
    _xmin_py = _return_scalars.xmin
    _xmax_py = _return_scalars.xmax
    _rcomm_py = _rcomm
    _return_data = SummaryOnevarCombineReturnData(
        _pn_py,
        _xmean_py,
        _xsd_py,
        _xskew_py,
        _xkurt_py,
        _xmin_py,
        _xmax_py,
        _rcomm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ProbBinomialReturnData = _collections.namedtuple(
    'ProbBinomialReturnData',
    ['plek', 'pgtk', 'peqk'],
)

def prob_binomial(n, p, k):
    u"""
    Binomial distribution function.

    ``prob_binomial`` returns the lower tail, upper tail and point
    probabilities associated with a binomial distribution.

    For full information please refer to the NAG Library document for
    g01bj

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01bjf.html

    Parameters
    ----------
    n : int
        The parameter n of the binomial distribution.

    p : float
        The parameter p of the binomial distribution.

    k : int
        The integer k which defines the required probabilities.

    Returns
    -------
    plek : float
        The lower tail probability, Prob{X <= k}.

    pgtk : float
        The upper tail probability, Prob{X > k}.

    peqk : float
        The point probability, Prob{X = k}.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `p` = *<value>*.

            Constraint: `p` < 1.0.

        (`errno` 2)
            On entry, `p` = *<value>*.

            Constraint: `p` > 0.0.

        (`errno` 3)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 3)
            On entry, `k` = *<value>* and `n` = *<value>*.

            Constraint: `k` <= `n`.

        (`errno` 4)
            On entry, `n` is too large to be represented exactly as a
            double precision number.

        (`errno` 5)
            On entry, the variance (= np(1-p)) exceeds 10^6.

    Notes
    -----
    Let X denote a random variable having a binomial distribution with
    parameters n and p (n >= 0 and 0 < p < 1).
    Then

        Prob{X = k} = (n;k)p^k(1-p)^(n-k), k = 0,1,...,n.

    The mean of the distribution is np and the variance is np(1-p).

    ``prob_binomial`` computes for given n, p and k the probabilities:

        `plek` = Prob{X <= k};`pgtk` = Prob{X > k}; `peqk` = Prob{X = k}
        .

    The method is similar to the method for the Poisson distribution
    described in Knüsel (1986).

    References
    ----------
    Knüsel, L, 1986, `Computation of the chi-square and Poisson
    distribution`, SIAM J. Sci. Statist. Comput. (7), 1022--1036
    """

    _return_scalars = _b_stat.prob_binomial(
        n,
        p,
        k,
    )
    _plek_py = _return_scalars.plek
    _pgtk_py = _return_scalars.pgtk
    _peqk_py = _return_scalars.peqk
    _return_data = ProbBinomialReturnData(
        _plek_py,
        _pgtk_py,
        _peqk_py,
    )
    return _return_data

ProbPoissonReturnData = _collections.namedtuple(
    'ProbPoissonReturnData',
    ['plek', 'pgtk', 'peqk'],
)

def prob_poisson(rlamda, k):
    u"""
    Poisson distribution function.

    ``prob_poisson`` returns the lower tail, upper tail and point
    probabilities associated with a Poisson distribution.

    For full information please refer to the NAG Library document for
    g01bk

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01bkf.html

    Parameters
    ----------
    rlamda : float
        The parameter lambda of the Poisson distribution.

    k : int
        The integer k which defines the required probabilities.

    Returns
    -------
    plek : float
        The lower tail probability, Prob{X <= k}.

    pgtk : float
        The upper tail probability, Prob{X > k}.

    peqk : float
        The point probability, Prob{X = k}.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `rlamda` = *<value>*.

            Constraint: `rlamda` > 0.0.

        (`errno` 2)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 3)
            On entry, `rlamda` = *<value>*.

            Constraint: `rlamda` <= 10^6.

    Notes
    -----
    Let X denote a random variable having a Poisson distribution with
    parameter lambda (> 0).
    Then

        Prob{X = k} = e^(-lambda)(lambda^k)/k!, k = 0,1,2,...

    The mean and variance of the distribution are both equal to lambda.

    ``prob_poisson`` computes for given lambda and k the probabilities:

        `plek` = Prob{X <= k};`pgtk` = Prob{X > k};`peqk` = Prob{X = k}
        .

    The method is described in Knüsel (1986).

    References
    ----------
    Knüsel, L, 1986, `Computation of the chi-square and Poisson
    distribution`, SIAM J. Sci. Statist. Comput. (7), 1022--1036
    """

    _return_scalars = _b_stat.prob_poisson(
        rlamda,
        k,
    )
    _plek_py = _return_scalars.plek
    _pgtk_py = _return_scalars.pgtk
    _peqk_py = _return_scalars.peqk
    _return_data = ProbPoissonReturnData(
        _plek_py,
        _pgtk_py,
        _peqk_py,
    )
    return _return_data

ProbHypergeomReturnData = _collections.namedtuple(
    'ProbHypergeomReturnData',
    ['plek', 'pgtk', 'peqk'],
)

def prob_hypergeom(n, l, m, k):
    u"""
    Hypergeometric distribution function.

    ``prob_hypergeom`` returns the lower tail, upper tail and point
    probabilities associated with a hypergeometric distribution.

    For full information please refer to the NAG Library document for
    g01bl

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01blf.html

    Parameters
    ----------
    n : int
        The parameter n of the hypergeometric distribution.

    l : int
        The parameter l of the hypergeometric distribution.

    m : int
        The parameter m of the hypergeometric distribution.

    k : int
        The integer k which defines the required probabilities.

    Returns
    -------
    plek : float
        The lower tail probability, Prob{X <= k}.

    pgtk : float
        The upper tail probability, Prob{X > k}.

    peqk : float
        The point probability, Prob{X = k}.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `l` = *<value>* and `n` = *<value>*.

            Constraint: `l` <= `n`.

        (`errno` 2)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 0.

        (`errno` 3)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` <= `n`.

        (`errno` 3)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 4)
            On entry, `k` = *<value>*, `l` = *<value>*, `m` = *<value>*
            and `l`+`m`-`n` = *<value>*.

            Constraint: `k` >= `l`+`m`-`n`.

        (`errno` 4)
            On entry, `k` = *<value>* and `m` = *<value>*.

            Constraint: `k` <= `m`.

        (`errno` 4)
            On entry, `k` = *<value>* and `l` = *<value>*.

            Constraint: `k` <= `l`.

        (`errno` 4)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 5)
            On entry, `n` is too large to be represented exactly as a
            double precision number.

        (`errno` 6)
            On entry, the variance = l*m*(n-l)*(n-m)/(n^2*(n-1)) exceeds
            10^6.

    Notes
    -----
    Let X denote a random variable having a hypergeometric distribution
    with parameters n, l and m (n >= l >= 0, n >= m >= 0).
    Then

        Prob{X = k} = (m;k)(n-m;l-k)/((n;l)),

    where max(0, l-(n-m)) <= k <= min(l, m), 0 <= l <= n and 0 <= m <=
    n.

    The hypergeometric distribution may arise if in a population of size
    n a number m are marked.
    From this population a sample of size l is drawn and of these k are
    observed to be marked.

    The mean of the distribution = lm/n, and the variance =
    lm(n-l)(n-m)/n^2(n-1).

    ``prob_hypergeom`` computes for given n, l, m and k the
    probabilities:

        `plek` = Prob{X <= k};`pgtk` = Prob{X > k};`peqk` = Prob{X = k}
        .

    The method is similar to the method for the Poisson distribution
    described in Knüsel (1986).

    References
    ----------
    Knüsel, L, 1986, `Computation of the chi-square and Poisson
    distribution`, SIAM J. Sci. Statist. Comput. (7), 1022--1036
    """

    _return_scalars = _b_stat.prob_hypergeom(
        n,
        l,
        m,
        k,
    )
    _plek_py = _return_scalars.plek
    _pgtk_py = _return_scalars.pgtk
    _peqk_py = _return_scalars.peqk
    _return_data = ProbHypergeomReturnData(
        _plek_py,
        _pgtk_py,
        _peqk_py,
    )
    return _return_data

NormalScoresExactReturnData = _collections.namedtuple(
    'NormalScoresExactReturnData',
    ['pp', 'errest'],
)

def normal_scores_exact(n, etol):
    u"""
    Normal scores, accurate values.

    ``normal_scores_exact`` computes a set of Normal scores, i.e., the
    expected values of an ordered set of independent observations from a
    Normal distribution with mean 0.0 and standard deviation 1.0.

    For full information please refer to the NAG Library document for
    g01da

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01daf.html

    Parameters
    ----------
    n : int
        n, the size of the set.

    etol : float
        The maximum value for the estimated absolute error in the
        computed scores.

    Returns
    -------
    pp : float, ndarray, shape (`n`)
        The Normal scores. `pp`[i-1] contains the value E(x_(i)), for i
        = 1,2,...,n.

    errest : float
        A computed estimate of the maximum error in the computed scores
        (see [Accuracy]).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 2)
            On entry, `etol` = *<value>*.

            Constraint: `etol` > 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The function was unable to estimate the scores with
            estimated error less than `etol`. The best result obtained
            is returned together with the associated value of `errest`.

    Notes
    -----
    If a sample of n observations from any distribution (which may be
    denoted by x_1,x_2,...,x_n), is sorted into ascending order, the rth
    smallest value in the sample is often referred to as the rth
    '**order statistic**', sometimes denoted by x_(r) (see Kendall and
    Stuart (1969)).

    The order statistics, therefore, have the property

        x_(1) <= x_(2) <= ... <= x_(n).

    (If n = 2r+1, x_{r+1} is the sample median.)

    For samples originating from a known distribution, the distribution
    of each order statistic in a sample of given size may be determined.
    In particular, the expected values of the order statistics may be
    found by integration.
    If the sample arises from a Normal distribution, the expected values
    of the order statistics are referred to as the '**Normal scores**'.
    The Normal scores provide a set of reference values against which
    the order statistics of an actual data sample of the same size may
    be compared, to provide an indication of Normality for the sample.
    Normal scores have other applications; for instance, they are
    sometimes used as alternatives to ranks in nonparametric testing
    procedures.

    ``normal_scores_exact`` computes the rth Normal score for a given
    sample size n as

        E(x_(r)) = int_- infinity^infinity x_r dG_r,

    where

        dG_r = A_r^r-1(1-A_r)^(n-r)dA_r/beta(r,n-r+1), A_r = 1/(sqrt(2
        pi))int_- infinity^x_re^(-t^2/2)dt, r = 1,2,...,n,

    and beta denotes the complete beta function.

    The function attempts to evaluate the scores so that the estimated
    error in each score is less than the value `etol` specified by you.
    All integrations are performed in parallel and arranged so as to
    give good speed and reasonable accuracy.

    References
    ----------
    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin
    """

    fun_name = 'naginterfaces.library.stat.normal_scores_exact'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _pp = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = _b_stat.normal_scores_exact(
            _n,
            _pp,
            etol,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _pp_py = _pp
    _errest_py = _return_scalar
    _return_data = NormalScoresExactReturnData(
        _pp_py,
        _errest_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def normal_scores_approx(n):
    u"""
    Normal scores, approximate values.

    ``normal_scores_approx`` calculates an approximation to the set of
    Normal Scores, i.e., the expected values of an ordered set of
    independent observations from a Normal distribution with mean 0.0
    and standard deviation 1.0.

    For full information please refer to the NAG Library document for
    g01db

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01dbf.html

    Parameters
    ----------
    n : int
        n, the size of the sample.

    Returns
    -------
    pp : float, ndarray, shape (`n`)
        The Normal scores. `pp`[i-1] contains the value E(x_(i)), for i
        = 1,2,...,n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``normal_scores_approx`` is an adaptation of the Applied Statistics
    Algorithm AS 177.3, see Royston (1982).
    If you are particularly concerned with the accuracy with which
    ``normal_scores_approx`` computes the expected values of the order
    statistics (see [Accuracy]), then ``normal_scores_exact`` which is
    more accurate should be used instead at a cost of increased storage
    and computing time.

    Let x_(1),x_(2),...,x_(n) be the order statistics from a random
    sample of size n from the standard Normal distribution.
    Defining

        P_{r,n} = Phi(-E(x_(r)))

    and

        Q_{r,n} = (r-epsilon)/(n+gamma), r = 1,2,...,n,

    where E(x_(r)) is the expected value of x_(r), the current function
    approximates the Normal upper tail area corresponding to E(x_(r))
    as,

        P~_{r,n} = Q_{r,n}+(delta_1)/nQ_r,n^lambda+(delta_2)/nQ_r,n^2
        lambda-C_{r,n}.

    , for r = 1,2,...,3, and r >= 4.
    Estimates of epsilon, gamma, delta_1, delta_2 and lambda are
    obtained.
    A small correction C_{r,n} to P~_{r,n} is necessary when r <= 7 and
    n <= 20.

    The approximation to E(X_(r)) is thus given by

        E(x_(r)) = -Phi^-1(P ~_{r,n}), r = 1,2,...,n.

    Values of the inverse Normal probability integral Phi^-1 are
    obtained from ``inv_cdf_normal``.

    References
    ----------
    Royston, J P, 1982, `Algorithm AS 177: expected normal order
    statistics (exact and approximate)`, Appl. Statist. (31), 161--165
    """

    fun_name = 'naginterfaces.library.stat.normal_scores_approx'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _pp = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _b_stat.normal_scores_approx(
        _n,
        _pp,
    )
    _pp_py = _pp
    return _pp_py

def normal_scores_var(n, exp1, exp2, sumssq):
    u"""
    Normal scores, approximate variance-covariance matrix.

    ``normal_scores_var`` computes an approximation to the
    variance-covariance matrix of an ordered set of independent
    observations from a Normal distribution with mean 0.0 and standard
    deviation 1.0.

    For full information please refer to the NAG Library document for
    g01dc

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01dcf.html

    Parameters
    ----------
    n : int
        n, the sample size.

    exp1 : float
        The expected value of the largest Normal order statistic, m_n,
        from a sample of size n.

    exp2 : float
        The expected value of the second largest Normal order statistic,
        m_{n-1}, from a sample of size n.

    sumssq : float
        The sum of squares of the expected values of the Normal order
        statistics from a sample of size n.

    Returns
    -------
    vec : float, ndarray, shape (`n`*(`n`+1)/2)
        The upper triangle of the n*n variance-covariance matrix packed
        by column. Thus element V_{ij} is stored in
        `vec`[i+j*(j-1)/2-1], for 1 <= i <= j <= n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

    Notes
    -----
    ``normal_scores_var`` is an adaptation of the Applied Statistics
    Algorithm AS 128, see Davis and Stephens (1978).
    An approximation to the variance-covariance matrix, V, using a
    Taylor series expansion of the Normal distribution function is
    discussed in David and Johnson (1954).

    However, convergence is slow for extreme variances and covariances.
    The present function uses the David--Johnson approximation to
    provide an initial approximation and improves upon it by use of the
    following identities for the matrix.

    For a sample of size n, let m_i be the expected value of the ith
    largest order statistic, then:

    (a) for any i = 1,2,...,n, sum_j = 1^nV_{ij} = 1

    (#) V_12 = V_11+m_n^2-m_n m_{n-1}-1

    (#) the trace of V is tr(V) = n-sum_i = 1^nm_i^2

    (#) V_{ij} = V_{ji} = V_{rs} = V_{sr} where r = n+1-i, s = n+1-j and
        i,j = 1,2,...,n. Note that only the upper triangle of the matrix
        is calculated and returned column-wise in vector form.

    References
    ----------
    David, F N and Johnson, N L, 1954, `Statistical treatment of
    censored data, Part 1. Fundamental formulae`, Biometrika (41),
    228--240

    Davis, C S and Stephens, M A, 1978, `Algorithm AS 128: approximating
    the covariance matrix of Normal order statistics`, Appl. Statist.
    (27), 206--212
    """

    fun_name = 'naginterfaces.library.stat.normal_scores_var'

    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _vec = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py*(_n_py+1)//2,),
    )
    _b_stat.normal_scores_var(
        _n,
        exp1,
        exp2,
        sumssq,
        _vec,
    )
    _vec_py = _vec
    return _vec_py

TestShapiroWilkReturnData = _collections.namedtuple(
    'TestShapiroWilkReturnData',
    ['a', 'w', 'pw'],
)

def test_shapiro_wilk(x, a=None):
    u"""
    Shapiro and Wilk's W test for Normality.

    ``test_shapiro_wilk`` calculates Shapiro and Wilk's W statistic and
    its significance level for testing Normality.

    For full information please refer to the NAG Library document for
    g01dd

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01ddf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The ordered sample values, x_i, for i = 1,2,...,n.

    a : None or float, array-like, shape (n), optional
        If calwts has been set to False then before entry `a` must
        contain the n weights as calculated in a previous call to
        ``test_shapiro_wilk``, otherwise `a` need not be set.

    Returns
    -------
    a : float, ndarray, shape (n)
        The n weights required to calculate `w`.

    w : float
        The value of the statistic, `w`.

    pw : float
        The significance level of `w`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 3.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n <= 5000.

        (`errno` 3)
            On entry, elements of `x` not in order. `x`[*<value>*] =
            *<value>*, `x`[*<value>*] = *<value>*, `x`[*<value>*] =
            *<value>*.

        (`errno` 3)
            On entry, all elements of `x` are equal.

    Notes
    -----
    ``test_shapiro_wilk`` calculates Shapiro and Wilk's W statistic and
    its significance level for any sample size between 3 and 5000.
    It is an adaptation of the Applied Statistics Algorithm AS R94, see
    Royston (1995).
    The full description of the theory behind this algorithm is given in
    Royston (1992).

    Given a set of observations x_1,x_2,...,x_n sorted into either
    ascending or descending order (``sort.realvec_sort`` may be used to
    sort the data) this function calculates the value of Shapiro and
    Wilk's W statistic defined as:

        W = (sum_i = 1^na_i x_i)^2/sum_i = 1^n(x_i-x_)^2,

    where x_= 1/nsum_1^nx_i is the sample mean and a_i, for i =
    1,2,...,n, are a set of 'weights' whose values depend only on the
    sample size n.

    On exit, the values of a_i, for i = 1,2,...,n, are only of interest
    should you wish to call the function again to calculate `w` and its
    significance level for a different sample of the same size.

    It is recommended that the function is used in conjunction with a
    Normal (Q-Q) plot of the data. Functions ``normal_scores_exact`` and
    ``normal_scores_approx`` can be used to obtain the required Normal
    scores.

    References
    ----------
    Royston, J P, 1982, `Algorithm AS 181: the W test for normality`,
    Appl. Statist. (31), 176--180

    Royston, J P, 1986, `A remark on AS 181: the W test for normality`,
    Appl. Statist. (35), 232--234

    Royston, J P, 1992, `Approximating the Shapiro--Wilk's W test for
    non-normality`, Statistics & Computing (2), 117--119

    Royston, J P, 1995, `A remark on AS R94: A remark on Algorithm AS
    181: the W test for normality`, Appl. Statist. (44(4)), 547--551
    """

    fun_name = 'naginterfaces.library.stat.test_shapiro_wilk'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    if a is None:
        _calwts = True
        _calwts_py = _calwts
    else:
        _calwts = False
        _calwts_py = _calwts
    if a is None:
        _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_n_py,),
        )
    else:
        _a = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'a'},
            a,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    _return_scalars = _b_stat.test_shapiro_wilk(
        _x,
        _n,
        _calwts,
        _a,
    )
    _a_py = _a
    _w_py = _return_scalars.w
    _pw_py = _return_scalars.pw
    _return_data = TestShapiroWilkReturnData(
        _a_py,
        _w_py,
        _pw_py,
    )
    return _return_data

def ranks_and_scores(scores, ties, x):
    u"""
    Ranks, Normal scores, approximate Normal scores or exponential
    (Savage) scores.

    ``ranks_and_scores`` computes the ranks, Normal scores, an
    approximation to the Normal scores or the exponential scores as
    requested by you.

    For full information please refer to the NAG Library document for
    g01dh

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01dhf.html

    Parameters
    ----------
    scores : str, length 1
        Indicates which of the following scores are required.

        `scores` = 'R'

            The ranks.

        `scores` = 'N'

            The Normal scores, that is the expected value of the Normal
            order statistics.

        `scores` = 'B'

            The Blom version of the Normal scores.

        `scores` = 'T'

            The Tukey version of the Normal scores.

        `scores` = 'V'

            The van der Waerden version of the Normal scores.

        `scores` = 'S'

            The Savage scores, that is the expected value of the
            exponential order statistics.

    ties : str, length 1
        Indicates which of the following methods is to be used to assign
        scores to tied observations.

        `ties` = 'A'

            The average of the scores for tied observations is used.

        `ties` = 'L'

            The lowest score in the group of ties is used.

        `ties` = 'H'

            The highest score in the group of ties is used.

        `ties` = 'N'

            The nonrepeatable random number generator is used to
            randomly untie any group of tied observations.

        `ties` = 'R'

            The repeatable random number generator is used to randomly
            untie any group of tied observations.

        `ties` = 'I'

            Any ties are ignored, that is the scores are assigned to
            tied observations in the order that they appear in the data.

    x : float, array-like, shape (n)
        The sample of observations, x_i, for i = 1,2,...,n.

    Returns
    -------
    r : float, ndarray, shape (n)
        Contains the scores, s_i, for i = 1,2,...,n, as specified by
        `scores`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `scores` = *<value>*.

            Constraint: `scores` = 'R', 'N', 'B', 'T', 'V' or 'S'.

        (`errno` 1)
            On entry, `ties` = *<value>*.

            Constraint: `ties` = 'A', 'L', 'H', 'R' or 'I'.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

    Notes
    -----
    ``ranks_and_scores`` computes one of the following scores for a
    sample of observations, x_1,x_2,...,x_n.

    (1) **Rank Scores**

        The ranks are assigned to the data in ascending order, that is
        the ith observation has score s_i = k if it is the kth smallest
        observation in the sample.

    (#) **Normal Scores**

        The Normal scores are the expected values of the Normal order
        statistics from a sample of size n.
        If x_i is the kth smallest observation in the sample, then the
        score for that observation, s_i, is E(Z_k) where Z_k is the kth
        order statistic in a sample of size n from a standard Normal
        distribution and E is the expectation operator.

    (#) **Blom, Tukey and van der Waerden Scores**

        These scores are approximations to the Normal scores.
        The scores are obtained by evaluating the inverse cumulative
        Normal distribution function, Phi^-1(·), at the values of the
        ranks scaled into the interval (0,1) using different scaling
        transformations.

        The Blom scores use the scaling transformation (r_i-3/8)/(n+1/4)
        for the rank r_i, for i = 1,2,...,n.
        Thus the Blom score corresponding to the observation x_i is

            s_i = Phi^-1((r_i-3/8)/(n+1/4)).

        The Tukey scores use the scaling transformation
        (r_i-1/3)/(n+1/3); the Tukey score corresponding to the
        observation x_i is

            s_i = Phi^-1((r_i-1/3)/(n+1/3)).

        The van der Waerden scores use the scaling transformation
        (r_i)/(n+1); the van der Waerden score corresponding to the
        observation x_i is

            s_i = Phi^-1((r_i)/(n+1)).

        The van der Waerden scores may be used to carry out the van der
        Waerden test for testing for differences between several
        population distributions, see Conover (1980).

    (#) **Savage Scores**

        The Savage scores are the expected values of the exponential
        order statistics from a sample of size n.
        They may be used in a test discussed by Savage (1956) and
        Lehmann (1975).
        If x_i is the kth smallest observation in the sample, then the
        score for that observation is

            s_i = E(Y_k) = 1/n+1/(n-1)+...+1/(n-k+1),

        where Y_k is the kth order statistic in a sample of size n from
        a standard exponential distribution and E is the expectation
        operator.

    Ties may be handled in one of five ways.
    Let x_{t(i)}, for i = 1,2,...,m, denote m tied observations, that is
    x_{t(1)} = x_{t(2)} = ... = x_{t(m)} with t(1) < t(2) < ... < t(m).
    If the rank of x_{t(1)} is k, then if ties are ignored the rank of
    x_{t(j)} will be k+j-1.
    Let the scores ignoring ties be s_t(1)^*,s_t(2)^*,...,s_t(m)^*.
    Then the scores, s_{t(i)}, for i = 1,2,...,m, may be calculated as
    follows:

    -   if averages are used, then s_{t(i)} = sum_j = 1^ms_t(j)^*/m;

    -   if the lowest score is used, then s_{t(i)} = s_t(1)^*;

    -   if the highest score is used, then s_{t(i)} = s_t(m)^*;

    -   if ties are to be broken randomly, then s_{t(i)} = s_t(I)^*
        where I in {random permutation of 1,2,...,m};

    -   if ties are to be ignored, then s_{t(i)} = s_t(i)^*.

    References
    ----------
    Blom, G, 1958, `Statistical Estimates and Transformed
    Beta-variables`, Wiley

    Conover, W J, 1980, `Practical Nonparametric Statistics`, Wiley

    Lehmann, E L, 1975, `Nonparametrics: Statistical Methods Based on
    Ranks`, Holden--Day

    Savage, I R, 1956, `Contributions to the theory of rank order
    statistics -- the two-sample case`, Ann. Math. Statist. (27),
    590--615

    Tukey, J W, 1962, `The future of data analysis`, Ann. Math. Statist.
    (33), 1--67
    """

    fun_name = 'naginterfaces.library.stat.ranks_and_scores'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _b_stat.ranks_and_scores(
        scores,
        ties,
        _n,
        _x,
        _r,
    )
    _r_py = _r
    return _r_py

def prob_normal(x, tail='L'):
    u"""
    Computes probabilities for the standard Normal distribution.

    ``prob_normal`` returns a one or two tail probability for the
    standard Normal distribution.

    For full information please refer to the NAG Library document for
    g01ea

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01eaf.html

    Parameters
    ----------
    x : float
        x, the value of the standard Normal variate.

    tail : str, length 1, optional
        Indicates which tail the returned probability should represent.

        `tail` = 'L'

            The lower tail probability is returned, i.e., P(X <= x).

        `tail` = 'U'

            The upper tail probability is returned, i.e., P(X >= x).

        `tail` = 'S'

            The two tail (significance level) probability is returned,
            i.e., P(X >= |x|)+P(X <= -|x|).

        `tail` = 'C'

            The two tail (confidence interval) probability is returned,
            i.e., P(X <= |x|)-P(X <= -|x|).

    Returns
    -------
    p : float
        A one or two tail probability for the standard Normal
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'L', 'U', 'S' or 'C'.

    Notes
    -----
    The lower tail probability for the standard Normal distribution, P(X
    <= x) is defined by:

        P(X <= x) = int_- infinity^xZ(X)dX,

    where

        Z(X) = 1/(sqrt(2 pi))e^(-X^2/2), - infinity < X < infinity.

    The relationship

        P(X <= x) = 1/2erfc((-x)/(sqrt(2)))

    is used, where erfc is the complementary error function, and is
    computed using ``specfun.erfc_real``.
    For the upper tail probability the relationship P(X >= x) = P(X <=
    -x) is used and for the two tail significance level probability
    twice the probability obtained from the absolute value of x is
    returned.

    When the two tail confidence probability is required the
    relationship

        P(X <= |x|)-P(X <= -|x|) = erf(|x|/(sqrt(2))),

    is used, where erf is the error function, and is computed using
    ``specfun.erf_real``.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth
    """

    _return_scalar = _b_stat.prob_normal(
        tail,
        x,
    )
    return _return_scalar

def prob_students_t(t, df, tail='L'):
    u"""
    Computes probabilities for Student's t-distribution.

    ``prob_students_t`` returns the lower tail, upper tail or two tail
    probability for the Student's t-distribution with real degrees of
    freedom.

    For full information please refer to the NAG Library document for
    g01eb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01ebf.html

    Parameters
    ----------
    t : float
        t, the value of the Student's t variate.

    df : float
        nu, the degrees of freedom of the Student's t-distribution.

    tail : str, length 1, optional
        Indicates which tail the returned probability should represent.

        `tail` = 'U'

            The upper tail probability is returned, i.e., P(T >= t:nu).

        `tail` = 'S'

            The two tail (significance level) probability is returned,
            i.e., P(T >= |t|:nu)+P(T <= -|t|:nu).

        `tail` = 'C'

            The two tail (confidence interval) probability is returned,
            i.e., P(T <= |t|:nu)-P(T <= -|t|:nu).

        `tail` = 'L'

            The lower tail probability is returned, i.e., P(T <= t:nu).

    Returns
    -------
    p : float
        Either the lower tail, upper tail or two tail probability for
        the Student's t-distribution, depending on the value of `tail`
        supplied.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'L', 'U', 'S' or 'C'.

        (`errno` 2)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 1.0.

    Notes
    -----
    The lower tail probability for the Student's t-distribution with nu
    degrees of freedom, P(T <= t:nu) is defined by:

        P(T <= t:nu) = Gamma((nu+1)/2)/(sqrt(pinu)Gamma(nu/2))int_-
        infinity^t[1+(T^2)/nu]^(-(nu+1)/2)dT, nu >= 1.

    Computationally, there are two situations:

    (i) when nu < 20, a transformation of the beta distribution,
        P_beta(B <= beta:a,b) is used

            P(T <= t:nu) = 1/2P_beta(B <= nu/(nu+t^2):nu/2,1/2) when t <
            0.0

        or

            P(T <= t:nu) = 1/2+1/2P_beta(B >= nu/(nu+t^2):nu/2,1/2) when
            t > 0.0;

    (#) when nu >= 20, an asymptotic normalizing expansion of the
        Cornish--Fisher type is used to evaluate the probability, see
        Hill (1970).

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth

    Hill, G W, 1970, `Student's t-distribution`, Comm. ACM (13(10)),
    617--619
    """

    _return_scalar = _b_stat.prob_students_t(
        tail,
        t,
        df,
    )
    return _return_scalar

def prob_chisq(x, df, tail='L'):
    u"""
    Computes probabilities for chi^2 distribution.

    ``prob_chisq`` returns the lower or upper tail probability for the
    chi^2-distribution with real degrees of freedom.

    For full information please refer to the NAG Library document for
    g01ec

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01ecf.html

    Parameters
    ----------
    x : float
        x, the value of the chi^2 variate with nu degrees of freedom.

    df : float
        nu, the degrees of freedom of the chi^2-distribution.

    tail : str, length 1, optional
        Indicates whether the upper or lower tail probability is
        required.

        `tail` = 'L'

            The lower tail probability is returned, i.e., P(X <= x:nu).

        `tail` = 'U'

            The upper tail probability is returned, i.e., P(X >= x:nu).

    Returns
    -------
    p : float
        The lower or upper tail probability for the chi^2-distribution,
        depending on the value of `tail` supplied.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'L' or 'U'.

        (`errno` 2)
            On entry, `x` = *<value>*.

            Constraint: `x` >= 0.0.

        (`errno` 3)
            On entry, `df` = *<value>*.

            Constraint: `df` > 0.0.

        (`errno` 4)
            The series used to calculate the gamma probabilities has
            failed to converge. The result returned should represent an
            approximation to the solution.

    Notes
    -----
    The lower tail probability for the chi^2-distribution with nu
    degrees of freedom, P(X <= x:nu) is defined by:

        P(X <= x:nu) =
        1/(2^(nu/2)Gamma(nu/2))int_0.0^xX^(nu/2-1)e^(-X/2)dX, x >= 0,nu
        > 0.

    To calculate P(X <= x:nu) a transformation of a gamma distribution
    is employed, i.e., a chi^2-distribution with nu degrees of freedom
    is equal to a gamma distribution with scale parameter 2 and shape
    parameter nu/2.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth
    """

    _return_scalar = _b_stat.prob_chisq(
        tail,
        x,
        df,
    )
    return _return_scalar

def prob_f(f, df1, df2, tail='L'):
    u"""
    Computes probabilities for F-distribution.

    ``prob_f`` returns the probability for the lower or upper tail of
    the F or variance-ratio distribution with real degrees of freedom.

    For full information please refer to the NAG Library document for
    g01ed

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01edf.html

    Parameters
    ----------
    f : float
        f, the value of the F variate.

    df1 : float
        The degrees of freedom of the numerator variance, nu_1.

    df2 : float
        The degrees of freedom of the denominator variance, nu_2.

    tail : str, length 1, optional
        Indicates whether an upper or lower tail probability is
        required.

        `tail` = 'L'

            The lower tail probability is returned, i.e., P(F <=
            f:nu_1,nu_2).

        `tail` = 'U'

            The upper tail probability is returned, i.e., P(F >=
            f:nu_1,nu_2).

    Returns
    -------
    p : float
        The probability for the lower or upper tail of the F or
        variance-ratio distribution with real degrees of freedom.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'L' or 'U'.

        (`errno` 2)
            On entry, `f` = *<value>*.

            Constraint: `f` >= 0.0.

        (`errno` 3)
            On entry, `df1` = *<value>* and `df2` = *<value>*.

            Constraint: `df1` > 0.0 and `df2` > 0.0.

        (`errno` 4)
            The probability is too close to 0.0 or 1.0. `f` is too far
            out into the tails for the probability to be evaluated
            exactly. The result tends to approach 1.0 if f is large, or
            0.0 if f is small. The result returned is a good
            approximation to the required solution.

    Notes
    -----
    The lower tail probability for the F, or variance-ratio
    distribution, with nu_1 and nu_2 degrees of freedom, P(F <=
    f:nu_1,nu_2), is defined by:

        P(F <= f:nu_1,nu_2) = nu_1^nu_1/2 nu_2^nu_2/2
        Gamma((nu_1+nu_2)/2)/(Gamma(nu_1/2)Gamma(nu_2/2))int_0^fF^((nu_1-2)/2)(nu_1
        F+nu_2)^(-(nu_1+nu_2)/2)dF,

    for nu_1, nu_2 > 0, f >= 0.

    The probability is computed by means of a transformation to a beta
    distribution, P_beta(B <= beta:a,b):

        P(F <= f:nu_1,nu_2) = P_beta(B <= nu_1 f/(nu_1
        f+nu_2):nu_1/2,nu_2/2)

    and using a call to ``prob_beta``.

    For very large values of both nu_1 and nu_2, greater than 10^5, a
    normal approximation is used.
    If only one of nu_1 or nu_2 is greater than 10^5 then a chi^2
    approximation is used, see Abramowitz and Stegun (1972).

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth
    """

    _return_scalar = _b_stat.prob_f(
        tail,
        f,
        df1,
        df2,
    )
    return _return_scalar

ProbBetaReturnData = _collections.namedtuple(
    'ProbBetaReturnData',
    ['p', 'q', 'pdf'],
)

def prob_beta(x, a, b):
    u"""
    Computes upper and lower tail probabilities and probability density
    function for the beta distribution.

    ``prob_beta`` computes the upper and lower tail probabilities and
    the probability density function of the beta distribution with
    parameters a and b.

    For full information please refer to the NAG Library document for
    g01ee

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01eef.html

    Parameters
    ----------
    x : float
        beta, the value of the beta variate.

    a : float
        a, the first parameter of the required beta distribution.

    b : float
        b, the second parameter of the required beta distribution.

    Returns
    -------
    p : float
        The lower tail probability, P(B <= beta:a,b).

    q : float
        The upper tail probability, P(B >= beta:a,b).

    pdf : float
        The probability density function, f(B:a,b).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` >= 0.0.

        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` <= 1.0.

        (`errno` 2)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `a` > 0.0.

        (`errno` 2)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `b` > 0.0.

    Notes
    -----
    The probability density function of the beta distribution with
    parameters a and b is:

        f(B:a,b) = Gamma(a+b)/(Gamma(a)Gamma(b))B^(a-1)(1-B)^(b-1), 0 <=
        B <= 1;a,b > 0.

    The lower tail probability, P(B <= beta:a,b) is defined by

        P(B <= beta:a,b) = Gamma(a+b)/(Gamma(a)Gamma(b))int_0^beta
        B^(a-1)(1-B)^(b-1)dB = I_beta(a,b), 0 <= beta <= 1;a,b > 0.

    The function I_x(a,b), also known as the incomplete beta function is
    calculated using ``specfun.beta_incomplete``.

    References
    ----------
    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth
    """

    _return_scalars = _b_stat.prob_beta(
        x,
        a,
        b,
    )
    _p_py = _return_scalars.p
    _q_py = _return_scalars.q
    _pdf_py = _return_scalars.pdf
    _return_data = ProbBetaReturnData(
        _p_py,
        _q_py,
        _pdf_py,
    )
    return _return_data

ProbGammaReturnData = _collections.namedtuple(
    'ProbGammaReturnData',
    ['p'],
)

def prob_gamma(g, a, b, tail='L'):
    u"""
    Computes probabilities for the gamma distribution.

    ``prob_gamma`` returns the lower or upper tail probability of the
    gamma distribution, with parameters alpha and beta.

    For full information please refer to the NAG Library document for
    g01ef

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01eff.html

    Parameters
    ----------
    g : float
        g, the value of the gamma variate.

    a : float
        The parameter alpha of the gamma distribution.

    b : float
        The parameter beta of the gamma distribution.

    tail : str, length 1, optional
        Indicates whether an upper or lower tail probability is
        required.

        `tail` = 'L'

            The lower tail probability is returned, that is P(G <=
            g:alpha,beta).

        `tail` = 'U'

            The upper tail probability is returned, that is P(G >=
            g:alpha,beta).

    Returns
    -------
    p : float
        The lower or upper tail probability of the gamma distribution,
        with parameters alpha and beta.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'L' or 'U'.

        (`errno` 2)
            On entry, `g` = *<value>*.

            Constraint: `g` >= 0.0.

        (`errno` 3)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `a` > 0.0 and `b` > 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The algorithm has failed to converge in *<value>*
            iterations. The probability returned should be a reasonable
            approximation to the solution.

    Notes
    -----
    The lower tail probability for the gamma distribution with
    parameters alpha and beta, P(G <= g), is defined by:

        P(G <= g;alpha,beta) =
        1/(beta^alphaGamma(alpha))int_0^gG^(alpha-1)e^(-G/beta)dG, alpha
        > 0.0, beta > 0.0.

    The mean of the distribution is alphabeta and its variance is
    alphabeta^2.
    The transformation Z = G/beta is applied to yield the following
    incomplete gamma function in normalized form,

        P(G <= g;alpha,beta) = P(Z <= g/beta:alpha,1.0) =
        1/Gamma(alpha)int_0^g/beta Z^(alpha-1)e^(-Z)dZ.

    This is then evaluated using ``specfun.gamma_incomplete``.

    References
    ----------
    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth
    """

    try:
        _return_scalar = _b_stat.prob_gamma(
            tail,
            g,
            a,
            b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = ProbGammaReturnData(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

ProbStudentizedRangeReturnData = _collections.namedtuple(
    'ProbStudentizedRangeReturnData',
    ['p'],
)

def prob_studentized_range(q, v, ir):
    u"""
    Computes probability for the Studentized range statistic.

    ``prob_studentized_range`` returns the probability associated with
    the lower tail of the distribution of the Studentized range
    statistic.

    For full information please refer to the NAG Library document for
    g01em

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01emf.html

    Parameters
    ----------
    q : float
        q, the Studentized range statistic.

    v : float
        v, the number of degrees of freedom for the experimental error.

    ir : int
        r, the number of groups.

    Returns
    -------
    p : float
        The probability associated with the lower tail of the
        distribution of the Studentized range statistic.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `q` = *<value>*.

            Constraint: `q` > 0.0.

        (`errno` 1)
            On entry, `ir` = *<value>*.

            Constraint: `ir` >= 2.

        (`errno` 1)
            On entry, `v` = *<value>*.

            Constraint: `v` >= 1.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            There is some doubt as to whether full accuracy has been
            achieved. The returned value should be a reasonable estimate
            of the true value.

    Notes
    -----
    The externally Studentized range, q, for a sample, x_1,x_2,...,x_r,
    is defined as:

        q = (max(x_i)-min(x_i))/(sigma^_e),

    where sigma^_e is an independent estimate of the standard error of
    the x_i's.
    The most common use of this statistic is in the testing of means
    from a balanced design.
    In this case for a set of group means, T__1,T__2,...,T__r, the
    Studentized range statistic is defined to be the difference between
    the largest and smallest means, T__largest and T__smallest, divided
    by the square root of the mean-square experimental error, MS_error,
    over the number of observations in each group, n, i.e.,

        q = (T__largest-T__smallest)/sqrt(MS_error/n).

    The Studentized range statistic can be used as part of a multiple
    comparisons procedure such as the Newman--Keuls procedure or
    Duncan's multiple range test (see Montgomery (1984) and Winer
    (1970)).

    For a Studentized range statistic the probability integral,
    P(q;v,r), for v degrees of freedom and r groups can be written as:

        P(q;v,r) = Cint_0^infinity x^(v-1)e^(-vx^2/2){rint_-
        infinity^infinity phi(y)[Phi(y)-Phi(y-qx)]^(r-1)dy}dx,

    where

        C = (v^(v/2))/(Gamma(v/2)2^(v/2-1)), phi(y) = 1/(sqrt(2
        pi))e^(-y^2/2) and Phi(y) = int_- infinity^y phi(t)dt.

    The above two-dimensional integral is evaluated using
    ``quad.dim2_fin`` with the upper and lower limits computed to give
    stated accuracy (see [Accuracy]).

    If the degrees of freedom v are greater than 2000 the probability
    integral can be approximated by its asymptotic form:

        P(q;r) = rint_- infinity^infinity
        phi(y)[Phi(y)-Phi(y-q)]^(r-1)dy.

    This integral is evaluated using ``quad.dim1_inf``.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Lund, R E and Lund, J R, 1983, `Algorithm AS 190: probabilities and
    upper quartiles for the studentized range`, Appl. Statist. (32(2)),
    204--210

    Montgomery, D C, 1984, `Design and Analysis of Experiments`, Wiley

    Winer, B J, 1970, `Statistical Principles in Experimental Design`,
    McGraw--Hill
    """

    try:
        _return_scalar = _b_stat.prob_studentized_range(
            q,
            v,
            ir,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = ProbStudentizedRangeReturnData(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

ProbDurbinWatsonReturnData = _collections.namedtuple(
    'ProbDurbinWatsonReturnData',
    ['pdl', 'pdu'],
)

def prob_durbin_watson(n, ip, d):
    u"""
    Computes bounds for the significance of a Durbin--Watson statistic.

    ``prob_durbin_watson`` calculates upper and lower bounds for the
    significance of a Durbin--Watson statistic.

    For full information please refer to the NAG Library document for
    g01ep

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01epf.html

    Parameters
    ----------
    n : int
        n, the number of observations used in calculating the
        Durbin--Watson statistic.

    ip : int
        p, the number of independent variables in the regression model,
        including the mean.

    d : float
        d, the Durbin--Watson statistic.

    Returns
    -------
    pdl : float
        Lower bound for the significance of the Durbin--Watson
        statistic, p_l.

    pdu : float
        Upper bound for the significance of the Durbin--Watson
        statistic, p_u.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>* and `ip` = *<value>*.

            Constraint: `n` > `ip`.

        (`errno` 2)
            On entry, `d` = *<value>*.

            Constraint: `d` >= 0.0.

    Notes
    -----
    Let r = (r_1,r_2,...,r_n)^T be the residuals from a linear
    regression of y on p independent variables, including the mean,
    where the y values y_1,y_2,...,y_n can be considered as a time
    series.
    The Durbin--Watson test (see Durbin and Watson (1950), Durbin and
    Watson (1951) and Durbin and Watson (1971)) can be used to test for
    serial correlation in the error term in the regression.

    The Durbin--Watson test statistic is:

        d = sum_i = 1^n-1(r_{i+1}-r_i)^2/sum_i = 1^nr_i^2,

    which can be written as

        d = r^TAr/r^Tr,

    where the n*n matrix A is given by

        A = [[table omitted]]

    with the nonzero eigenvalues of the matrix A being lambda_j =
    (1-cos(pi j/n)), for j = 1,2,...,n-1.

    Durbin and Watson show that the exact distribution of d depends on
    the eigenvalues of a matrix HA, where H is the hat matrix of
    independent variables, i.e., the matrix such that the vector of
    fitted values, y^, can be written as y^= Hy.
    However, bounds on the distribution can be obtained, the lower bound
    being

        d_l = sum_i = 1^n-p lambda_i u_i^2/sum_i = 1^n-pu_i^2

    and the upper bound being

        d_u = sum_i = 1^n-p lambda_{i-1+p}u_i^2/sum_i = 1^n-pu_i^2,

    where u_i are independent standard Normal variables.

    Two algorithms are used to compute the lower tail (significance
    level) probabilities, p_l and p_u, associated with d_l and d_u.
    If n <= 60 the procedure due to Pan (1964) is used, see Farebrother
    (1980), otherwise Imhof's method (see Imhof (1961)) is used.

    The bounds are for the usual test of positive correlation; if a test
    of negative correlation is required the value of d should be
    replaced by 4-d.

    References
    ----------
    Durbin, J and Watson, G S, 1950, `Testing for serial correlation in
    least squares regression. I`, Biometrika (37), 409--428

    Durbin, J and Watson, G S, 1951, `Testing for serial correlation in
    least squares regression. II`, Biometrika (38), 159--178

    Durbin, J and Watson, G S, 1971, `Testing for serial correlation in
    least squares regression. III`, Biometrika (58), 1--19

    Farebrother, R W, 1980, `Algorithm AS 153. Pan's procedure for the
    tail probabilities of the Durbin--Watson statistic`, Appl. Statist.
    (29), 224--227

    Imhof, J P, 1961, `Computing the distribution of quadratic forms in
    Normal variables`, Biometrika (48), 419--426

    Newbold, P, 1988, `Statistics for Business and Economics`,
    Prentice--Hall

    Pan, Jie--Jian, 1964, `Distributions of the noncircular serial
    correlation coefficients`, Shuxue Jinzhan (7), 328--337
    """

    _return_scalars = _b_stat.prob_durbin_watson(
        n,
        ip,
        d,
    )
    _pdl_py = _return_scalars.pdl
    _pdu_py = _return_scalars.pdu
    _return_data = ProbDurbinWatsonReturnData(
        _pdl_py,
        _pdu_py,
    )
    return _return_data

def prob_vonmises(t, vk):
    u"""
    Computes probability for von Mises distribution.

    ``prob_vonmises`` returns the probability associated with the lower
    tail of the von Mises distribution between -pi and pi through the
    function name.

    For full information please refer to the NAG Library document for
    g01er

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01erf.html

    Parameters
    ----------
    t : float
        theta, the observed von Mises statistic measured in radians.

    vk : float
        The concentration parameter kappa, of the von Mises
        distribution.

    Returns
    -------
    p : float
        The probability associated with the lower tail of the von Mises
        distribution between -pi and pi.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `vk` = *<value>*.

            Constraint: `vk` >= 0.0.

    Notes
    -----
    The von Mises distribution is a symmetric distribution used in the
    analysis of circular data.
    The lower tail area of this distribution on the circle with mean
    direction mu_0 = 0 and concentration parameter kappa, kappa, can be
    written as

        Pr(Theta <= theta:kappa) = 1/(2 pi I_0(kappa))int_-pi^theta
        e^(kappa cos(Theta))d Theta,

    where theta is reduced modulo 2 pi so that -pi <= theta < pi and
    kappa >= 0.
    Note that if theta = pi then ``prob_vonmises`` returns a probability
    of 1.
    For very small kappa the distribution is almost the uniform
    distribution, whereas for kappa→ infinity all the probability is
    concentrated at one point.

    The method of calculation for small kappa involves backwards
    recursion through a series expansion in terms of modified Bessel
    functions, while for large kappa an asymptotic Normal approximation
    is used.

    In the case of small kappa the series expansion of Pr(Theta <=
    theta: kappa) can be expressed as

        Pr(Theta <= theta:kappa) = 1/2+theta/((2 pi))+1/(pi
        I_0(kappa))sum_n = 1^infinity n^-1I_n(kappa)sin(n)theta,

    where I_n(kappa) is the modified Bessel function.
    This series expansion can be represented as a nested expression of
    terms involving the modified Bessel function ratio R_n,

        R_n(kappa) = I_n(kappa)/I_{n-1}(kappa), n = 1,2,3,...,

    which is calculated using backwards recursion.

    For large values of kappa (see [Accuracy]) an asymptotic Normal
    approximation is used.
    The angle Theta is transformed to the nearly Normally distributed
    variate Z,

        Z = b(kappa)sin(Theta/2),

    where

        b(kappa) = sqrt(2/pi)e^kappa/I_0(kappa)

    and b(kappa) is computed from a continued fraction approximation.
    An approximation to order kappa^-4 of the asymptotic normalizing
    series for z is then used.
    Finally the Normal probability integral is evaluated.

    For a more detailed analysis of the methods used see Hill (1977).

    References
    ----------
    Hill, G W, 1977, `Algorithm 518: Incomplete Bessel function I_0: The
    Von Mises distribution`, ACM Trans. Math. Software (3), 279--284

    Mardia, K V, 1972, `Statistics of Directional Data`, Academic Press
    """

    _return_scalar = _b_stat.prob_vonmises(
        t,
        vk,
    )
    return _return_scalar

def prob_landau(x):
    u"""
    Landau distribution function.

    ``prob_landau`` returns the value of the Landau distribution
    function Phi(lambda).

    For full information please refer to the NAG Library document for
    g01et

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01etf.html

    Parameters
    ----------
    x : float
        The argument lambda of the function.

    Returns
    -------
    p : float
        The value of the Landau distribution function Phi(lambda).

    Notes
    -----
    ``prob_landau`` evaluates an approximation to the Landau
    distribution function Phi(lambda) given by

        Phi(lambda) = int_- infinity^lambdaphi(lambda)d lambda,

    where phi(lambda) is described in ``pdf_landau``, using piecewise
    approximation by rational functions.
    Further details can be found in Kölbig and Schorr (1984).

    References
    ----------
    Kölbig, K S and Schorr, B, 1984, `A program package for the Landau
    distribution`, Comp. Phys. Comm. (31), 97--111
    """

    _return_scalar = _b_stat.prob_landau(
        x,
    )
    return _return_scalar

def prob_vavilov(x, comm):
    u"""
    Vavilov distribution function.

    ``prob_vavilov`` returns the value of the Vavilov distribution
    function Phi_V(lambda;kappa,beta^2).

    It is intended to be used after a call to ``init_vavilov``.

    For full information please refer to the NAG Library document for
    g01eu

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01euf.html

    Parameters
    ----------
    x : float
        The argument lambda of the function.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``init_vavilov``.

    Returns
    -------
    p : float
        The value of the Vavilov distribution function
        Phi_V(lambda;kappa,beta^2).

    Raises
    ------
    NagValueError
        (`errno` 1)
            Either the initialization function has not been called prior
            to the first call of this function or a communication array
            has become corrupted.

    Notes
    -----
    ``prob_vavilov`` evaluates an approximation to the Vavilov
    distribution function Phi_V(lambda;kappa,beta^2) given by

        Phi_V(lambda;kappa,beta^2) = int_-
        infinity^lambdaphi_V(lambda;kappa,beta^2)d lambda,

    where phi(lambda) is described in ``pdf_vavilov``.
    The method used is based on Fourier expansions.
    Further details can be found in Schorr (1974).

    References
    ----------
    Schorr, B, 1974, `Programs for the Landau and the Vavilov
    distributions and the corresponding random numbers`, Comp. Phys.
    Comm. (7), 215--224
    """

    _return_scalar = _b_stat.prob_vavilov(
        x,
        comm,
    )
    return _return_scalar

ProbDickeyFullerUnitReturnData = _collections.namedtuple(
    'ProbDickeyFullerUnitReturnData',
    ['pint'],
)

def prob_dickey_fuller_unit(ts_type, n, ts, method=1, nsamp=100000, statecomm=None):
    u"""
    Computes probabilities for the Dickey--Fuller unit root test.

    ``prob_dickey_fuller_unit`` returns the probability associated with
    the lower tail of the distribution for the Dickey--Fuller unit root
    test statistic.

    For full information please refer to the NAG Library document for
    g01ew

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01ewf.html

    Parameters
    ----------
    ts_type : int
        The type of test statistic, supplied in `ts`.

    n : int
        n, the length of the time series used to calculate the test
        statistic.

    ts : float
        The Dickey--Fuller test statistic for which the probability is
        required. If

        `ts_type` = 1

            `ts` must contain tau.

        `ts_type` = 2

            `ts` must contain tau_mu.

        `ts_type` = 3

            `ts` must contain tau_tau.

        If the test statistic was calculated using
        ``tsa.uni_dickey_fuller_unit`` the value of `ts_type` and `n`
        must not change between calls to ``tsa.uni_dickey_fuller_unit``
        and ``prob_dickey_fuller_unit``.

    method : int, optional
        The method used to calculate the probability.

        `method` = 1

            The probability is interpolated from a look-up table, whose
            values were obtained via simulation.

        `method` = 2

            The probability is interpolated from a look-up table, whose
            values were obtained from Dickey (1976).

        `method` = 3

            The probability is obtained via simulation.

        The probability calculated from the look-up table should give
        sufficient accuracy for most applications.

    nsamp : int, optional
        If `method` = 3, the number of samples used in the simulation;
        otherwise `nsamp` is not referenced and need not be set.

    statecomm : None or dict, RNG communication object, optional,
    modified in place
        RNG communication structure.

        When `method` = 3, this argument must have been initialized by a
        prior call to ``rand.init_repeat`` or ``rand.init_nonrepeat``.

    Returns
    -------
    pint : float
        The probability associated with the lower tail of the
        distribution for the (augmented) Dickey--Fuller unit root test
        statistic supplied in `ts`.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `method` = *<value>*.

            Constraint: `method` = 1, 2 or 3.

        (`errno` 21)
            On entry, `ts_type` = *<value>*.

            Constraint: `ts_type` = 1, 2 or 3.

        (`errno` 31)
            On entry, `n` = *<value>*.

            Constraint: if `method` != 3, `n` > 0.

        (`errno` 31)
            On entry, `n` = *<value>*.

            Constraint: if `method` = 3 and `ts_type` = 1, `n` > 2.

        (`errno` 31)
            On entry, `n` = *<value>*.

            Constraint: if `method` = 3 and `ts_type` = 2, `n` > 3.

        (`errno` 31)
            On entry, `n` = *<value>*.

            Constraint: if `method` = 3 and `ts_type` = 3, `n` > 4.

        (`errno` 51)
            On entry, `nsamp` = *<value>*.

            Constraint: if `method` = 3, `nsamp` > 0.

        (`errno` 61)
            On entry, `method` = 3 and the `statecomm`['state'] vector
            has been corrupted or not initialized.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 201)
            The supplied input values were outside the range of at least
            one look-up table, therefore, extrapolation was used.

    Notes
    -----
    If the root of the characteristic equation for a time series is one
    then that series is said to have a unit root.
    Such series are nonstationary. ``prob_dickey_fuller_unit`` is
    designed to be called after ``tsa.uni_dickey_fuller_unit`` and
    returns the probability associated with one of three types of
    (augmented) Dickey--Fuller test statistic: tau, tau_mu or tau_tau,
    used to test for a unit root, a unit root with drift or a unit root
    with drift and a deterministic time trend, respectively.
    The three types of test statistic are constructed as follows:

    (1) To test whether a time series, y_t, for t = 1,2,...,n, has a
        unit root the regression model

            nabla y_t = beta_1 y_{t-1}+sum_i = 1^p-1 delta_inabla
            y_{t-i}+epsilon_t

        is fit and the test statistic tau constructed as

            tau = (beta^_1)/(sigma_{11})

        where nabla is the difference operator, with nabla y_t =
        y_t-y_{t-1}, and where beta^_1 and sigma_{11} are the least
        squares estimate and associated standard error for beta_1
        respectively.

    (#) To test for a unit root with drift the regression model

            nabla y_t = beta_1 y_{t-1}+sum_i = 1^p-1 delta_inabla
            y_{t-i}+alpha+epsilon_t

        is fit and the test statistic tau_mu constructed as

            tau_mu = (beta^_1)/(sigma_{11}).

    (#) To test for a unit root with drift and deterministic time trend
        the regression model

            nabla y_t = beta_1 y_{t-1}+sum_i = 1^p-1 delta_inabla
            y_{t-i}+alpha+beta_2 t+epsilon_t

        is fit and the test statistic tau_tau constructed as

            tau_tau = (beta^_1)/(sigma_{11}).

    All three test statistics: tau, tau_mu and tau_tau can be calculated
    using ``tsa.uni_dickey_fuller_unit``.

    The probability distributions of these statistics are nonstandard
    and are a function of the length of the series of interest, n.
    The probability associated with a given test statistic, for a given
    n, can, therefore, only be calculated by simulation as described in
    Dickey and Fuller (1979).
    However, such simulations require a significant number of iterations
    and are, therefore, prohibitively expensive in terms of the time
    taken.
    As such ``prob_dickey_fuller_unit`` also allows the probability to
    be interpolated from a look-up table.
    Two such tables are provided, one from Dickey (1976) and one
    constructed as described in [Further Comments].
    The three different methods of obtaining an estimate of the
    probability can be chosen via the `method` argument.
    Unless there is a specific reason for choosing otherwise, `method` =
    1 should be used.

    References
    ----------
    Dickey, A D, 1976, `Estimation and hypothesis testing in
    nonstationary time series`, PhD Thesis, Iowa State University, Ames,
    Iowa

    Dickey, A D and Fuller, W A, 1979, `Distribution of the estimators
    for autoregressive time series with a unit root`, J. Am. Stat.
    Assoc. (74 366), 427--431
    """

    fun_name = 'naginterfaces.library.stat.prob_dickey_fuller_unit'

    _method = _b_utils._EngineIntScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'},
    ).to_ctypes()
    _method_py = _b_utils._EngineIntScalarType.to_py(_method)
    if _method_py == 3:
        if statecomm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'statecomm'},
                'must not be None when method = 3',
            )
    try:
        _return_scalar = _b_stat.prob_dickey_fuller_unit(
            _method,
            ts_type,
            n,
            ts,
            nsamp,
            statecomm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = ProbDickeyFullerUnitReturnData(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def prob_kolmogorov1(n, d):
    u"""
    Computes probabilities for the one-sample Kolmogorov--Smirnov
    distribution.

    ``prob_kolmogorov1`` returns the upper tail probability associated
    with the one sample Kolmogorov--Smirnov distribution.

    For full information please refer to the NAG Library document for
    g01ey

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01eyf.html

    Parameters
    ----------
    n : int
        n, the number of observations in the sample.

    d : float
        Contains the test statistic, D_n^+ or D_n^-.

    Returns
    -------
    p : float
        The upper tail probability associated with the one sample
        Kolmogorov--Smirnov distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, `d` < 0.0 or `d` > 1.0: `d` = *<value>*.

    Notes
    -----
    Let S_n(x) be the sample cumulative distribution function and F_0(x)
    the hypothesised theoretical distribution function.

    ``prob_kolmogorov1`` returns the upper tail probability, p,
    associated with the one-sided Kolmogorov--Smirnov test statistic
    D_n^+ or D_n^-, where these one-sided statistics are defined as
    follows;

        [table omitted]

    If n <= 100 an exact method is used; for the details see Conover
    (1980).
    Otherwise a large sample approximation derived by Smirnov is used;
    see Feller (1948), Kendall and Stuart (1973) and Smirnov (1948).

    References
    ----------
    Conover, W J, 1980, `Practical Nonparametric Statistics`, Wiley

    Feller, W, 1948, `On the Kolmogorov--Smirnov limit theorems for
    empirical distributions`, Ann. Math. Statist. (19), 179--181

    Kendall, M G and Stuart, A, 1973, `The Advanced Theory of Statistics
    (Volume 2)`, (3rd Edition), Griffin

    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill

    Smirnov, N, 1948, `Table for estimating the goodness of fit of
    empirical distributions`, Ann. Math. Statist. (19), 279--281
    """

    _return_scalar = _b_stat.prob_kolmogorov1(
        n,
        d,
    )
    return _return_scalar

def prob_kolmogorov2(n1, n2, d):
    u"""
    Computes probabilities for the two-sample Kolmogorov--Smirnov
    distribution.

    ``prob_kolmogorov2`` returns the probability associated with the
    upper tail of the Kolmogorov--Smirnov two sample distribution.

    For full information please refer to the NAG Library document for
    g01ez

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01ezf.html

    Parameters
    ----------
    n1 : int
        The number of observations in the first sample, n_1.

    n2 : int
        The number of observations in the second sample, n_2.

    d : float
        The test statistic D_{n_1,n_2}, for the two sample
        Kolmogorov--Smirnov goodness-of-fit test, that is the maximum
        difference between the empirical cumulative distribution
        functions (CDFs) of the two samples.

    Returns
    -------
    p : float
        The probability associated with the upper tail of the
        Kolmogorov--Smirnov two sample distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n1` = *<value>* and `n2` = *<value>*.

            Constraint: `n1` >= 1 and `n2` >= 1.

        (`errno` 2)
            On entry, `d` < 0.0 or `d` > 1.0: `d` = *<value>*.

        (`errno` 3)
            The Smirnov approximation used for large samples did not
            converge in 200 iterations. The probability is set to 1.0.

    Notes
    -----
    Let F_n_1(x) and G_n_2(x) denote the empirical cumulative
    distribution functions for the two samples, where n_1 and n_2 are
    the sizes of the first and second samples respectively.

    The function ``prob_kolmogorov2`` computes the upper tail
    probability for the Kolmogorov--Smirnov two sample two-sided test
    statistic D_{n_1,n_2}, where

        D_{n_1,n_2} = sup_x|F_n_1(x)-G_n_2(x)|.

    The probability is computed exactly if n_1,n_2 <= 10000 and max(n_1,
    n_2) <= 2500 using a method given by Kim and Jenrich (1973).
    For the case where min(n_1, n_2) <= 10% of the max(n_1, n_2) and
    min(n_1, n_2) <= 80 the Smirnov approximation is used.
    For all other cases the Kolmogorov approximation is used.
    These two approximations are discussed in Kim and Jenrich (1973).

    References
    ----------
    Conover, W J, 1980, `Practical Nonparametric Statistics`, Wiley

    Feller, W, 1948, `On the Kolmogorov--Smirnov limit theorems for
    empirical distributions`, Ann. Math. Statist. (19), 179--181

    Kendall, M G and Stuart, A, 1973, `The Advanced Theory of Statistics
    (Volume 2)`, (3rd Edition), Griffin

    Kim, P J and Jenrich, R I, 1973, `Tables of exact sampling
    distribution of the two sample Kolmogorov--Smirnov criterion
    D_{mn}(m < n)`, Selected Tables in Mathematical Statistics (1),
    80--129, American Mathematical Society

    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill

    Smirnov, N, 1948, `Table for estimating the goodness of fit of
    empirical distributions`, Ann. Math. Statist. (19), 279--281
    """

    _return_scalar = _b_stat.prob_kolmogorov2(
        n1,
        n2,
        d,
    )
    return _return_scalar

def inv_cdf_normal(p, tail='L'):
    u"""
    Computes deviates for the standard Normal distribution.

    ``inv_cdf_normal`` returns the deviate associated with the given
    probability of the standard Normal distribution.

    For full information please refer to the NAG Library document for
    g01fa

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01faf.html

    Parameters
    ----------
    p : float
        p, the probability from the standard Normal distribution as
        defined by `tail`.

    tail : str, length 1, optional
        Indicates which tail the supplied probability represents.

        `tail` = 'L'

            The lower probability, i.e., P(X <= x_p).

        `tail` = 'U'

            The upper probability, i.e., P(X >= x_p).

        `tail` = 'S'

            The two tail (significance level) probability, i.e., P(X >=
            |x_p|)+P(X <= -|x_p|).

        `tail` = 'C'

            The two tail (confidence interval) probability, i.e., P(X <=
            |x_p|)-P(X <= -|x_p|).

    Returns
    -------
    x : float
        The deviate associated with the given probability of the
        standard Normal distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'L', 'U', 'S' or 'C'.

        (`errno` 2)
            On entry, `p` = *<value>*.

            Constraint: `p` < 1.0.

        (`errno` 2)
            On entry, `p` = *<value>*.

            Constraint: `p` > 0.0.

    Notes
    -----
    The deviate, x_p associated with the lower tail probability, p, for
    the standard Normal distribution is defined as the solution to

        P(X <= x_p) = p = int_- infinity^x_pZ(X)dX,

    where

        Z(X) = 1/(sqrt(2 pi))e^(-X^2/2), - infinity < X < infinity.

    The method used is an extension of that of Wichura (1988). p is
    first replaced by q = p-0.5.

    (a) If |q| <= 0.3, x_p is computed by a rational Chebyshev
        approximation

            x_p = sA(s^2)/B(s^2),

        where s = sqrt(2 pi)q and A, B are polynomials of degree 7.

    (#) If 0.3 < |q| <= 0.42, x_p is computed by a rational Chebyshev
        approximation

            x_p = sign(q)(C(t)/D(t)),

        where t = |q|-0.3 and C, D are polynomials of degree 5.

    (#) If |q| > 0.42, x_p is computed as

            x_p = sign(q)[(E(u)/F(u))+u],

        where u = sqrt(-2*log(min(p, 1-p))) and E, F are polynomials of
        degree 6.

    For the upper tail probability -x_p is returned, while for the two
    tail probabilities the value x_p* is returned, where p* is the
    required tail probability computed from the input value of p.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth

    Wichura, 1988, `Algorithm AS 241: the percentage points of the
    Normal distribution`, Appl. Statist. (37), 477--484
    """

    _return_scalar = _b_stat.inv_cdf_normal(
        tail,
        p,
    )
    return _return_scalar

InvCdfStudentsTReturnData = _collections.namedtuple(
    'InvCdfStudentsTReturnData',
    ['x'],
)

def inv_cdf_students_t(p, df, tail='L'):
    u"""
    Computes deviates for Student's t-distribution.

    ``inv_cdf_students_t`` returns the deviate associated with the given
    tail probability of Student's t-distribution with real degrees of
    freedom.

    For full information please refer to the NAG Library document for
    g01fb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01fbf.html

    Parameters
    ----------
    p : float
        p, the probability from the required Student's t-distribution as
        defined by `tail`.

    df : float
        nu, the degrees of freedom of the Student's t-distribution.

    tail : str, length 1, optional
        Indicates which tail the supplied probability represents.

        `tail` = 'U'

            The upper tail probability, i.e., P(T >= t_p:nu).

        `tail` = 'L'

            The lower tail probability, i.e., P(T <= t_p:nu).

        `tail` = 'S'

            The two tail (significance level) probability, i.e., P(T >=
            |t_p|:nu)+P(T <= -|t_p|:nu).

        `tail` = 'C'

            The two tail (confidence interval) probability, i.e., P(T <=
            |t_p|:nu)-P(T <= -|t_p|:nu).

    Returns
    -------
    x : float
        The deviate associated with the given tail probability of
        Student's t-distribution with real degrees of freedom.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'L', 'U', 'S' or 'C'.

        (`errno` 2)
            On entry, `p` = *<value>*.

            Constraint: `p` < 1.0.

        (`errno` 2)
            On entry, `p` = *<value>*.

            Constraint: `p` > 0.0.

        (`errno` 3)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 1.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The solution has failed to converge. However, the result
            should be a reasonable approximation.

    Notes
    -----
    The deviate, t_p associated with the lower tail probability, p, of
    the Student's t-distribution with nu degrees of freedom is defined
    as the solution to

        P(T < t_p:nu) = p = Gamma((nu+1)/2)/(sqrt(nupi)Gamma(nu/2))int_-
        infinity^t_p(1+(T^2)/nu)^(-(nu+1)/2)dT, nu >= 1; - infinity <
        t_p < infinity.

    For nu = 1 or 2 the integral equation is easily solved for t_p.

    For other values of nu < 3 a transformation to the beta distribution
    is used and the result obtained from ``inv_cdf_beta``.

    For nu >= 3 an inverse asymptotic expansion of Cornish--Fisher type
    is used.
    The algorithm is described by Hill (1970).

    References
    ----------
    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth

    Hill, G W, 1970, `Student's t-distribution`, Comm. ACM (13(10)),
    617--619
    """

    try:
        _return_scalar = _b_stat.inv_cdf_students_t(
            tail,
            p,
            df,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = InvCdfStudentsTReturnData(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

InvCdfChisqReturnData = _collections.namedtuple(
    'InvCdfChisqReturnData',
    ['x'],
)

def inv_cdf_chisq(p, df):
    u"""
    Computes deviates for the chi^2 distribution.

    ``inv_cdf_chisq`` returns the deviate associated with the given
    lower tail probability of the chi^2-distribution with real degrees
    of freedom.

    For full information please refer to the NAG Library document for
    g01fc

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01fcf.html

    Parameters
    ----------
    p : float
        p, the lower tail probability from the required
        chi^2-distribution.

    df : float
        nu, the degrees of freedom of the chi^2-distribution.

    Returns
    -------
    x : float
        The deviate associated with the given lower tail probability of
        the chi^2-distribution with real degrees of freedom.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` < 1.0.

        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` >= 0.0.

        (`errno` 2)
            On entry, `df` = *<value>*.

            Constraint: `df` > 0.0.

        (`errno` 3)
            The probability is too close to 0.0 or 1.0.

        (`errno` 5)
            The series used to calculate the gamma function has failed
            to converge. This is an unlikely error exit.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The algorithm has failed to converge in *<value>*
            iterations. The result should be a reasonable approximation.

    Notes
    -----
    The deviate, x_p, associated with the lower tail probability p of
    the chi^2-distribution with nu degrees of freedom is defined as the
    solution to

        P(X <= x_p:nu) = p =
        1/(2^(nu/2)Gamma(nu/2))int_0^x_pe^(-X/2)X^(v/2-1)dX, 0 <= x_p <
        infinity ;nu > 0.

    The required x_p is found by using the relationship between a
    chi^2-distribution and a gamma distribution, i.e., a
    chi^2-distribution with nu degrees of freedom is equal to a gamma
    distribution with scale parameter 2 and shape parameter nu/2.

    For very large values of nu, greater than 10^5, Wilson and
    Hilferty's normal approximation to the chi^2 is used; see Kendall
    and Stuart (1969).

    References
    ----------
    Best, D J and Roberts, D E, 1975, `Algorithm AS 91. The percentage
    points of the chi^2 distribution`, Appl. Statist. (24), 385--388

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth

    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin
    """

    try:
        _return_scalar = _b_stat.inv_cdf_chisq(
            p,
            df,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = InvCdfChisqReturnData(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

InvCdfFReturnData = _collections.namedtuple(
    'InvCdfFReturnData',
    ['x'],
)

def inv_cdf_f(p, df1, df2):
    u"""
    Computes deviates for the F-distribution.

    ``inv_cdf_f`` returns the deviate associated with the given lower
    tail probability of the F or variance-ratio distribution with real
    degrees of freedom.

    For full information please refer to the NAG Library document for
    g01fd

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01fdf.html

    Parameters
    ----------
    p : float
        p, the lower tail probability from the required F-distribution.

    df1 : float
        The degrees of freedom of the numerator variance, nu_1.

    df2 : float
        The degrees of freedom of the denominator variance, nu_2.

    Returns
    -------
    x : float
        The deviate associated with the given lower tail probability of
        the F or variance-ratio distribution with real degrees of
        freedom.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` < 1.0.

        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` >= 0.0.

        (`errno` 2)
            On entry, `df1` = *<value>* and `df2` = *<value>*.

            Constraint: `df1` > 0.0 and `df2` > 0.0.

        (`errno` 4)
            The probability is too close to 0.0 or 1.0. The value of f_p
            cannot be computed. This will only occur when the large
            sample approximations are used.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The solution has failed to converge. However, the result
            should be a reasonable approximation. Alternatively,
            ``inv_cdf_beta`` can be used with a suitable setting of the
            argument tol.

    Notes
    -----
    The deviate, f_p, associated with the lower tail probability, p, of
    the F-distribution with degrees of freedom nu_1 and nu_2 is defined
    as the solution to

        P(F <= f_p:nu_1,nu_2) = p = nu_1^1/2 nu_1 nu_2^1/2 nu_2
        Gamma((nu_1+nu_2)/2)/(Gamma((nu_1)/2)Gamma((nu_2)/2))int_0^f_pF^(1/2(nu_1-2))(nu_2+nu_1
        F)^(-1/2(nu_1+nu_2))dF,

    where nu_1,nu_2 > 0; 0 <= f_p < infinity.

    The value of f_p is computed by means of a transformation to a beta
    distribution, P_beta(B <= beta:a,b):

        P(F <= f:nu_1,nu_2) = P_beta(B <= nu_1 f/(nu_1
        f+nu_2):nu_1/2,nu_2/2)

    and using a call to ``inv_cdf_beta``.

    For very large values of both nu_1 and nu_2, greater than 10^5, a
    normal approximation is used.
    If only one of nu_1 or nu_2 is greater than 10^5 then a chi^2
    approximation is used; see Abramowitz and Stegun (1972).

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth
    """

    try:
        _return_scalar = _b_stat.inv_cdf_f(
            p,
            df1,
            df2,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = InvCdfFReturnData(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

InvCdfBetaReturnData = _collections.namedtuple(
    'InvCdfBetaReturnData',
    ['x'],
)

def inv_cdf_beta(p, a, b, tol=0.0):
    u"""
    Computes deviates for the beta distribution.

    ``inv_cdf_beta`` returns the deviate associated with the given lower
    tail probability of the beta distribution.

    For full information please refer to the NAG Library document for
    g01fe

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01fef.html

    Parameters
    ----------
    p : float
        p, the lower tail probability from the required beta
        distribution.

    a : float
        a, the first parameter of the required beta distribution.

    b : float
        b, the second parameter of the required beta distribution.

    tol : float, optional
        The relative accuracy required by you in the result. If
        ``inv_cdf_beta`` is entered with `tol` greater than or equal to
        1.0 or less than 10*machine precision (see
        ``machine.precision``), the value of 10*machine precision is
        used instead.

    Returns
    -------
    x : float
        The deviate associated with the given lower tail probability of
        the beta distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` >= 0.0.

        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` <= 1.0.

        (`errno` 2)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `a` > 0.0.

        (`errno` 2)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `a` <= 10^6.

        (`errno` 2)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `b` > 0.0.

        (`errno` 2)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `b` <= 10^6.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The solution has failed to converge. However, the result
            should be a reasonable approximation. Requested accuracy not
            achieved when calculating beta probability. You should try
            setting `tol` larger.

        (`errno` 4)
            The requested accuracy has not been achieved. Use a larger
            value of `tol`. There is doubt concerning the accuracy of
            the computed result. 100 iterations of the Newton--Raphson
            method have been performed without satisfying the accuracy
            criterion (see [Further Comments]). The result should be a
            reasonable approximation of the solution.

    Notes
    -----
    The deviate, beta_p, associated with the lower tail probability, p,
    of the beta distribution with parameters a and b is defined as the
    solution to

        P(B <= beta_p:a,b) = p =
        Gamma(a+b)/(Gamma(a)Gamma(b))int_0^beta_pB^(a-1)(1-B)^(b-1)dB, 0
        <= beta_p <= 1;a,b > 0.

    The algorithm is a modified version of the Newton--Raphson method,
    following closely that of Cran `et al.` (1977).

    An initial approximation, beta_0, to beta_p is found (see Cran `et
    al.` (1977)), and the Newton--Raphson iteration

        beta_i = beta_{i-1}-f(beta_{i-1})/f'(beta_{i-1}),

    where f(beta) = P(B <= beta:a,b)-p is used, with modifications to
    ensure that beta remains in the range (0,1).

    References
    ----------
    Cran, G W, Martin, K J and Thomas, G E, 1977, `Algorithm AS 109.
    Inverse of the incomplete beta function ratio`, Appl. Statist. (26),
    111--114

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth
    """

    try:
        _return_scalar = _b_stat.inv_cdf_beta(
            p,
            a,
            b,
            tol,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = InvCdfBetaReturnData(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

InvCdfGammaReturnData = _collections.namedtuple(
    'InvCdfGammaReturnData',
    ['x'],
)

def inv_cdf_gamma(p, a, b, tol=0.0):
    u"""
    Computes deviates for the gamma distribution.

    ``inv_cdf_gamma`` returns the deviate associated with the given
    lower tail probability of the gamma distribution.

    For full information please refer to the NAG Library document for
    g01ff

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01fff.html

    Parameters
    ----------
    p : float
        p, the lower tail probability from the required gamma
        distribution.

    a : float
        alpha, the shape parameter of the gamma distribution.

    b : float
        beta, the scale parameter of the gamma distribution.

    tol : float, optional
        The relative accuracy required by you in the results. The
        smallest recommended value is 50*delta, where delta =
        max(10^-18, machine precision). If ``inv_cdf_gamma`` is entered
        with `tol` less than 50*delta or greater or equal to 1.0, then
        50*delta is used instead.

    Returns
    -------
    x : float
        The deviate associated with the given lower tail probability of
        the gamma distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` < 1.0.

        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` >= 0.0.

        (`errno` 2)
            On entry, `a` = *<value>*.

            Constraint: `a` <= 10^6.

        (`errno` 2)
            On entry, `b` = *<value>*.

            Constraint: `b` > 0.0.

        (`errno` 2)
            On entry, `a` = *<value>*.

            Constraint: `a` > 0.0.

        (`errno` 3)
            The probability is too close to 0.0 for the given `a` to
            enable the result to be calculated.

        (`errno` 5)
            The series used to calculate the gamma function has failed
            to converge. This is an unlikely error exit.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The algorithm has failed to converge in 100 iterations. A
            larger value of `tol` should be tried. The result may be a
            reasonable approximation.

    Notes
    -----
    The deviate, g_p, associated with the lower tail probability, p, of
    the gamma distribution with shape parameter alpha and scale
    parameter beta, is defined as the solution to

        P(G <= g_p:alpha,beta) = p =
        1/(beta^alphaGamma(alpha))int_0^g_pe^(-G/beta)G^(alpha-1)dG, 0
        <= g_p < infinity ;alpha,beta > 0.

    The method used is described by Best and Roberts (1975) making use
    of the relationship between the gamma distribution and the
    chi^2-distribution.

    Let y = 2(g_p)/beta.
    The required y is found from the Taylor series expansion

        y = y_0+sum_r C_r(y_0)/r!(E/phi(y_0))^r,

    where y_0 is a starting approximation

        C_1(u) = 1,

        C_{r+1}(u) = (r Psi+d/du)C_r(u),

        Psi = 1/2-(alpha-1)/u,

        E = p-int_0^y_0 phi(u)du,

        phi(u) = 1/(2^alphaGamma(alpha))e^(-u/2)u^(alpha-1).

    For most values of p and alpha the starting value

        y_01 = 2 alpha(z*sqrt(1/9 alpha)+1-1/9 alpha)^3

    is used, where z is the deviate associated with a lower tail
    probability of p for the standard Normal distribution.

    For p close to zero,

        y_02 = (p alpha 2^alphaGamma(alpha))^(1/alpha)

    is used.

    For large p values, when y_01 > 4.4 alpha+6.0,

        y_03 = -2*[ln(1-p)-(alpha-1)ln(1/2y_01)+ln(Gamma(alpha))]

    is found to be a better starting value than y_01.

    For small alpha (alpha <= 0.16), p is expressed in terms of an
    approximation to the exponential integral and y_04 is found by
    Newton--Raphson iterations.

    Seven terms of the Taylor series are used to refine the starting
    approximation, repeating the process if necessary until the required
    accuracy is obtained.

    References
    ----------
    Best, D J and Roberts, D E, 1975, `Algorithm AS 91. The percentage
    points of the chi^2 distribution`, Appl. Statist. (24), 385--388
    """

    try:
        _return_scalar = _b_stat.inv_cdf_gamma(
            p,
            a,
            b,
            tol,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = InvCdfGammaReturnData(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

InvCdfStudentizedRangeReturnData = _collections.namedtuple(
    'InvCdfStudentizedRangeReturnData',
    ['x'],
)

def inv_cdf_studentized_range(p, v, ir):
    u"""
    Computes deviates for the Studentized range statistic.

    ``inv_cdf_studentized_range`` returns the deviate associated with
    the lower tail probability of the distribution of the Studentized
    range statistic.

    For full information please refer to the NAG Library document for
    g01fm

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01fmf.html

    Parameters
    ----------
    p : float
        The lower tail probability for the Studentized range statistic,
        p_0.

    v : float
        v, the number of degrees of freedom.

    ir : int
        r, the number of groups.

    Returns
    -------
    x : float
        The deviate associated with the lower tail probability of the
        distribution of the Studentized range statistic.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: 0.0 < `p` < 1.0.

        (`errno` 1)
            On entry, `ir` = *<value>*.

            Constraint: `ir` >= 2.

        (`errno` 1)
            On entry, `v` = *<value>*.

            Constraint: `v` >= 1.0.

        (`errno` 2)
            The function was unable to find an upper bound for the value
            of q_0. This will be caused by p_0 being too close to 1.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            There is some doubt as to whether full accuracy has been
            achieved. The returned value should be a reasonable estimate
            of the true value.

    Notes
    -----
    The externally Studentized range, q, for a sample, x_1,x_2,...,x_r,
    is defined as

        q = (max(x_i)-min(x_i))/(sigma^_e),

    where sigma^_e is an independent estimate of the standard error of
    the x_i.
    The most common use of this statistic is in the testing of means
    from a balanced design.
    In this case for a set of group means, T__1,T__2,...,T__r, the
    Studentized range statistic is defined to be the difference between
    the largest and smallest means, T__largest and T__smallest, divided
    by the square root of the mean-square experimental error, MS_error,
    over the number of observations in each group, n, i.e.,

        q = (T__largest-T__smallest)/sqrt(MS_error/n).

    The Studentized range statistic can be used as part of a multiple
    comparisons procedure such as the Newman--Keuls procedure or
    Duncan's multiple range test (see Montgomery (1984) and Winer
    (1970)).

    For a Studentized range statistic the probability integral,
    P(q;v,r), for v degrees of freedom and r groups, can be written as:

        P(q;v,r) = Cint_0^infinity x^(v-1)e^(-vx^2/2)(rint_-
        infinity^infinity phi(y)(Phi(y)-Phi(y-qx))^(r-1)dy)dx,

    where

        C = (v^(v/2))/(Gamma(v/2)2^(v/2-1)), phi(y) = 1/(sqrt(2
        pi))e^(-y^2/2) and Phi(y) = int_- infinity^y phi(t)dt.

    For a given probability p_0, the deviate q_0 is found as the
    solution to the equation

        P(q_0;v,r) = p_0,

    using ``roots.contfn_brent_rcomm``.
    Initial estimates are found using the approximation given in Lund
    and Lund (1983) and a simple search procedure.

    References
    ----------
    Lund, R E and Lund, J R, 1983, `Algorithm AS 190: probabilities and
    upper quartiles for the studentized range`, Appl. Statist. (32(2)),
    204--210

    Montgomery, D C, 1984, `Design and Analysis of Experiments`, Wiley

    Winer, B J, 1970, `Statistical Principles in Experimental Design`,
    McGraw--Hill
    """

    try:
        _return_scalar = _b_stat.inv_cdf_studentized_range(
            p,
            v,
            ir,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = InvCdfStudentizedRangeReturnData(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def inv_cdf_landau(x):
    u"""
    Landau inverse function Psi(x).

    ``inv_cdf_landau`` returns the value of the inverse Phi^-1(x) of the
    Landau distribution function.

    For full information please refer to the NAG Library document for
    g01ft

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01ftf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    invlan : float
        The value of the inverse Phi^-1(x) of the Landau distribution
        function.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` < 1.0.

        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` > 0.0.

    Notes
    -----
    ``inv_cdf_landau`` evaluates an approximation to the inverse
    Phi^-1(x) of the Landau distribution function given by

        Psi(x) = Phi^-1(x)

    (where Phi(lambda) is described in ``prob_landau`` and
    ``pdf_landau``), using either linear or quadratic interpolation or
    rational approximations which mimic the asymptotic behaviour.
    Further details can be found in Kölbig and Schorr (1984).

    It can also be used to generate Landau distributed random numbers in
    the range 0 < x < 1.

    References
    ----------
    Kölbig, K S and Schorr, B, 1984, `A program package for the Landau
    distribution`, Comp. Phys. Comm. (31), 97--111
    """

    _return_scalar = _b_stat.inv_cdf_landau(
        x,
    )
    return _return_scalar

ProbStudentsTNoncentralReturnData = _collections.namedtuple(
    'ProbStudentsTNoncentralReturnData',
    ['p'],
)

def prob_students_t_noncentral(t, df, delta, tol=0.0, maxit=100):
    u"""
    Computes probabilities for the non-central Student's t-distribution.

    ``prob_students_t_noncentral`` returns the lower tail probability
    for the noncentral Student's t-distribution.

    For full information please refer to the NAG Library document for
    g01gb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01gbf.html

    Parameters
    ----------
    t : float
        t, the deviate from the Student's t-distribution with nu degrees
        of freedom.

    df : float
        nu, the degrees of freedom of the Student's t-distribution.

    delta : float
        delta, the noncentrality parameter of the Students
        t-distribution.

    tol : float, optional
        The absolute accuracy required by you in the results. If
        ``prob_students_t_noncentral`` is entered with `tol` greater
        than or equal to 1.0 or less than 10*machine precision (see
        ``machine.precision``), the value of 10*machine precision is
        used instead.

    maxit : int, optional
        The maximum number of terms that are used in each of the
        summations.

    Returns
    -------
    p : float
        The lower tail probability for the noncentral Student's
        t-distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 1.0.

        (`errno` 2)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 1.

        (`errno` 4)
            Unable to calculate the probability as it is too close to
            zero or one.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            One of the series has failed to converge with `maxit` =
            *<value>* and `tol` = *<value>*. Reconsider the requested
            tolerance and/or the maximum number of iterations.

        (`errno` 4)
            The probability is too close to 0 or 1. The returned value
            should be a reasonable estimate of the true value.

    Notes
    -----
    The lower tail probability of the noncentral Student's
    t-distribution with nu degrees of freedom and noncentrality
    parameter delta, P(T <= t:nu;delta), is defined by

        P(T <= t:nu;delta) = C_nuint_0^infinity (1/(sqrt(2 pi))int_-
        infinity^alpha u-delta e^(-x^2/2)dx)u^(nu-1)e^(-u^2/2)du, nu >
        0.0

    with

        C_nu = 1/(Gamma(1/2 nu)2^((nu-2)/2)), alpha = t/(sqrt(nu)).

    The probability is computed in one of two ways.

    (i) When t = 0.0, the relationship to the normal is used:

            P(T <= t:nu;delta) = 1/(sqrt(2 pi))int_delta^infinity
            e^(-u^2/2)du.

    (#) Otherwise the series expansion described in Equation 9 of Amos
        (1964) is used. This involves the sums of confluent
        hypergeometric functions, the terms of which are computed using
        recurrence relationships.

    References
    ----------
    Amos, D E, 1964, `Representations of the central and non-central
    t-distributions`, Biometrika (51), 451--458
    """

    try:
        _return_scalar = _b_stat.prob_students_t_noncentral(
            t,
            df,
            delta,
            tol,
            maxit,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = ProbStudentsTNoncentralReturnData(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def prob_chisq_noncentral(x, df, rlamda, tol=0.0, maxit=100):
    u"""
    Computes probabilities for the non-central chi^2 distribution.

    ``prob_chisq_noncentral`` returns the probability associated with
    the lower tail of the noncentral chi^2-distribution.

    For full information please refer to the NAG Library document for
    g01gc

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01gcf.html

    Parameters
    ----------
    x : float
        The deviate from the noncentral chi^2-distribution with nu
        degrees of freedom and noncentrality parameter lambda.

    df : float
        nu, the degrees of freedom of the noncentral chi^2-distribution.

    rlamda : float
        lambda, the noncentrality parameter of the noncentral
        chi^2-distribution.

    tol : float, optional
        The required accuracy of the solution. If
        ``prob_chisq_noncentral`` is entered with `tol` greater than or
        equal to 1.0 or less than 10*machine precision (see
        ``machine.precision``), the value of 10*machine precision is
        used instead.

    maxit : int, optional
        The maximum number of iterations to be performed.

    Returns
    -------
    p : float
        The probability associated with the lower tail of the noncentral
        chi^2-distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `df` = 0.0 and `rlamda` = 0.0.

            Constraint: `rlamda` > 0.0 if `df` = 0.0.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 1.

        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: `x` >= 0.0.

        (`errno` 1)
            On entry, `rlamda` = *<value>*.

            Constraint: `rlamda` >= 0.0.

        (`errno` 1)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 0.0.

        (`errno` 2)
            The initial value of the Poisson weight used in the
            summation of Equation (1) (see [Notes]) was too small to be
            calculated. The computed probability is likely to be zero.

        (`errno` 3)
            The solution has failed to converge in *<value>* iterations.
            Consider increasing `maxit` or `tol`.

        (`errno` 4)
            The value of a term required in Equation (2) (see [Notes])
            is too large to be evaluated accurately. The most likely
            cause of this error is both `x` and `rlamda` are too large.

        (`errno` 5)
            The calculations for the central chi-square probability has
            failed to converge. A larger value of `tol` should be used.

    Notes
    -----
    The lower tail probability of the noncentral chi^2-distribution with
    nu degrees of freedom and noncentrality parameter lambda, P(X <=
    x:nu;lambda), is defined by

        P(X <= x:nu;lambda) = sum_j = 0^infinity
        e^(-lambda/2)((lambda/2)^j)/j!P(X <= x:nu+2j;0),

    where P(X <= x:nu+2j;0) is a central chi^2-distribution with nu+2j
    degrees of freedom.

    The value of j at which the Poisson weight,
    e^(-lambda/2)((lambda/2)^j)/j!, is greatest is determined and the
    summation (1) is made forward and backward from that value of j.

    The recursive relationship:

        P(X <= x:a+2;0) = P(X <= x:a;0)-(x^a/2)e^(-x/2)/Gamma(a+1)

    is used during the summation in (1).

    References
    ----------
    NIST Digital Library of Mathematical Functions
    """

    _return_scalar = _b_stat.prob_chisq_noncentral(
        x,
        df,
        rlamda,
        tol,
        maxit,
    )
    return _return_scalar

ProbFNoncentralReturnData = _collections.namedtuple(
    'ProbFNoncentralReturnData',
    ['p'],
)

def prob_f_noncentral(f, df1, df2, rlamda, tol=0.0, maxit=500):
    u"""
    Computes probabilities for the non-central F-distribution.

    ``prob_f_noncentral`` returns the probability associated with the
    lower tail of the noncentral F or variance-ratio distribution.

    For full information please refer to the NAG Library document for
    g01gd

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01gdf.html

    Parameters
    ----------
    f : float
        f, the deviate from the noncentral F-distribution.

    df1 : float
        The degrees of freedom of the numerator variance, nu_1.

    df2 : float
        The degrees of freedom of the denominator variance, nu_2.

    rlamda : float
        lambda, the noncentrality parameter.

    tol : float, optional
        The relative accuracy required by you in the results. If
        ``prob_f_noncentral`` is entered with `tol` greater than or
        equal to 1.0 or less than 10*machine precision (see
        ``machine.precision``), the value of 10*machine precision is
        used instead.

    maxit : int, optional
        The maximum number of iterations to be used.

    Returns
    -------
    p : float
        The probability associated with the lower tail of the noncentral
        F or variance-ratio distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `df2` = *<value>*.

            Constraint: `df2` > 0.0.

        (`errno` 1)
            On entry, `rlamda` = *<value>*.

            Constraint: 0.0 <= `rlamda` <= -2.0*log(U), where U is the
            safe range parameter as defined by ``machine.real_safe``.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 1.

        (`errno` 1)
            On entry, `df1` = *<value>*.

            Constraint: 0.0 < `df1` <= 10^6.

        (`errno` 1)
            On entry, `f` = *<value>*.

            Constraint: `f` > 0.0.

        (`errno` 1)
            On entry, `df1` = *<value>*.

            Constraint: `df1` > 0.0.

        (`errno` 2)
            The solution has failed to converge in *<value>* iterations.
            Consider increasing `maxit` or `tol`.

        (`errno` 3)
            The required probability cannot be computed accurately. This
            may happen if the result would be very close to zero or one.
            Alternatively the values of `df1` and `f` may be too large.
            In the latter case you could try using a normal
            approximation, see Abramowitz and Stegun (1972).

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The required accuracy was not achieved when calculating the
            initial value of the central F or chi^2 probability. You
            should try a larger value of `tol`. If the chi^2
            approximation is being used then ``prob_f_noncentral``
            returns zero otherwise the value returned should be an
            approximation to the correct value.

    Notes
    -----
    The lower tail probability of the noncentral F-distribution with
    nu_1 and nu_2 degrees of freedom and noncentrality parameter lambda,
    P(F <= f:nu_1,nu_2;lambda), is defined by

        P(F <= f:nu_1,nu_2;lambda) = int_0^xp(F:nu_1,nu_2;lambda)dF,

    where

        P(F:nu_1,nu_2;lambda) = sum_j = 0^infinity
        e^(-lambda/2)((lambda/2)^j)/j!*(nu_1+2j)^((nu_1+2j)/2)nu_2^nu_2/2/B((nu_1+2j)/2,nu_2/2)

        *u^((nu_1+2j-2)/2)[nu_2+(nu_1+2j)u]^(-(nu_1+2j+nu_2)/2)

    and B(·,·) is the beta function.

    The probability is computed by means of a transformation to a
    noncentral beta distribution:

        P(F <= f:nu_1,nu_2;lambda) = P_beta(X <= x:a,b;lambda),

    where x = nu_1 f/(nu_1 f+nu_2) and P_beta(X <= x:a,b;lambda) is the
    lower tail probability integral of the noncentral beta distribution
    with parameters a, b, and lambda.

    If nu_2 is very large, greater than 10^6, then a chi^2 approximation
    is used.

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications
    """

    try:
        _return_scalar = _b_stat.prob_f_noncentral(
            f,
            df1,
            df2,
            rlamda,
            tol,
            maxit,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = ProbFNoncentralReturnData(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

ProbBetaNoncentralReturnData = _collections.namedtuple(
    'ProbBetaNoncentralReturnData',
    ['p'],
)

def prob_beta_noncentral(x, a, b, rlamda, tol=0.0, maxit=500):
    u"""
    Computes probabilities for the non-central beta distribution.

    ``prob_beta_noncentral`` returns the probability associated with the
    lower tail of the noncentral beta distribution.

    For full information please refer to the NAG Library document for
    g01ge

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01gef.html

    Parameters
    ----------
    x : float
        beta, the deviate from the beta distribution, for which the
        probability P(B <= beta:a,b;lambda) is to be found.

    a : float
        a, the first parameter of the required beta distribution.

    b : float
        b, the second parameter of the required beta distribution.

    rlamda : float
        lambda, the noncentrality parameter of the required beta
        distribution.

    tol : float, optional
        The relative accuracy required by you in the results. If
        ``prob_beta_noncentral`` is entered with `tol` greater than or
        equal to 1.0 or less than 10*machine precision (see
        ``machine.precision``), the value of 10*machine precision is
        used instead.

        See [Accuracy] for the relationship between `tol` and `maxit`.

    maxit : int, optional
        The maximum number of iterations that the algorithm should use.

        See [Accuracy] for suggestions as to suitable values for `maxit`
        for different values of the arguments.

    Returns
    -------
    p : float
        The probability associated with the lower tail of the noncentral
        beta distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `b` = *<value>*.

            Constraint: 0.0 < `b` <= 10^6.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 1.

        (`errno` 1)
            On entry, `a` = *<value>*.

            Constraint: 0.0 < `a` <= 10^6.

        (`errno` 1)
            On entry, `x` = *<value>*.

            Constraint: 0.0 <= `x` <= 1.0.

        (`errno` 1)
            On entry, `rlamda` = *<value>*.

            Constraint: 0.0 <= `rlamda` <= -2.0log(U), where U is the
            safe range parameter as defined by ``machine.real_safe``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The solution has failed to converge in *<value>* iterations.
            Consider increasing `maxit` or `tol`. The returned value
            will be an approximation to the correct value.

        (`errno` 3)
            The probability is too close to 0.0 or 1.0 for the algorithm
            to be able to calculate the required probability.
            ``prob_beta_noncentral`` will return 0.0 or 1.0 as
            appropriate. This should be a reasonable approximation.

        (`errno` 4)
            The required accuracy was not achieved when calculating the
            initial value of the beta distribution. You should try a
            larger value of `tol`. The returned value will be an
            approximation to the correct value.

    Notes
    -----
    The lower tail probability for the noncentral beta distribution with
    parameters a and b and noncentrality parameter lambda, P(B <=
    beta:a,b;lambda), is defined by

        P(B <= beta:a,b;lambda) = sum_j = 0^infinity
        e^(-lambda/2)((lambda/2))/j!P(B <= beta:a,b;0),

    where

        P(B <= beta:a,b;0) = Gamma(a+b)/(Gamma(a)Gamma(b))int_0^beta
        B^(a-1)(1-B)^(b-1)dB,

    which is the central beta probability function or incomplete beta
    function.

    Recurrence relationships given in Abramowitz and Stegun (1972) are
    used to compute the values of P(B <= beta:a,b;0) for each step of
    the summation (1).

    The algorithm is discussed in Lenth (1987).

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Lenth, R V, 1987, `Algorithm AS 226: Computing noncentral beta
    probabilities`, Appl. Statist. (36), 241--244
    """

    try:
        _return_scalar = _b_stat.prob_beta_noncentral(
            x,
            a,
            b,
            rlamda,
            tol,
            maxit,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = ProbBetaNoncentralReturnData(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def prob_bivariate_normal(x, y, rho):
    u"""
    Computes probability for the bivariate Normal distribution.

    ``prob_bivariate_normal`` returns the lower tail probability for the
    bivariate Normal distribution.

    For full information please refer to the NAG Library document for
    g01ha

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01haf.html

    Parameters
    ----------
    x : float
        x, the first argument for which the bivariate Normal
        distribution function is to be evaluated.

    y : float
        y, the second argument for which the bivariate Normal
        distribution function is to be evaluated.

    rho : float
        rho, the correlation coefficient.

    Returns
    -------
    p : float
        The lower tail probability for the bivariate Normal
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `rho` = *<value>*.

            Constraint: `rho` <= 1.0.

        (`errno` 1)
            On entry, `rho` = *<value>*.

            Constraint: `rho` >= -1.0.

    Notes
    -----
    For the two random variables (X,Y) following a bivariate Normal
    distribution with

        E[X] = 0, E[Y] = 0, E[X^2] = 1, E[Y^2] = 1 and E[XY] = rho,

    the lower tail probability is defined by:

        P(X <= x,Y <= y:rho) = 1/(2 pi*sqrt(1-rho^2))int_-
        infinity^yint_- infinity^xexp(-((X^2-2*rho
        XY+Y^2))/(2*(1-rho^2)))dXdY.

    For a more detailed description of the bivariate Normal distribution
    and its properties see Abramowitz and Stegun (1972) and Kendall and
    Stuart (1969).
    The method used is described by Genz (2004).

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Genz, A, 2004, `Numerical computation of rectangular bivariate and
    trivariate Normal and t probabilities`, Statistics and Computing
    (14), 151--160

    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin
    """

    _return_scalar = _b_stat.prob_bivariate_normal(
        x,
        y,
        rho,
    )
    return _return_scalar

ProbMultiNormalReturnData = _collections.namedtuple(
    'ProbMultiNormalReturnData',
    ['p'],
)

def prob_multi_normal(xmu, sig, a=None, b=None, tol=0.0001):
    u"""
    Computes probabilities for the multivariate Normal distribution.

    ``prob_multi_normal`` returns the upper tail, lower tail or central
    probability associated with a multivariate Normal distribution of up
    to ten dimensions.

    For full information please refer to the NAG Library document for
    g01hb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01hbf.html

    Parameters
    ----------
    xmu : float, array-like, shape (n)
        mu, the mean vector of the multivariate Normal distribution.

    sig : float, array-like, shape (n, n)
        Sigma, the variance-covariance matrix of the multivariate Normal
        distribution. Only the lower triangle is referenced.

    a : None or float, array-like, shape (n), optional
        If upper tail or central probablilities are to be returned, `a`
        should supply the lower bounds, a_i, for i = 1,2,...,n.

    b : None or float, array-like, shape (n), optional
        If lower tail or central probablilities are to be returned, `b`
        should supply the upper bounds, b_i, for i = 1,2,...,n.

    tol : float, optional
        If n > 2 the relative accuracy required for the probability, and
        if the upper or the lower tail probability is requested then
        `tol` is also used to determine the cut-off points, see
        [Accuracy].

        If n = 1, `tol` is not referenced.

    Returns
    -------
    p : float
        The upper tail, lower tail or central probability associated
        with then multivariate Normal distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: 1 <= n <= 10.

        (`errno` 1)
            On entry, tail = *<value>*.

            Constraint: tail = 'L', 'U' or 'C'.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 2)
            On entry, the *<value>**<value>* value in `b` is less than
            or equal to the corresponding value in `a`.

        (`errno` 3)
            On entry, `sig` is not positive definite.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            Full accuracy not achieved, relative accuracy = *<value>*.

        (`errno` 5)
            Accuracy requested by `tol` is too strict: `tol` =
            *<value>*.

    Notes
    -----
    Let the vector random variable X = (X_1,X_2,...,X_n)^T follow an
    n-dimensional multivariate Normal distribution with mean vector mu
    and n*n variance-covariance matrix Sigma, then the probability
    density function, f(X:mu,Sigma), is given by

        f(X:mu,Sigma) = (2 pi)^(-(1/2)n)|Sigma|^(-1/2)exp(-1/2(X-mu)^T
        Sigma^-1(X-mu)).

    The lower tail probability is defined by:

        P(X_1 <= b_1,...,X_n <= b_n:mu,Sigma) = int_-
        infinity^b_1...int_- infinity^b_nf(X:mu,Sigma)dX_n...dX_1.

    The upper tail probability is defined by:

        P(X_1 >= a_1,...,X_n >= a_n:mu,Sigma) =
        int_a_1^infinity...int_a_n^infinity f(X:mu,Sigma)dX_n...dX_1.

    The central probability is defined by:

        P(a_1 <= X_1 <= b_1,...,a_n <= X_n <= b_n:mu,Sigma) =
        int_a_1^b_1...int_a_n^b_nf(X:mu,Sigma)dX_n...dX_1.

    To evaluate the probability for n >= 3, the probability density
    function of X_1,X_2,...,X_n is considered as the product of the
    conditional probability of X_1,X_2,...,X_{n-2} given X_{n-1} and X_n
    and the marginal bivariate Normal distribution of X_{n-1} and X_n.
    The bivariate Normal probability can be evaluated as described in
    ``prob_bivariate_normal`` and numerical integration is then used
    over the remaining n-2 dimensions.
    In the case of n = 3, ``quad.dim1_fin_bad`` is used and for n > 3
    ``quad.md_adapt`` is used.

    To evaluate the probability for n = 1 a direct call to
    ``prob_normal`` is made and for n = 2 calls to
    ``prob_bivariate_normal`` are made.

    References
    ----------
    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin
    """

    fun_name = 'naginterfaces.library.stat.prob_multi_normal'

    _xmu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xmu'},
        xmu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_xmu, {'fun_name': fun_name, 'entity': 'xmu'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _sig = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sig'},
        sig,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalar = _b_stat.prob_multi_normal(
            _n,
            _a,
            _b,
            _xmu,
            _sig,
            tol,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = ProbMultiNormalReturnData(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def prob_bivariate_students_t(df, rho, a=None, b=None):
    u"""
    Computes probabilities for the bivariate Student's t-distribution.

    ``prob_bivariate_students_t`` returns probabilities for the
    bivariate Student's t-distribution.

    For full information please refer to the NAG Library document for
    g01hc

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01hcf.html

    Parameters
    ----------
    df : int
        nu, the degrees of freedom of the bivariate Student's
        t-distribution.

    rho : float
        rho, the correlation of the bivariate Student's t-distribution.

    a : None or float, array-like, shape (2), optional
        If upper tail or central probablilities are to be returned, `a`
        should supply the lower bounds, a_i, for i = 1,2,...,2.

    b : None or float, array-like, shape (2), optional
        If lower tail or central probablilities are to be returned, `b`
        should supply the upper bounds, b_i, for i = 1,2,...,2.

    Returns
    -------
    p : float
        The probabilities for the bivariate Student's t-distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, tail = *<value>*.

            Constraint: tail = 'L', 'U' or 'C'.

        (`errno` 3)
            On entry, `b`[i-1] <= `a`[i-1] for central probability, for
            some i = 1,2.

        (`errno` 4)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 1.

        (`errno` 5)
            On entry, `rho` = *<value>*.

            Constraint: -1.0 <= `rho` <= 1.0.

    Notes
    -----
    Let the vector random variable X = (X_1,X_2)^T follow a bivariate
    Student's t-distribution with degrees of freedom nu and correlation
    rho, then the probability density function is given by

        (X:nu,rho) = 1/(2 pi sqrt(1-rho^2))(1+(X_1^2+X_2^2-2 rho X_1
        X_2)/nu(1-rho^2))^(-nu/2-1).

    The lower tail probability is defined by:

        (X_1 <= b_1,X_2 <= b_2:nu,rho) = int_- infinity^b_1int_-
        infinity^b_2(X:nu,rho)dX_2 dX_1.

    The upper tail probability is defined by:

        (X_1 >= a_1,X_2 >= a_2:nu,rho) = int_a_1^infinity
        int_a_2^infinity (X:nu,rho)dX_2 dX_1.

    The central probability is defined by:

        (a_1 <= X_1 <= b_1,a_2 <= X_2 <= b_2:nu,rho) =
        int_a_1^b_1int_a_2^b_2(X:nu,rho)dX_2 dX_1.

    Calculations use the Dunnett and Sobel (1954) method, as described
    by Genz (2004).

    References
    ----------
    Dunnett, C W and Sobel, M, 1954, `A bivariate generalization of
    Student's t-distribution, with tables for certain special cases`,
    Biometrika (41), 153--169

    Genz, A, 2004, `Numerical computation of rectangular bivariate and
    trivariate Normal and t probabilities`, Statistics and Computing
    (14), 151--160
    """

    fun_name = 'naginterfaces.library.stat.prob_bivariate_students_t'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_stat.prob_bivariate_students_t(
        _a,
        _b,
        df,
        rho,
    )
    return _return_scalar

ProbMultiStudentsTReturnData = _collections.namedtuple(
    'ProbMultiStudentsTReturnData',
    ['p', 'rc', 'errest'],
)

def prob_multi_students_t(tail, a, b, nu, delta, iscov, rc, epsabs=0.0, epsrel=0.001, numsub=350, nsampl=8, fmax=None):
    u"""
    Computes the probability for the multivariate Student's
    t-distribution.

    ``prob_multi_students_t`` returns a probability associated with a
    multivariate Student's t-distribution.

    For full information please refer to the NAG Library document for
    g01hd

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01hdf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (n)
        Defines the calculated probability, set `tail`[i-1] to:

        `tail`[i-1] = 'L'

            If the ith lower limit a_i is negative infinity.

        `tail`[i-1] = 'U'

            If the ith upper limit b_i is infinity.

        `tail`[i-1] = 'C'

            If both a_i and b_i are finite.

    a : float, array-like, shape (n)
        a_i, for i = 1,2,...,n, the lower integral limits of the
        calculation.

        If `tail`[i-1] = 'L', `a`[i-1] is not referenced and the ith
        lower limit of integration is - infinity.

    b : float, array-like, shape (n)
        b_i, for i = 1,2,...,n, the upper integral limits of the
        calculation.

        If `tail`[i-1] = 'U', `b`[i-1] is not referenced and the ith
        upper limit of integration is infinity.

    nu : float
        nu, the degrees of freedom.

    delta : float, array-like, shape (n)
        `delta`[i-1] the noncentrality parameter for the ith dimension,
        for i = 1,2,...,n; set `delta`[i-1] = 0 for the central
        probability.

    iscov : int
        Set `iscov` = 1 if the covariance matrix is supplied and `iscov`
        = 2 if the correlation matrix is supplied.

    rc : float, array-like, shape (n, n)
        The lower triangle of either the covariance matrix (if `iscov` =
        1) or the correlation matrix (if `iscov` = 2). In either case
        the array elements corresponding to the upper triangle of the
        matrix need not be set.

    epsabs : float, optional
        epsilon_a, the absolute accuracy requested in the approximation.
        If `epsabs` is negative, the absolute value is used.

    epsrel : float, optional
        epsilon_r, the relative accuracy requested in the approximation.
        If `epsrel` is negative, the absolute value is used.

    numsub : int, optional
        If quadrature is used, the number of sub-intervals used by the
        quadrature algorithm; otherwise `numsub` is not referenced.

    nsampl : int, optional
        If quadrature is used, `nsampl` is not referenced; otherwise
        `nsampl` is the number of samples used to estimate the error in
        the approximation.

    fmax : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 1000*n.

        If a number theoretic approach is used, the maximum number of
        evaluations for each integrand function.

    Returns
    -------
    p : float
        The probability associated with the multivariate Student's
        t-distribution.

    rc : float, ndarray, shape (n, n)
        The strict upper triangle of `rc` contains the correlation
        matrix used in the calculations.

    errest : float
        An estimate of the error in the calculated probability.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: 1 < n <= 1000.

        (`errno` 2)
            On entry, `tail`[k-1] = *<value>*.

            Constraint: `tail`[k-1] = 'L', 'U' or 'C'.

        (`errno` 4)
            On entry, k = *<value>*.

            Constraint: `b`[k-1] > `a`[k-1] for a central probability.

        (`errno` 5)
            On entry, `nu` = *<value>*.

            Constraint: degrees of freedom `nu` > 0.0.

        (`errno` 8)
            On entry, `iscov` = *<value>*.

            Constraint: `iscov` = 1 or 2.

        (`errno` 9)
            On entry, the information supplied in `rc` is invalid.

        (`errno` 12)
            On entry, `numsub` = *<value>*.

            Constraint: `numsub` >= 1.

        (`errno` 13)
            On entry, `nsampl` = *<value>*.

            Constraint: `nsampl` >= 1.

        (`errno` 14)
            On entry, `fmax` = *<value>*.

            Constraint: `fmax` >= 1.

    Notes
    -----
    A random vector x in R^n that follows a Student's t-distribution
    with nu degrees of freedom and covariance matrix Sigma has density:

Gamma((nu+n)/2)/(Gamma(nu/2)nu^(n/2)pi^(n/2)|Sigma|^(1/2)[1+1/nux^T
        Sigma^-1x]^((nu+n)/2)),

    and probability p given by:

        p =
        Gamma((nu+n)/2)/(Gamma(nu/2)sqrt(|Sigma|(pinu)^n))int_a_1^b_1int_a_2^b_2...int_a_n^b_n(1+x^T
        Sigma^-1x/nu)^(-(nu+n)/2)dx.

    The method of calculation depends on the dimension n and degrees of
    freedom nu.
    The method of Dunnett and Sobel (1954) is used in the bivariate case
    if nu is a whole number.
    A Plackett transform followed by quadrature method is adopted in
    other bivariate cases and trivariate cases.
    In dimensions higher than three a number theoretic approach to
    evaluating multidimensional integrals is adopted.

    Error estimates are supplied as the published accuracy in the
    Dunnett and Sobel (1954) case, a Monte Carlo standard error for
    multidimensional integrals, and otherwise the quadrature error
    estimate.

    A parameter delta allows for non-central probabilities.
    The number theoretic method is used if any delta is nonzero.

    In cases other than the central bivariate with whole nu,
    ``prob_multi_students_t`` attempts to evaluate probabilities within
    a requested accuracy max(epsilon_a, epsilon_r*I), for an approximate
    integral value I, absolute accuracy epsilon_a and relative accuracy
    epsilon_r.

    References
    ----------
    Dunnett, C W and Sobel, M, 1954, `A bivariate generalization of
    Student's t-distribution, with tables for certain special cases`,
    Biometrika (41), 153--169

    Genz, A and Bretz, F, 2002, `Methods for the computation of
    multivariate t-probabilities`, Journal of Computational and
    Graphical Statistics ((11)), 950--971
    """

    fun_name = 'naginterfaces.library.stat.prob_multi_students_t'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _n_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _delta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'delta'},
        delta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _rc = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rc'},
        rc,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _fmax = 1000*_n_py if fmax is None else fmax
    _return_scalars = _b_stat.prob_multi_students_t(
        _n,
        _tail,
        _a,
        _b,
        nu,
        _delta,
        iscov,
        _rc,
        epsabs,
        epsrel,
        numsub,
        nsampl,
        _fmax,
    )
    _rc_py = _rc
    _p_py = _return_scalars.p
    _errest_py = _return_scalars.errest
    _return_data = ProbMultiStudentsTReturnData(
        _p_py,
        _rc_py,
        _errest_py,
    )
    return _return_data

ProbChisqNoncentralLincombReturnData = _collections.namedtuple(
    'ProbChisqNoncentralLincombReturnData',
    ['p', 'pdf'],
)

def prob_chisq_noncentral_lincomb(a, mult, rlamda, c, tol=0.0, maxit=500):
    u"""
    Computes probability for a positive linear combination of chi^2
    variables.

    ``prob_chisq_noncentral_lincomb`` returns the lower tail probability
    of a distribution of a positive linear combination of chi^2 random
    variables.

    For full information please refer to the NAG Library document for
    g01jc

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01jcf.html

    Parameters
    ----------
    a : float, array-like, shape (n)
        The weights, a_1,a_2,...,a_n.

    mult : int, array-like, shape (n)
        The degrees of freedom, m_1,m_2,...,m_n.

    rlamda : float, array-like, shape (n)
        The noncentrality parameters, lambda_1,lambda_2,...,lambda_n.

    c : float
        c, the point for which the lower tail probability is to be
        evaluated.

    tol : float, optional
        The relative accuracy required by you in the results. If
        ``prob_chisq_noncentral_lincomb`` is entered with `tol` greater
        than or equal to 1.0 or less than 10*machine precision (see
        ``machine.precision``), the value of 10*machine precision is
        used instead.

    maxit : int, optional
        The maximum number of terms that should be used during the
        summation.

    Returns
    -------
    p : float
        The lower tail probability associated with the linear
        combination of n chi^2 random variables with m_j degrees of
        freedom, and noncentrality parameters lambda_j, for j =
        1,2,...,n.

    pdf : float
        The value of the probability density function of the linear
        combination of chi^2 variables.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 1.

        (`errno` 1)
            On entry, `c` = *<value>*.

            Constraint: `c` >= 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            On entry, `rlamda`[*<value>*] = *<value>*.

            Constraint: `rlamda`[i] >= 0.0, for i = 0,...,n-1.

        (`errno` 2)
            On entry, `mult`[*<value>*] = *<value>*.

            Constraint: `mult`[i] >= 1, for i = 0,...,n-1.

        (`errno` 2)
            On entry, `a`[*<value>*] = *<value>*.

            Constraint: `a`[i] > 0.0, for i = 0,...,n-1.

        (`errno` 3)
            The central chi^2 calculation has failed to converge. This
            is an unlikely exit. A larger value of `tol` should be
            tried.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The solution has failed to converge within `maxit`
            iterations. A larger value of `maxit` or `tol` should be
            used. The returned value should be a reasonable
            approximation to the correct value.

        (`errno` 5)
            The solution appears to be too close to 0 or 1 for accurate
            calculation. The value returned is 0 or 1 as appropriate.

    Notes
    -----
    For a linear combination of noncentral chi^2 random variables with
    integer degrees of freedom the lower tail probability is

        P(sum_j = 1^na_jchi^2(m_j,lambda_j) <= c),

    where a_j and c are positive constants and where chi^2(m_j,lambda_j)
    represents an independent chi^2 random variable with m_j degrees of
    freedom and noncentrality parameter lambda_j.
    The linear combination may arise from considering a quadratic form
    in Normal variables.

    Ruben's method as described in Farebrother (1984) is used.
    Ruben has shown that (1) may be expanded as an infinite series of
    the form

        sum_k = 0^infinity d_k F(m+2k,c/beta),

    where F(m+2k,c/beta) = P(chi^2(m+2k) < c/beta), i.e., the
    probability that a central chi^2 is less than c/beta.

    The value of beta is set at

        beta = beta_B = 2/(1/a_min+1/a_max)

    unless beta_B > 1.8a_min, in which case

        beta = beta_A = a_min

    is used, where a_min = min{a_j} and a_max = setmax(a_j), for j =
    1,2,...,n.

    References
    ----------
    Farebrother, R W, 1984, `The distribution of a positive linear
    combination of chi^2 random variables`, Appl. Statist. (33(3))
    """

    fun_name = 'naginterfaces.library.stat.prob_chisq_noncentral_lincomb'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _mult = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mult'},
        mult,
        copy=False,
    )
    _rlamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rlamda'},
        rlamda,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalars = _b_stat.prob_chisq_noncentral_lincomb(
            _a,
            _mult,
            _rlamda,
            _n,
            c,
            tol,
            maxit,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _p_py = _return_scalars.p
    _pdf_py = _return_scalars.pdf
    _return_data = ProbChisqNoncentralLincombReturnData(
        _p_py,
        _pdf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def prob_chisq_lincomb(rlam, d, c, method='D'):
    u"""
    Computes lower tail probability for a linear combination of
    (central) chi^2 variables.

    ``prob_chisq_lincomb`` calculates the lower tail probability for a
    linear combination of (central) chi^2 variables.

    For full information please refer to the NAG Library document for
    g01jd

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01jdf.html

    Parameters
    ----------
    rlam : float, array-like, shape (n)
        The weights, lambda_i, for i = 1,2,...,n, of the central chi^2
        variables.

    d : float
        d, the multiplier of the central chi^2 variables.

    c : float
        c, the value of the constant.

    method : str, length 1, optional
        Indicates whether Pan's, Imhof's or an appropriately selected
        procedure is to be used.

        `method` = 'P'

            Pan's method is used.

        `method` = 'I'

            Imhof's method is used.

        `method` = 'D'

            Pan's method is used if lambda_i^*, for i = 1,2,...,n are at
            least 1% distinct and n <= 60; otherwise Imhof's method is
            used.

    Returns
    -------
    prob : float
        The lower tail probability for the linear combination of central
        chi^2 variables.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 'P', 'I' or 'D'.

        (`errno` 1)
            On entry, `d` = *<value>*.

            Constraint: `d` >= 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            On entry, `rlam`[i-1] = `d` for all values of i, for i =
            1,2,...,n.

    Notes
    -----
    Let u_1,u_2,...,u_n be independent Normal variables with mean zero
    and unit variance, so that u_1^2,u_2^2,...,u_n^2 have independent
    chi^2-distributions with unit degrees of freedom.
    ``prob_chisq_lincomb`` evaluates the probability that

        lambda_1 u_1^2+lambda_2 u_2^2+...+lambda_n u_n^2 <
        d(u_1^2+u_2^2+...+u_n^2)+c.

    If c = 0.0 this is equivalent to the probability that

        (lambda_1 u_1^2+lambda_2 u_2^2+...+lambda_n
        u_n^2)/(u_1^2+u_2^2+...+u_n^2) < d.

    Alternatively let

        lambda_i^* = lambda_i-d, i = 1,2,...,n,

    then ``prob_chisq_lincomb`` returns the probability that

        lambda_1^*u_1^2+lambda_2^*u_2^2+...+lambda_n^*u_n^2 < c.

    Two methods are available.
    One due to Pan (1964) (see Farebrother (1980)) makes use of series
    approximations.
    The other method due to Imhof (1961) reduces the problem to a
    one-dimensional integral.
    If n >= 6 then a non-adaptive method described in
    ``quad.dim1_fin_smooth`` is used to compute the value of the
    integral otherwise ``quad.dim1_fin_bad`` is used.

    Pan's procedure can only be used if the lambda_i^* are sufficiently
    distinct; ``prob_chisq_lincomb`` requires the lambda_i^* to be at
    least 1% distinct; see [Further Comments].
    If the lambda_i^* are at least 1% distinct and n <= 60, then Pan's
    procedure is recommended; otherwise Imhof's procedure is
    recommended.

    References
    ----------
    Farebrother, R W, 1980, `Algorithm AS 153. Pan's procedure for the
    tail probabilities of the Durbin--Watson statistic`, Appl. Statist.
    (29), 224--227

    Imhof, J P, 1961, `Computing the distribution of quadratic forms in
    Normal variables`, Biometrika (48), 419--426

    Pan, Jie--Jian, 1964, `Distributions of the noncircular serial
    correlation coefficients`, Shuxue Jinzhan (7), 328--337
    """

    fun_name = 'naginterfaces.library.stat.prob_chisq_lincomb'

    _rlam = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rlam'},
        rlam,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_rlam, {'fun_name': fun_name, 'entity': 'rlam'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = _b_stat.prob_chisq_lincomb(
        method,
        _n,
        _rlam,
        d,
        c,
    )
    _prob_py = _return_scalar
    return _prob_py

def pdf_normal(x, xmean, xstd):
    u"""
    Calculates the value for the probability density function of the
    Normal distribution at a chosen point.

    ``pdf_normal`` returns the value of the probability density function
    (PDF) for the Normal (Gaussian) distribution with mean mu and
    variance sigma^2 at a point x.

    For full information please refer to the NAG Library document for
    g01ka

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01kaf.html

    Parameters
    ----------
    x : float
        x, the value at which the PDF is to be evaluated.

    xmean : float
        mu, the mean of the Normal distribution.

    xstd : float
        sigma, the standard deviation of the Normal distribution.

    Returns
    -------
    pdf : float
        The value of the PDF.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `xstd` = *<value>*.

            Constraint: `xstd`*sqrt(2.0*pi) > U, where U is the safe
            range parameter as defined by ``machine.real_safe``.

        (`errno` 2)
            Computation abandoned owing to underflow of
            1/(sigma*sqrt(2*pi)).

        (`errno` 3)
            Computation abandoned owing to an internal calculation
            overflowing.

    Notes
    -----
    The Normal distribution has probability density function (PDF)

        f(x) = 1/(sigma*sqrt(2 pi))e^(-(x-mu)^2/2 sigma^2), sigma > 0.
    """

    _return_scalar = _b_stat.pdf_normal(
        x,
        xmean,
        xstd,
    )
    return _return_scalar

def pdf_gamma(x, a, b):
    u"""
    Calculates the value for the probability density function of the
    gamma distribution at a chosen point.

    ``pdf_gamma`` returns the value of the probability density function
    (PDF) for the gamma distribution with shape parameter alpha and
    scale parameter beta at a point x.

    For full information please refer to the NAG Library document for
    g01kf

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01kff.html

    Parameters
    ----------
    x : float
        x, the value at which the PDF is to be evaluated.

    a : float
        alpha, the shape parameter of the gamma distribution.

    b : float
        beta, the scale parameter of the gamma distribution.

    Returns
    -------
    pdf : float
        The value of the PDF.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `a` = *<value>*.

            Constraint: `a` > 0.0.

        (`errno` 2)
            On entry, `b` = *<value>*.

            Constraint: `b` > 0.0.

        (`errno` 3)
            Computation abandoned owing to overflow due to extreme
            parameter values.

    Notes
    -----
    The gamma distribution has PDF

        [table omitted]

    If 0.01 <= x,alpha,beta <= 100 then an algorithm based directly on
    the gamma distribution's PDF is used.
    For values outside this range, the function is calculated via the
    Poisson distribution's PDF as described in Loader (2000) (see
    [Further Comments]).

    References
    ----------
    Loader, C, 2000, `Fast and accurate computation of binomial
    probabilities` ((not yet published))
    """

    _return_scalar = _b_stat.pdf_gamma(
        x,
        a,
        b,
    )
    return _return_scalar

PdfGammaVectorReturnData = _collections.namedtuple(
    'PdfGammaVectorReturnData',
    ['pdf', 'ivalid'],
)

def pdf_gamma_vector(ilog, x, a, b):
    u"""
    Computes a vector of values for the probability density function of
    the gamma distribution.

    ``pdf_gamma_vector`` returns a number of values of the probability
    density function (PDF), or its logarithm, for the gamma
    distribution.

    For full information please refer to the NAG Library document for
    g01kk

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01kkf.html

    Parameters
    ----------
    ilog : int
        The value of `ilog` determines whether the logarithmic value is
        returned in `pdf`.

        `ilog` = 0

            f(x_i, alpha_i, beta_i), the probability density function is
            returned.

        `ilog` = 1

            log(f(x_i, alpha_i, beta_i)), the logarithm of the
            probability density function is returned.

    x : float, array-like, shape (lx)
        x_i, the values at which the PDF is to be evaluated.

    a : float, array-like, shape (la)
        alpha_i, the shape parameter.

    b : float, array-like, shape (lb)
        beta_i, the scale parameter.

    Returns
    -------
    pdf : float, ndarray, shape (max(lx, la, lb))
        f(x_i, alpha_i, beta_i) or log(f(x_i, alpha_i, beta_i)).

    ivalid : int, ndarray, shape (max(lx, la, lb))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            alpha_i <= 0.0.

        `ivalid`[i-1] = 2

            beta_i <= 0.0.

        `ivalid`[i-1] = 3

            (x_i)/(beta_i) overflows, the value returned should be a
            reasonable approximation.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `ilog` = *<value>*.

            Constraint: `ilog` = 0 or 1.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lx > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: la > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: lb > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x`, `a` or `b` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    The gamma distribution with shape parameter alpha_i and scale
    parameter beta_i has PDF

        [table omitted]

    If 0.01 <= x_i,alpha_i,beta_i <= 100 then an algorithm based
    directly on the gamma distribution's PDF is used.
    For values outside this range, the function is calculated via the
    Poisson distribution's PDF as described in Loader (2000) (see
    [Further Comments]).

    The input arrays to this function are designed to allow maximum
    flexibility in the supply of vector arguments by re-using elements
    of any arrays that are shorter than the total number of evaluations
    required.
    See [the G01 Introduction] for further information.

    References
    ----------
    Loader, C, 2000, `Fast and accurate computation of binomial
    probabilities` ((not yet published))
    """

    fun_name = 'naginterfaces.library.stat.pdf_gamma_vector'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _lx = _b_utils._EngineIntScalarType(
        _lx_py, {'fun_name': fun_name, 'entity': '_lx'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _la_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _la = _b_utils._EngineIntScalarType(
        _la_py, {'fun_name': fun_name, 'entity': '_la'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lb_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=1)[0]
    _lb = _b_utils._EngineIntScalarType(
        _lb_py, {'fun_name': fun_name, 'entity': '_lb'},
    ).to_ctypes()
    _pdf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_lx_py, _la_py, _lb_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_lx_py, _la_py, _lb_py),),
    )
    try:
        _b_stat.pdf_gamma_vector(
            ilog,
            _lx,
            _x,
            _la,
            _a,
            _lb,
            _b,
            _pdf,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _pdf_py = _pdf
    _ivalid_py = _ivalid
    _return_data = PdfGammaVectorReturnData(
        _pdf_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

PdfNormalVectorReturnData = _collections.namedtuple(
    'PdfNormalVectorReturnData',
    ['pdf', 'ivalid'],
)

def pdf_normal_vector(ilog, x, xmu, xstd):
    u"""
    Computes a vector of values for the probability density function of
    the Normal distribution.

    ``pdf_normal_vector`` returns a number of values of the probability
    density function (PDF), or its logarithm, for the Normal (Gaussian)
    distributions.

    For full information please refer to the NAG Library document for
    g01kq

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01kqf.html

    Parameters
    ----------
    ilog : int
        The value of `ilog` determines whether the logarithmic value is
        returned in PDF.

        `ilog` = 0

            f(x_i, mu_i, sigma_i), the probability density function is
            returned.

        `ilog` = 1

            log(f(x_i, mu_i, sigma_i)), the logarithm of the probability
            density function is returned.

    x : float, array-like, shape (lx)
        x_i, the values at which the PDF is to be evaluated.

    xmu : float, array-like, shape (lxmu)
        mu_i, the means.

    xstd : float, array-like, shape (lxstd)
        sigma_i, the standard deviations.

    Returns
    -------
    pdf : float, ndarray, shape (max(lx, lxstd, lxmu))
        f(x_i, mu_i, sigma_i) or log(f(x_i, mu_i, sigma_i)).

    ivalid : int, ndarray, shape (max(lx, lxstd, lxmu))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            sigma_i < 0.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `ilog` = *<value>*.

            Constraint: `ilog` = 0 or 1.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lx > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: lxmu > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: lxstd > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `xstd` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    The Normal distribution with mean mu_i, variance sigma_i^2; has
    probability density function (PDF)

        f(x_i,mu_i,sigma_i) = 1/(sigma_i*sqrt(2 pi))e^(-(x_i-mu_i)^2/2
        sigma_i^2), sigma_i > 0.

    The input arrays to this function are designed to allow maximum
    flexibility in the supply of vector arguments by re-using elements
    of any arrays that are shorter than the total number of evaluations
    required.
    See [the G01 Introduction] for further information.
    """

    fun_name = 'naginterfaces.library.stat.pdf_normal_vector'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _lx = _b_utils._EngineIntScalarType(
        _lx_py, {'fun_name': fun_name, 'entity': '_lx'},
    ).to_ctypes()
    _xmu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xmu'},
        xmu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lxmu_py = _b_utils._EngineFloat64ArrayType.get_shape(_xmu, {'fun_name': fun_name, 'entity': 'xmu'}, exp_rank=1)[0]
    _lxmu = _b_utils._EngineIntScalarType(
        _lxmu_py, {'fun_name': fun_name, 'entity': '_lxmu'},
    ).to_ctypes()
    _xstd = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xstd'},
        xstd,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lxstd_py = _b_utils._EngineFloat64ArrayType.get_shape(_xstd, {'fun_name': fun_name, 'entity': 'xstd'}, exp_rank=1)[0]
    _lxstd = _b_utils._EngineIntScalarType(
        _lxstd_py, {'fun_name': fun_name, 'entity': '_lxstd'},
    ).to_ctypes()
    _pdf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_lx_py, _lxstd_py, _lxmu_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_lx_py, _lxstd_py, _lxmu_py),),
    )
    try:
        _b_stat.pdf_normal_vector(
            ilog,
            _lx,
            _x,
            _lxmu,
            _xmu,
            _lxstd,
            _xstd,
            _pdf,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _pdf_py = _pdf
    _ivalid_py = _ivalid
    _return_data = PdfNormalVectorReturnData(
        _pdf_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

PdfMultiNormalVectorReturnData = _collections.namedtuple(
    'PdfMultiNormalVectorReturnData',
    ['pdf', 'rank'],
)

def pdf_multi_normal_vector(ilog, x, xmu, iuld, sig):
    u"""
    Computes a vector of values for the probability density function of
    the multivariate Normal distribution.

    ``pdf_multi_normal_vector`` returns a number of values of the
    probability density function (PDF), or its logarithm, for the
    multivariate Normal (Gaussian) distribution.

    For full information please refer to the NAG Library document for
    g01lb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01lbf.html

    Parameters
    ----------
    ilog : int
        The value of `ilog` determines whether the logarithmic value is
        returned in PDF.

        `ilog` = 0

            f(X:mu, Sigma), the probability density function is
            returned.

        `ilog` = 1

            log(f(X:mu, Sigma)), the logarithm of the probability
            density function is returned.

    x : float, array-like, shape (n, k)
        X, the matrix of k points at which to evaluate the probability
        density function, with the ith dimension for the jth point held
        in `x`[i-1,j-1].

    xmu : float, array-like, shape (n)
        mu, the mean vector of the multivariate Normal distribution.

    iuld : int
        Indicates the form of Sigma and how it is stored in `sig`.

        `iuld` = 1

            `sig` holds the lower triangular portion of Sigma.

        `iuld` = 2

            `sig` holds the upper triangular portion of Sigma.

        `iuld` = 3

            Sigma is a diagonal matrix and `sig` only holds the diagonal
            elements.

        `iuld` = 4

            `sig` holds the lower Cholesky decomposition, L such that
            L*L^T = Sigma.

        `iuld` = 5

            `sig` holds the upper Cholesky decomposition, U such that
            U^T*U = Sigma.

    sig : float, array-like, shape (:, n)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `iuld`=3: 1; otherwise: n.

        Information defining the variance-covariance matrix, Sigma.

        `iuld` = 1 or 2

            `sig` must hold the lower or upper portion of Sigma, with
            Sigma_{ij} held in `sig`[i-1,j-1]. The supplied
            variance-covariance matrix must be positive semidefinite.

        `iuld` = 3

            Sigma is a diagonal matrix and the ith diagonal element,
            Sigma_{ii}, must be held in `sig`[0,i-1]

        `iuld` = 4 or 5

            `sig` must hold L or U, the lower or upper Cholesky
            decomposition of Sigma, with L_{ij} or U_{ij} held in
            `sig`[i-1,j-1], depending on the value of `iuld`. No check
            is made that L*L^T or U^T*U is a valid variance-covariance
            matrix. The diagonal elements of the supplied L or U must be
            greater than zero

    Returns
    -------
    pdf : float, ndarray, shape (k)
        f(X:mu, Sigma) or log(f(X:mu, Sigma)) depending on the value of
        `ilog`.

    rank : int
        r, rank of Sigma.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `ilog` = *<value>*.

            Constraint: `ilog` = 0 or 1.

        (`errno` 21)
            On entry, k = *<value>*.

            Constraint: k >= 0.

        (`errno` 31)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 71)
            On entry, `iuld` = *<value>*.

            Constraint: `iuld` = 1, 2, 3, 4 or 5.

        (`errno` 81)
            On entry, Sigma is not positive semidefinite.

        (`errno` 82)
            On entry, at least one diagonal element of Sigma is less
            than or equal to 0.

        (`errno` 83)
            On entry, Sigma is not positive definite and eigenvalue
            decomposition failed.

        (`errno` 92)
            On entry, ldsig = *<value>*.

            Constraint: if `iuld` != 3, ldsig >= n.

    Notes
    -----
    The probability density function, f(X:mu, Sigma) of an n-dimensional
    multivariate Normal distribution with mean vector mu and n*n
    variance-covariance matrix Sigma, is given by

        f(X:mu, Sigma) =
        ((2*pi)^n*|Sigma|)^(-1/2)*exp(-1/2*(X-mu)^T*Sigma^-1*(X-mu)).

    If the variance-covariance matrix, Sigma, is not of full rank then
    the probability density function, is calculated as

        f(X:mu, Sigma) =
        ((2*pi)^r*pdet(Sigma))^(-1/2)*exp(-1/2*(X-mu)^T*Sigma^-*(X-mu))

    where pdet(Sigma) is the pseudo-determinant, Sigma^- a generalized
    inverse of Sigma and r its rank.

    ``pdf_multi_normal_vector`` evaluates the PDF at k points with a
    single call.
    """

    fun_name = 'naginterfaces.library.stat.pdf_multi_normal_vector'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _xmu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xmu'},
        xmu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _sig = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sig'},
        sig,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _pdf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py,),
    )
    _return_scalar = _b_stat.pdf_multi_normal_vector(
        ilog,
        _k,
        _n,
        _x,
        _xmu,
        iuld,
        _sig,
        _pdf,
    )
    _pdf_py = _pdf
    _rank_py = _return_scalar
    _return_data = PdfMultiNormalVectorReturnData(
        _pdf_py,
        _rank_py,
    )
    return _return_data

def mills_ratio(x):
    u"""
    Computes reciprocal of Mills' Ratio.

    ``mills_ratio`` returns the reciprocal of Mills' Ratio.

    For full information please refer to the NAG Library document for
    g01mb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01mbf.html

    Parameters
    ----------
    x : float
        x, the argument of the reciprocal of Mills' Ratio.

    Returns
    -------
    rmr : float
        The reciprocal of Mills' Ratio.

    Notes
    -----
    ``mills_ratio`` calculates the reciprocal of Mills' Ratio, the
    hazard rate, lambda(x), for the standard Normal distribution.
    It is defined as the ratio of the ordinate to the upper tail area of
    the standard Normal distribution, that is,

        lambda(x) = Z(x)/Q(x) = 1/(sqrt(2 pi))e^(-(x^2/2))/(1/(sqrt(2
        pi))int_x^infinity e^(-(t^2/2))dt).

    The calculation is based on a Chebyshev expansion as described in
    ``specfun.erfcx_real``.

    References
    ----------
    Gross, A J and Clark, V A, 1975, `Survival Distributions:
    Reliability Applications in the Biomedical Sciences`, Wiley
    """

    _return_scalar = _b_stat.mills_ratio(
        x,
    )
    return _return_scalar

def pdf_landau(x):
    u"""
    Landau density function phi(lambda).

    ``pdf_landau`` returns the value of the Landau density function
    phi(lambda).

    For full information please refer to the NAG Library document for
    g01mt

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01mtf.html

    Parameters
    ----------
    x : float
        The argument lambda of the function.

    Returns
    -------
    pdf : float
        The value of the Landau density function phi(lambda).

    Notes
    -----
    ``pdf_landau`` evaluates an approximation to the Landau density
    function phi(lambda) given by

        phi(lambda) = 1/(2 pi i)int_c-i infinity^c+i infinity exp(lambda
        s+sln(s))ds,

    where c is an arbitrary real constant, using piecewise approximation
    by rational functions.
    Further details can be found in Kölbig and Schorr (1984).

    To obtain the value of phi'(lambda), ``pdf_landau_deriv`` can be
    used.

    References
    ----------
    Kölbig, K S and Schorr, B, 1984, `A program package for the Landau
    distribution`, Comp. Phys. Comm. (31), 97--111
    """

    _return_scalar = _b_stat.pdf_landau(
        x,
    )
    return _return_scalar

def pdf_vavilov(x, comm):
    u"""
    Vavilov density function phi_V(lambda;kappa,beta^2).

    ``pdf_vavilov`` returns the value of the Vavilov density function
    phi_V(lambda;kappa,beta^2).

    It is intended to be used after a call to ``init_vavilov``.

    For full information please refer to the NAG Library document for
    g01mu

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01muf.html

    Parameters
    ----------
    x : float
        The argument lambda of the function.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``init_vavilov``.

    Returns
    -------
    pdf : float
        The value of the Vavilov density function
        phi_V(lambda;kappa,beta^2).

    Raises
    ------
    NagValueError
        (`errno` 1)
            Either the initialization function has not been called prior
            to the first call of this function or a communication array
            has become corrupted.

    Notes
    -----
    ``pdf_vavilov`` evaluates an approximation to the Vavilov density
    function phi_V(lambda;kappa,beta^2) given by

        phi_V(lambda;kappa,beta^2) = 1/(2 pi i)int_c-i infinity^c+i
        infinity e^(lambda s)f(s;kappa,beta^2)ds,

    where kappa > 0 and 0 <= beta^2 <= 1, c is an arbitrary real
    constant and

        f(s;kappa,beta^2) =
        C(kappa,beta^2)exp{sln(kappa)+(s+kappabeta^2)[ln(s/kappa)+E_1(s/kappa)]-kappa
        exp(-s/kappa)}.

    E_1(x) = int_0^xt^-1(1-e^(-t))dt is the exponential integral,
    C(kappa,beta^2) = exp{kappa(1+gammabeta^2)} and gamma is Euler's
    constant.

    The method used is based on Fourier expansions.
    Further details can be found in Schorr (1974).

    For values of kappa <= 0.01, the Vavilov distribution can be
    replaced by the Landau distribution since lambda_V =
    (lambda_L-ln(kappa))/kappa.
    For values of kappa >= 10, the Vavilov distribution can be replaced
    by a Gaussian distribution with mean mu = gamma-1-beta^2-ln(kappa)
    and variance sigma^2 = (2-beta^2)/2 kappa.

    References
    ----------
    Schorr, B, 1974, `Programs for the Landau and the Vavilov
    distributions and the corresponding random numbers`, Comp. Phys.
    Comm. (7), 215--224
    """

    _return_scalar = _b_stat.pdf_vavilov(
        x,
        comm,
    )
    return _return_scalar

MomentsQuadFormReturnData = _collections.namedtuple(
    'MomentsQuadFormReturnData',
    ['rkum', 'rmom'],
)

def moments_quad_form(a, sigma, l, emu=None):
    u"""
    Cumulants and moments of quadratic forms in Normal variables.

    ``moments_quad_form`` computes the cumulants and moments of
    quadratic forms in Normal variates.

    For full information please refer to the NAG Library document for
    g01na

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01naf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n*n symmetric matrix A. Only the lower triangle is
        referenced.

    sigma : float, array-like, shape (n, n)
        The n*n variance-covariance matrix Sigma. Only the lower
        triangle is referenced.

    l : int
        The required number of cumulants, and moments if specified.

    emu : None or float, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if mean='M': n; otherwise: 1.

        If mean = 'M', `emu` must contain the n elements of the vector
        mu.

        If mean = 'Z', `emu` is not referenced.

    Returns
    -------
    rkum : float, ndarray, shape (`l`)
        The `l` cumulants of the quadratic form.

    rmom : float, ndarray, shape (:)
        If mom = 'M', the `l` moments of the quadratic form.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` <= 12.

        (`errno` 1)
            On entry, mean = *<value>*.

            Constraint: mean = 'Z' or 'M'.

        (`errno` 1)
            On entry, mom = *<value>*.

            Constraint: mom = 'C' or 'M'.

        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 1.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 1.

        (`errno` 2)
            On entry, `sigma` is not positive definite.

    Notes
    -----
    Let x have an n-dimensional multivariate Normal distribution with
    mean mu and variance-covariance matrix Sigma.
    Then for a symmetric matrix A, ``moments_quad_form`` computes up to
    the first 12 moments and cumulants of the quadratic form Q = x^TAx.
    The sth moment (about the origin) is defined as

        E(Q^s),

    where E denotes expectation.
    The sth moment of Q can also be found as the coefficient of t^s/s!
    in the expansion of E(e^(Qt)).
    The sth cumulant is defined as the coefficient of t^s/s! in the
    expansion of log(E(e^(Qt))).

    The function is based on the function CUM written by Magnus and
    Pesaran (1993a) and based on the theory given by Magnus (1978),
    Magnus (1979) and Magnus (1986).

    References
    ----------
    Magnus, J R, 1978, `The moments of products of quadratic forms in
    Normal variables`, Statist. Neerlandica (32), 201--210

    Magnus, J R, 1979, `The expectation of products of quadratic forms
    in Normal variables: the practice`, Statist. Neerlandica (33),
    131--136

    Magnus, J R, 1986, `The exact moments of a ratio of quadratic forms
    in Normal variables`, Ann. Économ. Statist. (4), 95--109

    Magnus, J R and Pesaran, B, 1993, `The evaluation of cumulants and
    moments of quadratic forms in Normal variables (CUM): Technical
    description`, Comput. Statist. (8), 39--45

    Magnus, J R and Pesaran, B, 1993, `The evaluation of moments of
    quadratic forms and ratios of quadratic forms in Normal variables:
    Background, motivation and examples`, Comput. Statist. (8), 47--55
    """

    fun_name = 'naginterfaces.library.stat.moments_quad_form'

    _mom = 'M'
    _mom_py = _mom
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _emu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'emu'},
        emu,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _sigma = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sigma'},
        sigma,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _l = _b_utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _l_py = _b_utils._EngineIntScalarType.to_py(_l)
    _rkum = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_l_py,),
    )
    if _mom_py in ('M', 'm'):
        _rmom_dim1 = _l_py
    else:
        _rmom_dim1 = 1
    exp_shape_rmom = (_rmom_dim1,)
    _rmom = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_rmom,
    )
    _b_stat.moments_quad_form(
        _mom,
        _n,
        _a,
        _emu,
        _sigma,
        _l,
        _rkum,
        _rmom,
    )
    _rkum_py = _rkum
    _rmom_py = _rmom
    _return_data = MomentsQuadFormReturnData(
        _rkum_py,
        _rmom_py,
    )
    return _return_data

MomentsRatioQuadFormsReturnData = _collections.namedtuple(
    'MomentsRatioQuadFormsReturnData',
    ['lmax', 'rmom', 'abserr'],
)

def moments_ratio_quad_forms(a, b, sigma, l1, l2, eps, c=None, ela=None, emu=None):
    u"""
    Moments of ratios of quadratic forms in Normal variables, and
    related statistics.

    ``moments_ratio_quad_forms`` computes the moments of ratios of
    quadratic forms in Normal variables and related statistics.

    For full information please refer to the NAG Library document for
    g01nb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01nbf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n*n symmetric matrix A. Only the lower triangle is
        referenced.

    b : float, array-like, shape (n, n)
        The n*n positive semidefinite symmetric matrix B. Only the lower
        triangle is referenced.

    sigma : float, array-like, shape (n, n)
        The n*n variance-covariance matrix Sigma. Only the lower
        triangle is referenced.

    l1 : int
        The first moment to be computed, l_1.

    l2 : int
        The last moment to be computed, l_2.

    eps : float
        The relative accuracy required for the moments, this value is
        also used in the checks for the existence of the moments.

        If `eps` = 0.0, a value of sqrt(epsilon) where epsilon is the
        machine precision used.

    c : None or float, array-like, shape (:, n), optional
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `c` is not None: n; otherwise: 0.

        If `c` is not None, `c` must contain the n*n symmetric matrix C;
        only the lower triangle is referenced.

    ela : None or float, array-like, shape (n), optional
        If `ela` is not None, `ela` must contain the vector a of length
        n, otherwise `ela` is not referenced.

    emu : None or float, array-like, shape (n), optional
        If `emu` is not None, `emu` must contain the n elements of the
        vector mu.

    Returns
    -------
    lmax : int
        The highest moment computed, l_MAX. This will be l_2 if no
        exception or warning is raised on exit.

    rmom : float, ndarray, shape (`l2`-`l1`+1)
        The l_1 to l_MAX moments.

    abserr : float
        The estimated maximum absolute error in any computed moment.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `l2` = *<value>*.

            Constraint: `l2` <= 12.

        (`errno` 1)
            On entry, `l1` = *<value>* and `l2` = *<value>*.

            Constraint: `l2` >= `l1`.

        (`errno` 1)
            On entry, `l1` = *<value>*.

            Constraint: `l1` >= 1.

        (`errno` 1)
            On entry, `eps` = *<value>*.

            Constraint: if `eps` != 0.0, `eps` >= machine precision.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 1.

        (`errno` 2)
            On entry, `sigma` is not positive definite.

        (`errno` 2)
            On entry, `b` is not positive semidefinite or is null.

        (`errno` 3)
            Only *<value>* moments exist, less than `l1` = *<value>*,
            therefore, none of the required moments can be computed.

        (`errno` 4)
            The matrix L^TBL is not positive semidefinite or is null.

        (`errno` 5)
            The computation to compute the eigenvalues required in the
            calculation of moments has failed to converge: this is an
            unlikely error exit.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            Only some of the required moments have been computed, the
            highest is given by `lmax`.

        (`errno` 7)
            The required accuracy has not been achieved in the
            integration. An estimate of the accuracy is returned in
            `abserr`.

    Notes
    -----
    Let x have an n-dimensional multivariate Normal distribution with
    mean mu and variance-covariance matrix Sigma.
    Then for a symmetric matrix A and symmetric positive semidefinite
    matrix B, ``moments_ratio_quad_forms`` computes a subset, l_1 to
    l_2, of the first 12 moments of the ratio of quadratic forms

        R = x^TAx/x^TBx.

    The sth moment (about the origin) is defined as

        E(R^s),

    where E denotes the expectation.
    Alternatively, this function will compute the following
    expectations:

        E(R^s(a^Tx))

    and

        E(R^s(x^TCx)),

    where a is a vector of length n and C is an n*n symmetric matrix, if
    they exist.
    In the case of (2) the moments are zero if mu = 0.

    The conditions of theorems 1, 2 and 3 of Magnus (1986) and Magnus
    (1990) are used to check for the existence of the moments.
    If all the requested moments do not exist, the computations are
    carried out for those moments that are requested up to the maximum
    that exist, l_MAX.

    This function is based on the function QRMOM written by Magnus and
    Pesaran (1993a) and based on the theory given by Magnus (1986) and
    Magnus (1990).
    The computation of the moments requires first the computation of the
    eigenvectors of the matrix L^TBL, where LL^T = Sigma.
    The matrix L^TBL must be positive semidefinite and not null.
    Given the eigenvectors of this matrix, a function which has to be
    integrated over the range zero to infinity can be computed.
    This integration is performed using ``quad.dim1_inf``.

    References
    ----------
    Magnus, J R, 1986, `The exact moments of a ratio of quadratic forms
    in Normal variables`, Ann. Économ. Statist. (4), 95--109

    Magnus, J R, 1990, `On certain moments relating to quadratic forms
    in Normal variables: Further results`, Sankhyā, Ser. B (52), 1--13

    Magnus, J R and Pesaran, B, 1993, `The evaluation of cumulants and
    moments of quadratic forms in Normal variables (CUM): Technical
    description`, Comput. Statist. (8), 39--45

    Magnus, J R and Pesaran, B, 1993, `The evaluation of moments of
    quadratic forms and ratios of quadratic forms in Normal variables:
    Background, motivation and examples`, Comput. Statist. (8), 47--55
    """

    fun_name = 'naginterfaces.library.stat.moments_ratio_quad_forms'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ela = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ela'},
        ela,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _emu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'emu'},
        emu,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _sigma = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'sigma'},
        sigma,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _l2 = _b_utils._EngineIntScalarType(
        l2, {'fun_name': fun_name, 'entity': 'l2'},
    ).to_ctypes()
    _l2_py = _b_utils._EngineIntScalarType.to_py(_l2)
    _l1 = _b_utils._EngineIntScalarType(
        l1, {'fun_name': fun_name, 'entity': 'l1'},
    ).to_ctypes()
    _l1_py = _b_utils._EngineIntScalarType.to_py(_l1)
    _rmom = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_l2_py-_l1_py+1,),
    )
    try:
        _return_scalars = _b_stat.moments_ratio_quad_forms(
            _n,
            _a,
            _b,
            _c,
            _ela,
            _emu,
            _sigma,
            _l1,
            _l2,
            _rmom,
            eps,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _lmax_py = _return_scalars.lmax
    _rmom_py = _rmom
    _abserr_py = _return_scalars.abserr
    _return_data = MomentsRatioQuadFormsReturnData(
        _lmax_py,
        _rmom_py,
        _abserr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def pdf_landau_moment1(x):
    u"""
    Landau first moment function Phi_1(x).

    ``pdf_landau_moment1`` returns the value of the first moment
    Phi_1(x) of the Landau density function.

    For full information please refer to the NAG Library document for
    g01pt

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01ptf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    mom1 : float
        The value of the first moment Phi_1(x) of the Landau density
        function.

    Notes
    -----
    ``pdf_landau_moment1`` evaluates an approximation to the first
    moment Phi_1(x) of the Landau density function given by

        Phi_1(x) = 1/Phi(x)int_- infinity^x lambdaphi(lambda)d lambda,

    where phi(lambda) is described in ``pdf_landau``, using piecewise
    approximation by rational functions.
    Further details can be found in Kölbig and Schorr (1984).

    To obtain the value of Phi_2(x), ``pdf_landau_moment2`` can be used.

    References
    ----------
    Kölbig, K S and Schorr, B, 1984, `A program package for the Landau
    distribution`, Comp. Phys. Comm. (31), 97--111
    """

    _return_scalar = _b_stat.pdf_landau_moment1(
        x,
    )
    return _return_scalar

def pdf_landau_moment2(x):
    u"""
    Landau second moment function Phi_2(x).

    ``pdf_landau_moment2`` returns the value of the second moment
    Phi_2(x) of the Landau density function.

    For full information please refer to the NAG Library document for
    g01qt

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01qtf.html

    Parameters
    ----------
    x : float
        The argument x of the function.

    Returns
    -------
    mom2 : float
        The value of the second moment Phi_2(x) of the Landau density
        function.

    Notes
    -----
    ``pdf_landau_moment2`` evaluates an approximation to the second
    moment Phi_2(x) of the Landau density function given by

        Phi_2(x) = 1/Phi(x)int_- infinity^x lambda^2 phi(lambda)d
        lambda,

    where phi(lambda) is described in ``pdf_landau``, using piecewise
    approximation by rational functions.
    Further details can be found in Kölbig and Schorr (1984).

    To obtain the value of Phi_1(x), ``pdf_landau_moment1`` can be used.

    References
    ----------
    Kölbig, K S and Schorr, B, 1984, `A program package for the Landau
    distribution`, Comp. Phys. Comm. (31), 97--111
    """

    _return_scalar = _b_stat.pdf_landau_moment2(
        x,
    )
    return _return_scalar

def pdf_landau_deriv(x):
    u"""
    Landau derivative function phi'(lambda).

    ``pdf_landau_deriv`` returns the value of the derivative
    phi'(lambda) of the Landau density function.

    For full information please refer to the NAG Library document for
    g01rt

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01rtf.html

    Parameters
    ----------
    x : float
        The argument lambda of the function.

    Returns
    -------
    deriv : float
        The value of the derivative phi'(lambda) of the Landau density
        function.

    Notes
    -----
    ``pdf_landau_deriv`` evaluates an approximation to the derivative
    phi'(lambda) of the Landau density function given by

        phi'(lambda) = d phi(lambda)/d lambda,

    where phi(lambda) is described in ``pdf_landau``, using piecewise
    approximation by rational functions.
    Further details can be found in Kölbig and Schorr (1984).

    To obtain the value of phi(lambda), ``pdf_landau`` can be used.

    References
    ----------
    Kölbig, K S and Schorr, B, 1984, `A program package for the Landau
    distribution`, Comp. Phys. Comm. (31), 97--111
    """

    _return_scalar = _b_stat.pdf_landau_deriv(
        x,
    )
    return _return_scalar

ProbNormalVectorReturnData = _collections.namedtuple(
    'ProbNormalVectorReturnData',
    ['p', 'ivalid'],
)

def prob_normal_vector(tail, x, xmu, xstd):
    u"""
    Computes a vector of probabilities for the standard Normal
    distribution.

    ``prob_normal_vector`` returns a number of one or two tail
    probabilities for the Normal distribution.

    For full information please refer to the NAG Library document for
    g01sa

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01saf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates which tail the returned probabilities should
        represent. Letting Z denote a variate from a standard Normal
        distribution, and z_i = (x_i-mu_i)/(sigma_i), then for j =
        (mod(i-1, ltail)) +1 -1, for i = 1,2,...,max(lx, ltail, lxmu,
        lxstd):

        `tail`[j] = 'L'

            The lower tail probability is returned, i.e., p_i = P(Z <=
            z_i).

        `tail`[j] = 'U'

            The upper tail probability is returned, i.e., p_i = P(Z >=
            z_i).

        `tail`[j] = 'C'

            The two tail (confidence interval) probability is returned,
            i.e., p_i = P(Z <= |z_i|)-P(Z <= -|z_i|).

        `tail`[j] = 'S'

            The two tail (significance level) probability is returned,
            i.e., p_i = P(Z >= |z_i|)+P(Z <= -|z_i|).

    x : float, array-like, shape (lx)
        x_i, the Normal variate values.

    xmu : float, array-like, shape (lxmu)
        mu_i, the means.

    xstd : float, array-like, shape (lxstd)
        sigma_i, the standard deviations.

    Returns
    -------
    p : float, ndarray, shape (max(lx, ltail, lxmu, lxstd))
        p_i, the probabilities for the Normal distribution.

    ivalid : int, ndarray, shape (max(lx, ltail, lxmu, lxstd))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, invalid value supplied in `tail` when calculating
            p_i.

        `ivalid`[i-1] = 2

            On entry, sigma_i <= 0.0.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, ltail = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, lx = *<value>*.

            Constraint: lx > 0.

        (`errno` 4)
            On entry, lxmu = *<value>*.

            Constraint: lxmu > 0.

        (`errno` 5)
            On entry, lxstd = *<value>*.

            Constraint: lxstd > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `tail` or `xstd` was
            invalid.

            Check `ivalid` for more information.

    Notes
    -----
    The lower tail probability for the Normal distribution, P(X_i <=
    x_i) is defined by:

        P(X_i <= x_i) = int_- infinity^x_iZ_i(X_i)dX_i,

    where

        Z_i(X_i) = 1/(sqrt(2 pisigma_i^2))e^(-(X_i-mu_i)^2/(2
        sigma_i^2)), - infinity < X_i < infinity.

    The relationship

        P(X_i <= x_i) = 1/2erfc((-(x_i-mu_i))/sqrt(2)sigma_i)

    is used, where erfc is the complementary error function, and is
    computed using ``specfun.erfc_real``.

    When the two tail confidence probability is required the
    relationship

        P(X_i <= |x_i|)-P(X_i <= -|x_i|) =
        erf(|x_i-mu_i|/sqrt(2)sigma_i),

    is used, where erf is the error function, and is computed using
    ``specfun.erf_real``.

    The input arrays to this function are designed to allow maximum
    flexibility in the supply of vector arguments by re-using elements
    of any arrays that are shorter than the total number of evaluations
    required.
    See [the G01 Introduction] for further information.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth
    """

    fun_name = 'naginterfaces.library.stat.prob_normal_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _lx = _b_utils._EngineIntScalarType(
        _lx_py, {'fun_name': fun_name, 'entity': '_lx'},
    ).to_ctypes()
    _xmu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xmu'},
        xmu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lxmu_py = _b_utils._EngineFloat64ArrayType.get_shape(_xmu, {'fun_name': fun_name, 'entity': 'xmu'}, exp_rank=1)[0]
    _lxmu = _b_utils._EngineIntScalarType(
        _lxmu_py, {'fun_name': fun_name, 'entity': '_lxmu'},
    ).to_ctypes()
    _xstd = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xstd'},
        xstd,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lxstd_py = _b_utils._EngineFloat64ArrayType.get_shape(_xstd, {'fun_name': fun_name, 'entity': 'xstd'}, exp_rank=1)[0]
    _lxstd = _b_utils._EngineIntScalarType(
        _lxstd_py, {'fun_name': fun_name, 'entity': '_lxstd'},
    ).to_ctypes()
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_lx_py, _ltail_py, _lxmu_py, _lxstd_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_lx_py, _ltail_py, _lxmu_py, _lxstd_py),),
    )
    try:
        _b_stat.prob_normal_vector(
            _ltail,
            _tail,
            _lx,
            _x,
            _lxmu,
            _xmu,
            _lxstd,
            _xstd,
            _p,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _p_py = _p
    _ivalid_py = _ivalid
    _return_data = ProbNormalVectorReturnData(
        _p_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ProbStudentsTVectorReturnData = _collections.namedtuple(
    'ProbStudentsTVectorReturnData',
    ['p', 'ivalid'],
)

def prob_students_t_vector(tail, t, df):
    u"""
    Computes a vector of probabilities for the Student's t-distribution.

    ``prob_students_t_vector`` returns a number of one or two tail
    probabilities for the Student's t-distribution with real degrees of
    freedom.

    For full information please refer to the NAG Library document for
    g01sb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01sbf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates which tail the returned probabilities should
        represent. For j = (mod(i-1, ltail)) +1 -1, for i =
        1,2,...,max(ltail, lt, ldf):

        `tail`[j] = 'L'

            The lower tail probability is returned, i.e., p_i = (T_i <=
            t_i:nu_i).

        `tail`[j] = 'U'

            The upper tail probability is returned, i.e., p_i = (T_i >=
            t_i:nu_i).

        `tail`[j] = 'C'

            The two tail (confidence interval) probability is returned,
            i.e., p_i = (T_i <= |t_i|:nu_i)-(T_i <= -|t_i|:nu_i).

        `tail`[j] = 'S'

            The two tail (significance level) probability is returned,
            i.e., p_i = (T_i >= |t_i|:nu_i)+(T_i <= -|t_i|:nu_i).

    t : float, array-like, shape (lt)
        t_i, the values of the Student's t variates.

    df : float, array-like, shape (ldf)
        nu_i, the degrees of freedom of the Student's t-distribution.

    Returns
    -------
    p : float, ndarray, shape (max(ltail, lt, ldf))
        p_i, the probabilities for the Student's t distribution.

    ivalid : int, ndarray, shape (max(ltail, lt, ldf))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, invalid value supplied in `tail` when calculating
            p_i.

        `ivalid`[i-1] = 2

            On entry, nu_i < 1.0.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lt > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: ldf > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `tail` or `df` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    The lower tail probability for the Student's t-distribution with
    nu_i degrees of freedom, (T_i <= t_i:nu_i) is defined by:

        (T_i <= t_i:nu_i) =
        Gamma((nu_i+1)/2)/(sqrt(pinu_i)Gamma(nu_i/2))int_-
        infinity^t_i[1+(T_i^2)/(nu_i)]^(-(nu_i+1)/2)dT_i, nu_i >= 1.

    Computationally, there are two situations:

    (i) when nu_i < 20, a transformation of the beta distribution, P
        beta_i(B_i <= beta_i:a_i,b_i) is used

            (T_i <= t_i:nu_i) = 1/2P beta_i(B_i <=
            (nu_i)/(nu_i+t_i^2):nu_i/2,1/2) when t_i < 0.0

        or

            (T_i <= t_i:nu_i) = 1/2+1/2P beta_i(B_i >=
            nu_i/(nu_i+t_i^2):nu_i/2,1/2) when t_i > 0.0;

    (#) when nu_i >= 20, an asymptotic normalizing expansion of the
        Cornish--Fisher type is used to evaluate the probability, see
        Hill (1970).

    The input arrays to this function are designed to allow maximum
    flexibility in the supply of vector arguments by re-using elements
    of any arrays that are shorter than the total number of evaluations
    required.
    See [the G01 Introduction] for further information.

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth

    Hill, G W, 1970, `Student's t-distribution`, Comm. ACM (13(10)),
    617--619
    """

    fun_name = 'naginterfaces.library.stat.prob_students_t_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _t = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 't'},
        t,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lt_py = _b_utils._EngineFloat64ArrayType.get_shape(_t, {'fun_name': fun_name, 'entity': 't'}, exp_rank=1)[0]
    _lt = _b_utils._EngineIntScalarType(
        _lt_py, {'fun_name': fun_name, 'entity': '_lt'},
    ).to_ctypes()
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ldf_py = _b_utils._EngineFloat64ArrayType.get_shape(_df, {'fun_name': fun_name, 'entity': 'df'}, exp_rank=1)[0]
    _ldf = _b_utils._EngineIntScalarType(
        _ldf_py, {'fun_name': fun_name, 'entity': '_ldf'},
    ).to_ctypes()
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lt_py, _ldf_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lt_py, _ldf_py),),
    )
    try:
        _b_stat.prob_students_t_vector(
            _ltail,
            _tail,
            _lt,
            _t,
            _ldf,
            _df,
            _p,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _p_py = _p
    _ivalid_py = _ivalid
    _return_data = ProbStudentsTVectorReturnData(
        _p_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ProbChisqVectorReturnData = _collections.namedtuple(
    'ProbChisqVectorReturnData',
    ['p', 'ivalid'],
)

def prob_chisq_vector(tail, x, df):
    u"""
    Computes a vector of probabilities for chi^2 distribution.

    ``prob_chisq_vector`` returns a number of lower or upper tail
    probabilities for the chi^2-distribution with real degrees of
    freedom.

    For full information please refer to the NAG Library document for
    g01sc

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01scf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates whether the lower or upper tail probabilities are
        required. For j = (mod(i-1, ltail)) +1 -1, for i =
        1,2,...,max(ltail, lx, ldf):

        `tail`[j] = 'L'

            The lower tail probability is returned, i.e., p_i = (X_i <=
            x_i:nu_i).

        `tail`[j] = 'U'

            The upper tail probability is returned, i.e., p_i = (X_i >=
            x_i:nu_i).

    x : float, array-like, shape (lx)
        x_i, the values of the chi^2 variates with nu_i degrees of
        freedom.

    df : float, array-like, shape (ldf)
        nu_i, the degrees of freedom of the chi^2-distribution.

    Returns
    -------
    p : float, ndarray, shape (max(ltail, ldf, lx))
        p_i, the probabilities for the chi^2 distribution.

    ivalid : int, ndarray, shape (max(ltail, ldf, lx))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, invalid value supplied in `tail` when calculating
            p_i.

        `ivalid`[i-1] = 2

            On entry, x_i < 0.0.

        `ivalid`[i-1] = 3

            On entry, nu_i <= 0.0.

        `ivalid`[i-1] = 4

            The solution has failed to converge while calculating the
            gamma variate. The result returned should represent an
            approximation to the solution.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lx > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: ldf > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `x`, `df` or `tail` was
            invalid, or the solution failed to converge.

            Check `ivalid` for more information.

    Notes
    -----
    The lower tail probability for the chi^2-distribution with nu_i
    degrees of freedom, P = (X_i <= x_i:nu_i) is defined by:

        P = (X_i <= x_i:nu_i) =
        1/(2^(nu_i/2)Gamma(nu_i/2))int_0.0^x_iX_i^(nu_i/2-1)e^(-X_i/2)dX_i,
        x_i >= 0, nu_i > 0.

    To calculate P = (X_i <= x_i:nu_i) a transformation of a gamma
    distribution is employed, i.e., a chi^2-distribution with nu_i
    degrees of freedom is equal to a gamma distribution with scale
    parameter 2 and shape parameter nu_i/2.

    The input arrays to this function are designed to allow maximum
    flexibility in the supply of vector arguments by re-using elements
    of any arrays that are shorter than the total number of evaluations
    required.
    See [the G01 Introduction] for further information.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth
    """

    fun_name = 'naginterfaces.library.stat.prob_chisq_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _lx = _b_utils._EngineIntScalarType(
        _lx_py, {'fun_name': fun_name, 'entity': '_lx'},
    ).to_ctypes()
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ldf_py = _b_utils._EngineFloat64ArrayType.get_shape(_df, {'fun_name': fun_name, 'entity': 'df'}, exp_rank=1)[0]
    _ldf = _b_utils._EngineIntScalarType(
        _ldf_py, {'fun_name': fun_name, 'entity': '_ldf'},
    ).to_ctypes()
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _ldf_py, _lx_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _ldf_py, _lx_py),),
    )
    try:
        _b_stat.prob_chisq_vector(
            _ltail,
            _tail,
            _lx,
            _x,
            _ldf,
            _df,
            _p,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _p_py = _p
    _ivalid_py = _ivalid
    _return_data = ProbChisqVectorReturnData(
        _p_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ProbFVectorReturnData = _collections.namedtuple(
    'ProbFVectorReturnData',
    ['p', 'ivalid'],
)

def prob_f_vector(tail, f, df1, df2):
    u"""
    Computes a vector of probabilities for F-distribution.

    ``prob_f_vector`` returns a number of lower or upper tail
    probabilities for the F or variance-ratio distribution with real
    degrees of freedom.

    For full information please refer to the NAG Library document for
    g01sd

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01sdf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates whether the lower or upper tail probabilities are
        required. For j = (mod(i-1, ltail)) +1 -1, for i =
        1,2,...,max(ltail, lf, ldf1, ldf2):

        `tail`[j] = 'L'

            The lower tail probability is returned, i.e., p_i = (F_i <=
            f_i:u_i,v_i).

        `tail`[j] = 'U'

            The upper tail probability is returned, i.e., p_i = (F_i >=
            f_i:u_i,v_i).

    f : float, array-like, shape (lf)
        f_i, the value of the F variate.

    df1 : float, array-like, shape (ldf1)
        u_i, the degrees of freedom of the numerator variance.

    df2 : float, array-like, shape (ldf2)
        v_i, the degrees of freedom of the denominator variance.

    Returns
    -------
    p : float, ndarray, shape (max(ltail, lf, ldf1, ldf2))
        p_i, the probabilities for the F-distribution.

    ivalid : int, ndarray, shape (max(ltail, lf, ldf1, ldf2))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, invalid value supplied in `tail` when calculating
            p_i.

        `ivalid`[i-1] = 2

            On entry, f_i < 0.0.

        `ivalid`[i-1] = 3

            On entry, u_i <= 0.0, or, v_i <= 0.0.

        `ivalid`[i-1] = 4

            The solution has failed to converge. The result returned
            should represent an approximation to the solution.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lf > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: ldf1 > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: ldf2 > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `f`, `df1`, `df2` or `tail`
            was invalid, or the solution failed to converge.

            Check `ivalid` for more information.

    Notes
    -----
    The lower tail probability for the F, or variance-ratio,
    distribution with u_i and v_i degrees of freedom, (F_i <=
    f_i:u_i,v_i), is defined by:

        (F_i <= f_i:u_i,v_i) =
        u_i^(u_i/2)v_i^(v_i/2)Gamma((u_i+v_i)/2)/(Gamma(u_i/2)Gamma(v_i/2))int_0^f_iF_i^((u_i-2)/2)(u_i
        F_i+v_i)^(-(u_i+v_i)/2)dF_i,

    for u_i, v_i > 0, f_i >= 0.

    The probability is computed by means of a transformation to a beta
    distribution, P beta_i(B_i <= beta_i:a_i,b_i):

        (F_i <= f_i:u_i,v_i) = P beta_i(B_i <= u_i f_i/(u_i
        f_i+v_i):u_i/2,v_i/2)

    and using a call to ``prob_beta``.

    For very large values of both u_i and v_i, greater than 10^5, a
    normal approximation is used.
    If only one of u_i or v_i is greater than 10^5 then a chi^2
    approximation is used, see Abramowitz and Stegun (1972).

    The input arrays to this function are designed to allow maximum
    flexibility in the supply of vector arguments by re-using elements
    of any arrays that are shorter than the total number of evaluations
    required.
    See [the G01 Introduction] for further information.

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth
    """

    fun_name = 'naginterfaces.library.stat.prob_f_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lf_py = _b_utils._EngineFloat64ArrayType.get_shape(_f, {'fun_name': fun_name, 'entity': 'f'}, exp_rank=1)[0]
    _lf = _b_utils._EngineIntScalarType(
        _lf_py, {'fun_name': fun_name, 'entity': '_lf'},
    ).to_ctypes()
    _df1 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df1'},
        df1,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ldf1_py = _b_utils._EngineFloat64ArrayType.get_shape(_df1, {'fun_name': fun_name, 'entity': 'df1'}, exp_rank=1)[0]
    _ldf1 = _b_utils._EngineIntScalarType(
        _ldf1_py, {'fun_name': fun_name, 'entity': '_ldf1'},
    ).to_ctypes()
    _df2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df2'},
        df2,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ldf2_py = _b_utils._EngineFloat64ArrayType.get_shape(_df2, {'fun_name': fun_name, 'entity': 'df2'}, exp_rank=1)[0]
    _ldf2 = _b_utils._EngineIntScalarType(
        _ldf2_py, {'fun_name': fun_name, 'entity': '_ldf2'},
    ).to_ctypes()
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lf_py, _ldf1_py, _ldf2_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lf_py, _ldf1_py, _ldf2_py),),
    )
    try:
        _b_stat.prob_f_vector(
            _ltail,
            _tail,
            _lf,
            _f,
            _ldf1,
            _df1,
            _ldf2,
            _df2,
            _p,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _p_py = _p
    _ivalid_py = _ivalid
    _return_data = ProbFVectorReturnData(
        _p_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ProbBetaVectorReturnData = _collections.namedtuple(
    'ProbBetaVectorReturnData',
    ['p', 'ivalid'],
)

def prob_beta_vector(tail, beta, a, b):
    u"""
    Computes a vector of probabilities for the beta distribution.

    ``prob_beta_vector`` computes a number of lower or upper tail
    probabilities for the beta distribution.

    For full information please refer to the NAG Library document for
    g01se

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01sef.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates whether a lower or upper tail probabilities are
        required. For j = (mod(i-1, ltail)) +1 -1, for i =
        1,2,...,max(ltail, lbeta, la, lb):

        `tail`[j] = 'L'

            The lower tail probability is returned, i.e., p_i = (B_i <=
            beta_i:a_i,b_i).

        `tail`[j] = 'U'

            The upper tail probability is returned, i.e., p_i = (B_i >=
            beta_i:a_i,b_i).

    beta : float, array-like, shape (lbeta)
        beta_i, the value of the beta variate.

    a : float, array-like, shape (la)
        a_i, the first parameter of the required beta distribution.

    b : float, array-like, shape (lb)
        b_i, the second parameter of the required beta distribution.

    Returns
    -------
    p : float, ndarray, shape (max(ltail, lbeta, la, lb))
        p_i, the probabilities for the beta distribution.

    ivalid : int, ndarray, shape (max(ltail, lbeta, la, lb))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, invalid value supplied in `tail` when calculating
            p_i.

        `ivalid`[i-1] = 2

            On entry, beta_i < 0.0, or, beta_i > 1.0.

        `ivalid`[i-1] = 3

            On entry, a_i <= 0.0, or, b_i <= 0.0.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lbeta > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: la > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: lb > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `beta`, `a`, `b` or `tail`
            was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    The lower tail probability, (B_i <= beta_i:a_i,b_i) is defined by

        (B_i <= beta_i:a_i,b_i) =
        Gamma(a_i+b_i)/(Gamma(a_i)Gamma(b_i))int_0^beta_iB_i^(a_i-1)(1-B_i)^(b_i-1)dB_i
        = I_beta_i(a_i,b_i), 0 <= beta_i <= 1; a_i,b_i > 0.

    The function I_beta_i(a_i,b_i), also known as the incomplete beta
    function is calculated using ``specfun.beta_incomplete``.

    The input arrays to this function are designed to allow maximum
    flexibility in the supply of vector arguments by re-using elements
    of any arrays that are shorter than the total number of evaluations
    required.
    See [the G01 Introduction] for further information.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth

    Majumder, K L and Bhattacharjee, G P, 1973, `Algorithm AS 63. The
    incomplete beta integral`, Appl. Statist. (22), 409--411
    """

    fun_name = 'naginterfaces.library.stat.prob_beta_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _beta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'beta'},
        beta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lbeta_py = _b_utils._EngineFloat64ArrayType.get_shape(_beta, {'fun_name': fun_name, 'entity': 'beta'}, exp_rank=1)[0]
    _lbeta = _b_utils._EngineIntScalarType(
        _lbeta_py, {'fun_name': fun_name, 'entity': '_lbeta'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _la_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _la = _b_utils._EngineIntScalarType(
        _la_py, {'fun_name': fun_name, 'entity': '_la'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lb_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=1)[0]
    _lb = _b_utils._EngineIntScalarType(
        _lb_py, {'fun_name': fun_name, 'entity': '_lb'},
    ).to_ctypes()
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lbeta_py, _la_py, _lb_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lbeta_py, _la_py, _lb_py),),
    )
    try:
        _b_stat.prob_beta_vector(
            _ltail,
            _tail,
            _lbeta,
            _beta,
            _la,
            _a,
            _lb,
            _b,
            _p,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _p_py = _p
    _ivalid_py = _ivalid
    _return_data = ProbBetaVectorReturnData(
        _p_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ProbGammaVectorReturnData = _collections.namedtuple(
    'ProbGammaVectorReturnData',
    ['p', 'ivalid'],
)

def prob_gamma_vector(tail, g, a, b):
    u"""
    Computes a vector of probabilities for the gamma distribution.

    ``prob_gamma_vector`` returns a number of lower or upper tail
    probabilities for the gamma distribution.

    For full information please refer to the NAG Library document for
    g01sf

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01sff.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates whether a lower or upper tail probability is required.
        For j = (mod(i-1, ltail)) +1 -1, for i = 1,2,...,max(ltail, lg,
        la, lb):

        `tail`[j] = 'L'

            The lower tail probability is returned, i.e., p_i = (G_i <=
            g_i:alpha_i,beta_i).

        `tail`[j] = 'U'

            The upper tail probability is returned, i.e., p_i = (G_i >=
            g_i:alpha_i,beta_i).

    g : float, array-like, shape (lg)
        g_i, the value of the gamma variate.

    a : float, array-like, shape (la)
        The parameter alpha_i of the gamma distribution.

    b : float, array-like, shape (lb)
        The parameter beta_i of the gamma distribution.

    Returns
    -------
    p : float, ndarray, shape (max(lg, la, lb, ltail))
        p_i, the probabilities of the beta distribution.

    ivalid : int, ndarray, shape (max(lg, la, lb, ltail))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, invalid value supplied in `tail` when calculating
            p_i.

        `ivalid`[i-1] = 2

            On entry, g_i < 0.0.

        `ivalid`[i-1] = 3

            On entry, alpha_i <= 0.0, or, beta_i <= 0.0.

        `ivalid`[i-1] = 4

            The solution did not converge in 600 iterations, see
            ``specfun.gamma_incomplete`` for details. The probability
            returned should be a reasonable approximation to the
            solution.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lg > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: la > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: lb > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `g`, `a`, `b` or `tail` was
            invalid, or the solution did not converge.

            Check `ivalid` for more information.

    Notes
    -----
    The lower tail probability for the gamma distribution with
    parameters alpha_i and beta_i, P(G_i <= g_i), is defined by:

        (G_i <= g_i:alpha_i,beta_i) =
        1/(beta_i^(alpha_i)Gamma(alpha_i))int_0^g_iG_i^(alpha_i-1)e^(-G_i/beta_i)dG_i,
        alpha_i > 0.0, beta_i > 0.0.

    The mean of the distribution is alpha_ibeta_i and its variance is
    alpha_ibeta_i^2.
    The transformation Z_i = (G_i)/(beta_i) is applied to yield the
    following incomplete gamma function in normalized form,

        (G_i <= g_i:alpha_i,beta_i) = (Z_i <= g_i/beta_i:alpha_i,1.0) =
        1/Gamma(alpha_i)int_0^g_i/beta_iZ_i^(alpha_i-1)e^(-Z_i)dZ_i.

    This is then evaluated using ``specfun.gamma_incomplete``.

    The input arrays to this function are designed to allow maximum
    flexibility in the supply of vector arguments by re-using elements
    of any arrays that are shorter than the total number of evaluations
    required.
    See [the G01 Introduction] for further information.

    References
    ----------
    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth
    """

    fun_name = 'naginterfaces.library.stat.prob_gamma_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _g = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'g'},
        g,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lg_py = _b_utils._EngineFloat64ArrayType.get_shape(_g, {'fun_name': fun_name, 'entity': 'g'}, exp_rank=1)[0]
    _lg = _b_utils._EngineIntScalarType(
        _lg_py, {'fun_name': fun_name, 'entity': '_lg'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _la_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _la = _b_utils._EngineIntScalarType(
        _la_py, {'fun_name': fun_name, 'entity': '_la'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lb_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=1)[0]
    _lb = _b_utils._EngineIntScalarType(
        _lb_py, {'fun_name': fun_name, 'entity': '_lb'},
    ).to_ctypes()
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_lg_py, _la_py, _lb_py, _ltail_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_lg_py, _la_py, _lb_py, _ltail_py),),
    )
    try:
        _b_stat.prob_gamma_vector(
            _ltail,
            _tail,
            _lg,
            _g,
            _la,
            _a,
            _lb,
            _b,
            _p,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _p_py = _p
    _ivalid_py = _ivalid
    _return_data = ProbGammaVectorReturnData(
        _p_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ProbBinomialVectorReturnData = _collections.namedtuple(
    'ProbBinomialVectorReturnData',
    ['plek', 'pgtk', 'peqk', 'ivalid'],
)

def prob_binomial_vector(n, p, k):
    u"""
    Computes a vector of probabilities for the binomial distribution.

    ``prob_binomial_vector`` returns a number of the lower tail, upper
    tail and point probabilities for the binomial distribution.

    For full information please refer to the NAG Library document for
    g01sj

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01sjf.html

    Parameters
    ----------
    n : int, array-like, shape (ln)
        n_i, the first parameter of the binomial distribution.

    p : float, array-like, shape (lp)
        p_i, the second parameter of the binomial distribution.

    k : int, array-like, shape (lk)
        k_i, the integer which defines the required probabilities.

    Returns
    -------
    plek : float, ndarray, shape (max(ln, lp, lk))
        Prob{X_i <= k_i}, the lower tail probabilities.

    pgtk : float, ndarray, shape (max(ln, lp, lk))
        Prob{X_i > k_i}, the upper tail probabilities.

    peqk : float, ndarray, shape (max(ln, lp, lk))
        Prob{X_i = k_i}, the point probabilities.

    ivalid : int, ndarray, shape (max(ln, lp, lk))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, n_i < 0.

        `ivalid`[i-1] = 2

            On entry, p_i <= 0.0, or, p_i >= 1.0.

        `ivalid`[i-1] = 3

            On entry, k_i < 0, or, k_i > n_i.

        `ivalid`[i-1] = 4

            On entry, n_i is too large to be represented exactly as a
            real number.

        `ivalid`[i-1] = 5

            On entry, the variance (= n_i p_i(1-p_i)) exceeds 10^6.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ln > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lp > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: lk > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `n`, `p` or `k` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    Let X = {X_i:i = 1,2,...,m} denote a vector of random variables each
    having a binomial distribution with parameters n_i and p_i (n_i >= 0
    and 0 < p_i < 1).
    Then

        Prob{X_i = k_i} = (n_i;k_i)p_i^(k_i)(1-p_i)^(n_i-k_i), k_i =
        0,1,...,n_i.

    The mean of the each distribution is given by n_i p_i and the
    variance by n_i p_i(1-p_i).

    ``prob_binomial_vector`` computes, for given n_i, p_i and k_i, the
    probabilities: Prob{X_i <= k_i}, Prob{X_i > k_i} and Prob{X_i = k_i}
    using an algorithm similar to that described in Knüsel (1986) for
    the Poisson distribution.

    The input arrays to this function are designed to allow maximum
    flexibility in the supply of vector arguments by re-using elements
    of any arrays that are shorter than the total number of evaluations
    required.
    See [the G01 Introduction] for further information.

    References
    ----------
    Knüsel, L, 1986, `Computation of the chi-square and Poisson
    distribution`, SIAM J. Sci. Statist. Comput. (7), 1022--1036
    """

    fun_name = 'naginterfaces.library.stat.prob_binomial_vector'

    _n = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'n'},
        n,
        copy=False,
    )
    _ln_py = _b_utils._EngineIntArrayType.get_shape(_n, {'fun_name': fun_name, 'entity': 'n'}, exp_rank=1)[0]
    _ln = _b_utils._EngineIntScalarType(
        _ln_py, {'fun_name': fun_name, 'entity': '_ln'},
    ).to_ctypes()
    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lp_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _lp = _b_utils._EngineIntScalarType(
        _lp_py, {'fun_name': fun_name, 'entity': '_lp'},
    ).to_ctypes()
    _k = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'k'},
        k,
        copy=False,
    )
    _lk_py = _b_utils._EngineIntArrayType.get_shape(_k, {'fun_name': fun_name, 'entity': 'k'}, exp_rank=1)[0]
    _lk = _b_utils._EngineIntScalarType(
        _lk_py, {'fun_name': fun_name, 'entity': '_lk'},
    ).to_ctypes()
    _plek = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ln_py, _lp_py, _lk_py),),
    )
    _pgtk = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ln_py, _lp_py, _lk_py),),
    )
    _peqk = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ln_py, _lp_py, _lk_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ln_py, _lp_py, _lk_py),),
    )
    try:
        _b_stat.prob_binomial_vector(
            _ln,
            _n,
            _lp,
            _p,
            _lk,
            _k,
            _plek,
            _pgtk,
            _peqk,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _plek_py = _plek
    _pgtk_py = _pgtk
    _peqk_py = _peqk
    _ivalid_py = _ivalid
    _return_data = ProbBinomialVectorReturnData(
        _plek_py,
        _pgtk_py,
        _peqk_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ProbPoissonVectorReturnData = _collections.namedtuple(
    'ProbPoissonVectorReturnData',
    ['plek', 'pgtk', 'peqk', 'ivalid'],
)

def prob_poisson_vector(l, k):
    u"""
    Computes a vector of probabilities for the Poisson distribution.

    ``prob_poisson_vector`` returns a number of the lower tail, upper
    tail and point probabilities for the Poisson distribution.

    For full information please refer to the NAG Library document for
    g01sk

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01skf.html

    Parameters
    ----------
    l : float, array-like, shape (ll)
        lambda_i, the parameter of the Poisson distribution.

    k : int, array-like, shape (lk)
        k_i, the integer which defines the required probabilities.

    Returns
    -------
    plek : float, ndarray, shape (max(ll, lk))
        Prob{X_i <= k_i}, the lower tail probabilities.

    pgtk : float, ndarray, shape (max(ll, lk))
        Prob{X_i > k_i}, the upper tail probabilities.

    peqk : float, ndarray, shape (max(ll, lk))
        Prob{X_i = k_i}, the point probabilities.

    ivalid : int, ndarray, shape (max(ll, lk))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, lambda_i <= 0.0.

        `ivalid`[i-1] = 2

            On entry, k_i < 0.

        `ivalid`[i-1] = 3

            On entry, lambda_i > 10^6.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ll > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lk > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `l` or `k` was invalid.

            Check `ivalid` for more information.

    Notes
    -----
    Let X = {X_i:i = 1,2,...,m} denote a vector of random variables each
    having a Poisson distribution with parameter lambda_i (> 0).
    Then

        Prob{X_i = k_i} = e^(-lambda_i)(lambda_i^(k_i))/k_i!, k_i =
        0,1,2,...

    The mean and variance of each distribution are both equal to
    lambda_i.

    ``prob_poisson_vector`` computes, for given lambda_i and k_i the
    probabilities: Prob{X_i <= k_i}, Prob{X_i > k_i} and Prob{X_i = k_i}
    using the algorithm described in Knüsel (1986).

    The input arrays to this function are designed to allow maximum
    flexibility in the supply of vector arguments by re-using elements
    of any arrays that are shorter than the total number of evaluations
    required.
    See [the G01 Introduction] for further information.

    References
    ----------
    Knüsel, L, 1986, `Computation of the chi-square and Poisson
    distribution`, SIAM J. Sci. Statist. Comput. (7), 1022--1036
    """

    fun_name = 'naginterfaces.library.stat.prob_poisson_vector'

    _l = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'l'},
        l,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ll_py = _b_utils._EngineFloat64ArrayType.get_shape(_l, {'fun_name': fun_name, 'entity': 'l'}, exp_rank=1)[0]
    _ll = _b_utils._EngineIntScalarType(
        _ll_py, {'fun_name': fun_name, 'entity': '_ll'},
    ).to_ctypes()
    _k = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'k'},
        k,
        copy=False,
    )
    _lk_py = _b_utils._EngineIntArrayType.get_shape(_k, {'fun_name': fun_name, 'entity': 'k'}, exp_rank=1)[0]
    _lk = _b_utils._EngineIntScalarType(
        _lk_py, {'fun_name': fun_name, 'entity': '_lk'},
    ).to_ctypes()
    _plek = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ll_py, _lk_py),),
    )
    _pgtk = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ll_py, _lk_py),),
    )
    _peqk = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ll_py, _lk_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ll_py, _lk_py),),
    )
    try:
        _b_stat.prob_poisson_vector(
            _ll,
            _l,
            _lk,
            _k,
            _plek,
            _pgtk,
            _peqk,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _plek_py = _plek
    _pgtk_py = _pgtk
    _peqk_py = _peqk
    _ivalid_py = _ivalid
    _return_data = ProbPoissonVectorReturnData(
        _plek_py,
        _pgtk_py,
        _peqk_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

ProbHypergeomVectorReturnData = _collections.namedtuple(
    'ProbHypergeomVectorReturnData',
    ['plek', 'pgtk', 'peqk', 'ivalid'],
)

def prob_hypergeom_vector(n, l, m, k):
    u"""
    Computes a vector of probabilities for the hypergeometric
    distribution.

    ``prob_hypergeom_vector`` returns a number of the lower tail, upper
    tail and point probabilities for the hypergeometric distribution.

    For full information please refer to the NAG Library document for
    g01sl

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01slf.html

    Parameters
    ----------
    n : int, array-like, shape (ln)
        n_i, the parameter of the hypergeometric distribution.

    l : int, array-like, shape (ll)
        l_i, the parameter of the hypergeometric distribution.

    m : int, array-like, shape (lm)
        m_i, the parameter of the hypergeometric distribution.

    k : int, array-like, shape (lk)
        k_i, the integer which defines the required probabilities.

    Returns
    -------
    plek : float, ndarray, shape (max(ln, ll, lm, lk))
        Prob{X_i <= k_i}, the lower tail probabilities.

    pgtk : float, ndarray, shape (max(ln, ll, lm, lk))
        Prob{X_i > k_i}, the upper tail probabilities.

    peqk : float, ndarray, shape (max(ln, ll, lm, lk))
        Prob{X_i = k_i}, the point probabilities.

    ivalid : int, ndarray, shape (max(ln, ll, lm, lk))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, n_i < 0.

        `ivalid`[i-1] = 2

            On entry, l_i < 0, or, l_i > n_i.

        `ivalid`[i-1] = 3

            On entry, m_i < 0, or, m_i > n_i.

        `ivalid`[i-1] = 4

            On entry, k_i < 0, or, k_i > l_i, or, k_i > m_i, or, k_i <
            l_i+m_i-n_i.

        `ivalid`[i-1] = 5

            On entry, n_i is too large to be represented exactly as a
            real number.

        `ivalid`[i-1] = 6

            On entry, the variance (see [Notes]) exceeds 10^6.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ln > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: ll > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: lm > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: lk > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `n`, `l`, `m` or `k` was
            invalid, or the variance was too large.

            Check `ivalid` for more information.

    Notes
    -----
    Let X = {X_i:i = 1,2,...,r} denote a vector of random variables
    having a hypergeometric distribution with parameters n_i, l_i and
    m_i.
    Then

        Prob{X_i = k_i} = (m_i;k_i)(n_i - m_i ; l_i - k_i)/((n_i;l_i)),

    where max(0, l_i+m_i-n_i) <= k_i <= min(l_i, m_i), 0 <= l_i <= n_i
    and 0 <= m_i <= n_i.

    The hypergeometric distribution may arise if in a population of size
    n_i a number m_i are marked.
    From this population a sample of size l_i is drawn and of these k_i
    are observed to be marked.

    The mean of the distribution = l_i m_i/(n_i), and the variance = l_i
    m_i(n_i-l_i)(n_i-m_i)/n_i^2(n_i-1).

    ``prob_hypergeom_vector`` computes for given n_i, l_i, m_i and k_i
    the probabilities: Prob{X_i <= k_i}, Prob{X_i > k_i} and Prob{X_i =
    k_i} using an algorithm similar to that described in Knüsel (1986)
    for the Poisson distribution.

    The input arrays to this function are designed to allow maximum
    flexibility in the supply of vector arguments by re-using elements
    of any arrays that are shorter than the total number of evaluations
    required.
    See [the G01 Introduction] for further information.

    References
    ----------
    Knüsel, L, 1986, `Computation of the chi-square and Poisson
    distribution`, SIAM J. Sci. Statist. Comput. (7), 1022--1036
    """

    fun_name = 'naginterfaces.library.stat.prob_hypergeom_vector'

    _n = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'n'},
        n,
        copy=False,
    )
    _ln_py = _b_utils._EngineIntArrayType.get_shape(_n, {'fun_name': fun_name, 'entity': 'n'}, exp_rank=1)[0]
    _ln = _b_utils._EngineIntScalarType(
        _ln_py, {'fun_name': fun_name, 'entity': '_ln'},
    ).to_ctypes()
    _l = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'l'},
        l,
        copy=False,
    )
    _ll_py = _b_utils._EngineIntArrayType.get_shape(_l, {'fun_name': fun_name, 'entity': 'l'}, exp_rank=1)[0]
    _ll = _b_utils._EngineIntScalarType(
        _ll_py, {'fun_name': fun_name, 'entity': '_ll'},
    ).to_ctypes()
    _m = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'm'},
        m,
        copy=False,
    )
    _lm_py = _b_utils._EngineIntArrayType.get_shape(_m, {'fun_name': fun_name, 'entity': 'm'}, exp_rank=1)[0]
    _lm = _b_utils._EngineIntScalarType(
        _lm_py, {'fun_name': fun_name, 'entity': '_lm'},
    ).to_ctypes()
    _k = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'k'},
        k,
        copy=False,
    )
    _lk_py = _b_utils._EngineIntArrayType.get_shape(_k, {'fun_name': fun_name, 'entity': 'k'}, exp_rank=1)[0]
    _lk = _b_utils._EngineIntScalarType(
        _lk_py, {'fun_name': fun_name, 'entity': '_lk'},
    ).to_ctypes()
    _plek = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ln_py, _ll_py, _lm_py, _lk_py),),
    )
    _pgtk = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ln_py, _ll_py, _lm_py, _lk_py),),
    )
    _peqk = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ln_py, _ll_py, _lm_py, _lk_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ln_py, _ll_py, _lm_py, _lk_py),),
    )
    try:
        _b_stat.prob_hypergeom_vector(
            _ln,
            _n,
            _ll,
            _l,
            _lm,
            _m,
            _lk,
            _k,
            _plek,
            _pgtk,
            _peqk,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _plek_py = _plek
    _pgtk_py = _pgtk
    _peqk_py = _peqk
    _ivalid_py = _ivalid
    _return_data = ProbHypergeomVectorReturnData(
        _plek_py,
        _pgtk_py,
        _peqk_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

InvCdfNormalVectorReturnData = _collections.namedtuple(
    'InvCdfNormalVectorReturnData',
    ['x', 'ivalid'],
)

def inv_cdf_normal_vector(tail, p, xmu, xstd):
    u"""
    Computes a vector of deviates for the standard Normal distribution.

    ``inv_cdf_normal_vector`` returns a number of deviates associated
    with given probabilities of the Normal distribution.

    For full information please refer to the NAG Library document for
    g01ta

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01taf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates which tail the supplied probabilities represent.
        Letting Z denote a variate from a standard Normal distribution,
        and z_i = (x_p_i-mu_i)/(sigma_i), then for j = (mod(i-1, ltail))
        +1 -1, for i = 1,2,...,max(ltail, lp, lxmu, lxstd):

        `tail`[j] = 'L'

            The lower tail probability, i.e., p_i = P(Z <= z_i).

        `tail`[j] = 'U'

            The upper tail probability, i.e., p_i = P(Z >= z_i).

        `tail`[j] = 'C'

            The two tail (confidence interval) probability, i.e., p_i =
            P(Z <= |z_i|)-P(Z <= -|z_i|).

        `tail`[j] = 'S'

            The two tail (significance level) probability, i.e., p_i =
            P(Z >= |z_i|)+P(Z <= -|z_i|).

    p : float, array-like, shape (lp)
        p_i, the probabilities for the Normal distribution as defined by
        `tail` with p_i = `p`[j], j = mod(i-1, lp).

    xmu : float, array-like, shape (lxmu)
        mu_i, the means.

    xstd : float, array-like, shape (lxstd)
        sigma_i, the standard deviations.

    Returns
    -------
    x : float, ndarray, shape (max(ltail, lxmu, lxstd, lp))
        x_p_i, the deviates for the Normal distribution.

    ivalid : int, ndarray, shape (max(ltail, lxmu, lxstd, lp))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, invalid value supplied in `tail` when calculating
            x_p_i.

        `ivalid`[i-1] = 2

            On entry, p_i <= 0.0, or, p_i >= 1.0.

        `ivalid`[i-1] = 3

            On entry, sigma_i <= 0.0.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lp > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: lxmu > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: lxstd > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `tail`, `xstd` or `p` was
            invalid.

            Check `ivalid` for more information.

    Notes
    -----
    The deviate, x_p_i associated with the lower tail probability, p_i,
    for the Normal distribution is defined as the solution to

        P(X_i <= x_p_i) = p_i = int_- infinity^x_p_iZ_i(X_i)dX_i,

    where

        Z_i(X_i) = 1/(sqrt(2 pisigma_i^2))e^(-(X_i-mu_i)^2/(2
        sigma_i^2)), - infinity < X_i < infinity.

    The method used is an extension of that of Wichura (1988). p_i is
    first replaced by q_i = p_i-0.5.

    (a) If |q_i| <= 0.3, z_i is computed by a rational Chebyshev
        approximation

            z_i = s_i A_i(s_i^2)/B_i(s_i^2),

        where s_i = sqrt(2 pi)q_i and A_i, B_i are polynomials of degree
        7.

    (#) If 0.3 < |q_i| <= 0.42, z_i is computed by a rational Chebyshev
        approximation

            z_i = sign(q_i)(C_i(t_i)/D_i(t_i)),

        where t_i = |q_i|-0.3 and C_i, D_i are polynomials of degree 5.

    (#) If |q_i| > 0.42, z_i is computed as

            z_i = sign(q_i)[(E_i(u_i)/F_i(u_i))+u_i],

        where u_i = sqrt(-2*log(min(p_i, 1-p_i))) and E_i, F_i are
        polynomials of degree 6.

    x_p_i is then calculated from z_i, using the relationsship z_p_i =
    (x_i-mu_i)/(sigma_i).

    For the upper tail probability -x_p_i is returned, while for the two
    tail probabilities the value x_{ip_i*} is returned, where p_i* is
    the required tail probability computed from the input value of p_i.

    The input arrays to this function are designed to allow maximum
    flexibility in the supply of vector arguments by re-using elements
    of any arrays that are shorter than the total number of evaluations
    required.
    See [the G01 Introduction] for further information.

    References
    ----------
    NIST Digital Library of Mathematical Functions

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth

    Wichura, 1988, `Algorithm AS 241: the percentage points of the
    Normal distribution`, Appl. Statist. (37), 477--484
    """

    fun_name = 'naginterfaces.library.stat.inv_cdf_normal_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lp_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _lp = _b_utils._EngineIntScalarType(
        _lp_py, {'fun_name': fun_name, 'entity': '_lp'},
    ).to_ctypes()
    _xmu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xmu'},
        xmu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lxmu_py = _b_utils._EngineFloat64ArrayType.get_shape(_xmu, {'fun_name': fun_name, 'entity': 'xmu'}, exp_rank=1)[0]
    _lxmu = _b_utils._EngineIntScalarType(
        _lxmu_py, {'fun_name': fun_name, 'entity': '_lxmu'},
    ).to_ctypes()
    _xstd = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xstd'},
        xstd,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lxstd_py = _b_utils._EngineFloat64ArrayType.get_shape(_xstd, {'fun_name': fun_name, 'entity': 'xstd'}, exp_rank=1)[0]
    _lxstd = _b_utils._EngineIntScalarType(
        _lxstd_py, {'fun_name': fun_name, 'entity': '_lxstd'},
    ).to_ctypes()
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lxmu_py, _lxstd_py, _lp_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lxmu_py, _lxstd_py, _lp_py),),
    )
    try:
        _b_stat.inv_cdf_normal_vector(
            _ltail,
            _tail,
            _lp,
            _p,
            _lxmu,
            _xmu,
            _lxstd,
            _xstd,
            _x,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _ivalid_py = _ivalid
    _return_data = InvCdfNormalVectorReturnData(
        _x_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

InvCdfStudentsTVectorReturnData = _collections.namedtuple(
    'InvCdfStudentsTVectorReturnData',
    ['t', 'ivalid'],
)

def inv_cdf_students_t_vector(tail, p, df):
    u"""
    Computes a vector of deviates for Student's t-distribution.

    ``inv_cdf_students_t_vector`` returns a number of deviates
    associated with given probabilities of Student's t-distribution with
    real degrees of freedom.

    For full information please refer to the NAG Library document for
    g01tb

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01tbf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates which tail the supplied probabilities represent. For j
        = (mod(i-1, ltail)) +1 -1, for i = 1,2,...,max(ltail, lp, ldf):

        `tail`[j] = 'L'

            The lower tail probability, i.e., p_i = (T_i <= t_p_i:nu_i).

        `tail`[j] = 'U'

            The upper tail probability, i.e., p_i = (T_i >= t_p_i:nu_i).

        `tail`[j] = 'C'

            The two tail (confidence interval) probability, i.e., p_i =
            (T_i <= |t_p_i|:nu_i)-(T_i <= -|t_p_i|:nu_i).

        `tail`[j] = 'S'

            The two tail (significance level) probability, i.e., p_i =
            (T_i >= |t_p_i|:nu_i)+(T_i <= -|t_p_i|:nu_i).

    p : float, array-like, shape (lp)
        p_i, the probability of the required Student's t-distribution as
        defined by `tail`.

    df : float, array-like, shape (ldf)
        nu_i, the degrees of freedom of the Student's t-distribution.

    Returns
    -------
    t : float, ndarray, shape (max(ltail, lp, ldf))
        t_p_i, the deviates for the Student's t-distribution.

    ivalid : int, ndarray, shape (max(ltail, lp, ldf))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, invalid value supplied in `tail` when calculating
            t_p_i.

        `ivalid`[i-1] = 2

            On entry, p_i <= 0.0, or, p_i >= 1.0.

        `ivalid`[i-1] = 3

            On entry, nu_i < 1.0.

        `ivalid`[i-1] = 4

            The solution has failed to converge. The result returned
            should represent an approximation to the solution.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lp > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: ldf > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `tail`, `p` or `df` was
            invalid, or the solution failed to converge.

            Check `ivalid` for more information.

    Notes
    -----
    The deviate, t_p_i associated with the lower tail probability, p_i,
    of the Student's t-distribution with nu_i degrees of freedom is
    defined as the solution to

        (T_i < t_p_i:nu_i) = p_i =
        Gamma((nu_i+1)/2)/(sqrt(nu_ipi)Gamma(nu_i/2))int_-
        infinity^t_p_i(1+(T_i^2)/(nu_i))^(-(nu_i+1)/2)dT_i, nu_i >= 1; -
        infinity < t_p_i < infinity.

    For nu_i = 1 or 2 the integral equation is easily solved for t_p_i.

    For other values of nu_i < 3 a transformation to the beta
    distribution is used and the result obtained from ``inv_cdf_beta``.

    For nu_i >= 3 an inverse asymptotic expansion of Cornish--Fisher
    type is used.
    The algorithm is described by Hill (1970).

    The input arrays to this function are designed to allow maximum
    flexibility in the supply of vector arguments by re-using elements
    of any arrays that are shorter than the total number of evaluations
    required.
    See [the G01 Introduction] for further information.

    References
    ----------
    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth

    Hill, G W, 1970, `Student's t-distribution`, Comm. ACM (13(10)),
    617--619
    """

    fun_name = 'naginterfaces.library.stat.inv_cdf_students_t_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lp_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _lp = _b_utils._EngineIntScalarType(
        _lp_py, {'fun_name': fun_name, 'entity': '_lp'},
    ).to_ctypes()
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ldf_py = _b_utils._EngineFloat64ArrayType.get_shape(_df, {'fun_name': fun_name, 'entity': 'df'}, exp_rank=1)[0]
    _ldf = _b_utils._EngineIntScalarType(
        _ldf_py, {'fun_name': fun_name, 'entity': '_ldf'},
    ).to_ctypes()
    _t = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _ldf_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _ldf_py),),
    )
    try:
        _b_stat.inv_cdf_students_t_vector(
            _ltail,
            _tail,
            _lp,
            _p,
            _ldf,
            _df,
            _t,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _t_py = _t
    _ivalid_py = _ivalid
    _return_data = InvCdfStudentsTVectorReturnData(
        _t_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

InvCdfChisqVectorReturnData = _collections.namedtuple(
    'InvCdfChisqVectorReturnData',
    ['x', 'ivalid'],
)

def inv_cdf_chisq_vector(tail, p, df):
    u"""
    Computes a vector of deviates for chi^2 distribution.

    ``inv_cdf_chisq_vector`` returns a number of deviates associated
    with the given probabilities of the chi^2-distribution with real
    degrees of freedom.

    For full information please refer to the NAG Library document for
    g01tc

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01tcf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates which tail the supplied probabilities represent. For j
        = (mod(i-1, ltail)) +1 -1, for i = 1,2,...,max(ltail, lp, ldf):

        `tail`[j] = 'L'

            The lower tail probability, i.e., p_i = (X_i <= x_p_i:nu_i).

        `tail`[j] = 'U'

            The upper tail probability, i.e., p_i = (X_i >= x_p_i:nu_i).

    p : float, array-like, shape (lp)
        p_i, the probability of the required chi^2-distribution as
        defined by `tail`.

    df : float, array-like, shape (ldf)
        nu_i, the degrees of freedom of the chi^2-distribution.

    Returns
    -------
    x : float, ndarray, shape (max(ltail, lp, ldf))
        x_p_i, the deviates for the chi^2-distribution.

    ivalid : int, ndarray, shape (max(ltail, lp, ldf))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, invalid value supplied in `tail` when calculating
            x_p_i.

        `ivalid`[i-1] = 2

            On entry, invalid value for p_i.

        `ivalid`[i-1] = 3

            On entry, nu_i <= 0.0.

        `ivalid`[i-1] = 4

            p_i is too close to 0.0 or 1.0 for the result to be
            calculated.

        `ivalid`[i-1] = 5

            The solution has failed to converge. The result should be a
            reasonable approximation.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lp > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: ldf > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `tail`, `p` or `df` was
            invalid, or the solution failed to converge.

            Check `ivalid` for more information.

    Notes
    -----
    The deviate, x_p_i, associated with the lower tail probability p_i
    of the chi^2-distribution with nu_i degrees of freedom is defined as
    the solution to

        (X_i <= x_p_i:nu_i) = p_i =
        1/(2^(nu_i/2)Gamma(nu_i/2))int_0^x_p_ie^(-X_i/2)X_i^(v_i/2-1)dX_i,
        0 <= x_p_i < infinity ; nu_i > 0.

    The required x_p_i is found by using the relationship between a
    chi^2-distribution and a gamma distribution, i.e., a
    chi^2-distribution with nu_i degrees of freedom is equal to a gamma
    distribution with scale parameter 2 and shape parameter nu_i/2.

    For very large values of nu_i, greater than 10^5, Wilson and
    Hilferty's Normal approximation to the chi^2 is used; see Kendall
    and Stuart (1969).

    The input arrays to this function are designed to allow maximum
    flexibility in the supply of vector arguments by re-using elements
    of any arrays that are shorter than the total number of evaluations
    required.
    See [the G01 Introduction] for further information.

    References
    ----------
    Best, D J and Roberts, D E, 1975, `Algorithm AS 91. The percentage
    points of the chi^2 distribution`, Appl. Statist. (24), 385--388

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth

    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin
    """

    fun_name = 'naginterfaces.library.stat.inv_cdf_chisq_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lp_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _lp = _b_utils._EngineIntScalarType(
        _lp_py, {'fun_name': fun_name, 'entity': '_lp'},
    ).to_ctypes()
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ldf_py = _b_utils._EngineFloat64ArrayType.get_shape(_df, {'fun_name': fun_name, 'entity': 'df'}, exp_rank=1)[0]
    _ldf = _b_utils._EngineIntScalarType(
        _ldf_py, {'fun_name': fun_name, 'entity': '_ldf'},
    ).to_ctypes()
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _ldf_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _ldf_py),),
    )
    try:
        _b_stat.inv_cdf_chisq_vector(
            _ltail,
            _tail,
            _lp,
            _p,
            _ldf,
            _df,
            _x,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _ivalid_py = _ivalid
    _return_data = InvCdfChisqVectorReturnData(
        _x_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

InvCdfFVectorReturnData = _collections.namedtuple(
    'InvCdfFVectorReturnData',
    ['f', 'ivalid'],
)

def inv_cdf_f_vector(tail, p, df1, df2):
    u"""
    Computes a vector of deviates for F-distribution.

    ``inv_cdf_f_vector`` returns a number of deviates associated with
    given probabilities of the F or variance-ratio distribution with
    real degrees of freedom.

    For full information please refer to the NAG Library document for
    g01td

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01tdf.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates which tail the supplied probabilities represent. For j
        = (mod(i-1, ltail)) +1 -1, for i = 1,2,...,max(ltail, lp, ldf1,
        ldf2):

        `tail`[j] = 'L'

            The lower tail probability, i.e., p_i = (F_i <=
            f_p_i:u_i,v_i).

        `tail`[j] = 'U'

            The upper tail probability, i.e., p_i = (F_i >=
            f_p_i:u_i,v_i).

    p : float, array-like, shape (lp)
        p_i, the probability of the required F-distribution as defined
        by `tail`.

    df1 : float, array-like, shape (ldf1)
        u_i, the degrees of freedom of the numerator variance.

    df2 : float, array-like, shape (ldf2)
        v_i, the degrees of freedom of the denominator variance.

    Returns
    -------
    f : float, ndarray, shape (max(ltail, lp, ldf1, ldf2))
        f_p_i, the deviates for the F-distribution.

    ivalid : int, ndarray, shape (max(ltail, lp, ldf1, ldf2))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, invalid value supplied in `tail` when calculating
            f_p_i.

        `ivalid`[i-1] = 2

            On entry, invalid value for p_i.

        `ivalid`[i-1] = 3

            On entry, u_i <= 0.0, or, v_i <= 0.0.

        `ivalid`[i-1] = 4

            The solution has not converged. The result should still be a
            reasonable approximation to the solution.

        `ivalid`[i-1] = 5

            The value of p_i is too close to 0.0 or 1.0 for the result
            to be computed. This will only occur when the large sample
            approximations are used.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lp > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: ldf1 > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: ldf2 > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `tail`, `p`, `df1`, `df2`
            was invalid, or the solution failed to converge.

            Check `ivalid` for more information.

    Notes
    -----
    The deviate, f_p_i, associated with the lower tail probability, p_i,
    of the F-distribution with degrees of freedom u_i and v_i is defined
    as the solution to

        (F_i <= f_p_i:u_i,v_i) = p_i = u_i^1/2u_iv_i^1/2v_i
        Gamma((u_i+v_i)/2)/(Gamma((u_i)/2)Gamma((v_i)/2))int_0^f_p_iF_i^(1/2(u_i-2))(v_i+u_i
        F_i)^(-1/2(u_i+v_i))dF_i,

    where u_i,v_i > 0; 0 <= f_p_i < infinity.

    The value of f_p_i is computed by means of a transformation to a
    beta distribution, Pi beta_i(B_i <= beta_i:a_i,b_i):

        (F_i <= f_p_i:u_i,v_i) = Pi beta_i(B_i <= u_i f_p_i/(u_i
        f_p_i+v_i):u_i/2,v_i/2)

    and using a call to ``inv_cdf_beta_vector``.

    For very large values of both u_i and v_i, greater than 10^5, a
    Normal approximation is used.
    If only one of u_i or v_i is greater than 10^5 then a chi^2
    approximation is used; see Abramowitz and Stegun (1972).

    The input arrays to this function are designed to allow maximum
    flexibility in the supply of vector arguments by re-using elements
    of any arrays that are shorter than the total number of evaluations
    required.
    See [the G01 Introduction] for further information.

    References
    ----------
    Abramowitz, M and Stegun, I A, 1972, `Handbook of Mathematical
    Functions`, (3rd Edition), Dover Publications

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth
    """

    fun_name = 'naginterfaces.library.stat.inv_cdf_f_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lp_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _lp = _b_utils._EngineIntScalarType(
        _lp_py, {'fun_name': fun_name, 'entity': '_lp'},
    ).to_ctypes()
    _df1 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df1'},
        df1,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ldf1_py = _b_utils._EngineFloat64ArrayType.get_shape(_df1, {'fun_name': fun_name, 'entity': 'df1'}, exp_rank=1)[0]
    _ldf1 = _b_utils._EngineIntScalarType(
        _ldf1_py, {'fun_name': fun_name, 'entity': '_ldf1'},
    ).to_ctypes()
    _df2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df2'},
        df2,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ldf2_py = _b_utils._EngineFloat64ArrayType.get_shape(_df2, {'fun_name': fun_name, 'entity': 'df2'}, exp_rank=1)[0]
    _ldf2 = _b_utils._EngineIntScalarType(
        _ldf2_py, {'fun_name': fun_name, 'entity': '_ldf2'},
    ).to_ctypes()
    _f = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _ldf1_py, _ldf2_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _ldf1_py, _ldf2_py),),
    )
    try:
        _b_stat.inv_cdf_f_vector(
            _ltail,
            _tail,
            _lp,
            _p,
            _ldf1,
            _df1,
            _ldf2,
            _df2,
            _f,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _f_py = _f
    _ivalid_py = _ivalid
    _return_data = InvCdfFVectorReturnData(
        _f_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

InvCdfBetaVectorReturnData = _collections.namedtuple(
    'InvCdfBetaVectorReturnData',
    ['beta', 'ivalid'],
)

def inv_cdf_beta_vector(tail, p, a, b, tol=0.0):
    u"""
    Computes a vector of deviates for the beta distribution.

    ``inv_cdf_beta_vector`` returns a number of deviates associated with
    given probabilities of the beta distribution.

    For full information please refer to the NAG Library document for
    g01te

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01tef.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates which tail the supplied probabilities represent. For j
        = (mod(i-1, ltail)) +1 -1, for i = 1,2,...,max(ltail, lp, la,
        lb):

        `tail`[j] = 'L'

            The lower tail probability, i.e., p_i = (B_i <=
            beta_p_i:a_i,b_i).

        `tail`[j] = 'U'

            The upper tail probability, i.e., p_i = (B_i >=
            beta_p_i:a_i,b_i).

    p : float, array-like, shape (lp)
        p_i, the probability of the required beta distribution as
        defined by `tail`.

    a : float, array-like, shape (la)
        a_i, the first parameter of the required beta distribution.

    b : float, array-like, shape (lb)
        b_i, the second parameter of the required beta distribution.

    tol : float, optional
        The relative accuracy required by you in the results. If
        ``inv_cdf_beta_vector`` is entered with `tol` greater than or
        equal to 1.0 or less than 10*machine precision (see
        ``machine.precision``), the value of 10*machine precision is
        used instead.

    Returns
    -------
    beta : float, ndarray, shape (max(ltail, lp, la, lb))
        beta_p_i, the deviates for the beta distribution.

    ivalid : int, ndarray, shape (max(ltail, lp, la, lb))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, invalid value supplied in `tail` when calculating
            beta_p_i.

        `ivalid`[i-1] = 2

            On entry, p_i < 0.0, or, p_i > 1.0.

        `ivalid`[i-1] = 3

            On entry, a_i <= 0.0, or, a_i > 10^6, or, b_i <= 0.0, or,
            b_i > 10^6.

        `ivalid`[i-1] = 4

            The solution has not converged but the result should be a
            reasonable approximation to the solution.

        `ivalid`[i-1] = 5

            Requested accuracy not achieved when calculating the beta
            probability. The result should be a reasonable approximation
            to the correct solution.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lp > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: la > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: lb > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `tail`, `p`, `a`, or `b` was
            invalid, or the solution failed to converge.

            Check `ivalid` for more information.

    Notes
    -----
    The deviate, beta_p_i, associated with the lower tail probability,
    p_i, of the beta distribution with parameters a_i and b_i is defined
    as the solution to

        (B_i <= beta_p_i:a_i,b_i) = p_i =
        Gamma(a_i+b_i)/(Gamma(a_i)Gamma(b_i))int_0^beta_p_iB_i^(a_i-1)(1-B_i)^(b_i-1)dB_i,
        0 <= beta_p_i <= 1; a_i,b_i > 0.

    The algorithm is a modified version of the Newton--Raphson method,
    following closely that of Cran `et al.` (1977).

    An initial approximation, beta_{i0}, to beta_p_i is found (see Cran
    `et al.` (1977)), and the Newton--Raphson iteration

        beta_k = beta_{k-1}-f_i(beta_{k-1})/f_i'(beta_{k-1}),

    where f_i(beta_k) = (B_i <= beta_k:a_i,b_i)-p_i is used, with
    modifications to ensure that beta_k remains in the range (0,1).

    The input arrays to this function are designed to allow maximum
    flexibility in the supply of vector arguments by re-using elements
    of any arrays that are shorter than the total number of evaluations
    required.
    See [the G01 Introduction] for further information.

    References
    ----------
    Cran, G W, Martin, K J and Thomas, G E, 1977, `Algorithm AS 109.
    Inverse of the incomplete beta function ratio`, Appl. Statist. (26),
    111--114

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth
    """

    fun_name = 'naginterfaces.library.stat.inv_cdf_beta_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lp_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _lp = _b_utils._EngineIntScalarType(
        _lp_py, {'fun_name': fun_name, 'entity': '_lp'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _la_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _la = _b_utils._EngineIntScalarType(
        _la_py, {'fun_name': fun_name, 'entity': '_la'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lb_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=1)[0]
    _lb = _b_utils._EngineIntScalarType(
        _lb_py, {'fun_name': fun_name, 'entity': '_lb'},
    ).to_ctypes()
    _beta = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _la_py, _lb_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _la_py, _lb_py),),
    )
    try:
        _b_stat.inv_cdf_beta_vector(
            _ltail,
            _tail,
            _lp,
            _p,
            _la,
            _a,
            _lb,
            _b,
            tol,
            _beta,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _beta_py = _beta
    _ivalid_py = _ivalid
    _return_data = InvCdfBetaVectorReturnData(
        _beta_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

InvCdfGammaVectorReturnData = _collections.namedtuple(
    'InvCdfGammaVectorReturnData',
    ['g', 'ivalid'],
)

def inv_cdf_gamma_vector(tail, p, a, b, tol=0.0):
    u"""
    Computes a vector of deviates for the gamma distribution.

    ``inv_cdf_gamma_vector`` returns a number of deviates associated
    with given probabilities of the gamma distribution.

    For full information please refer to the NAG Library document for
    g01tf

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01tff.html

    Parameters
    ----------
    tail : str, length 1, array-like, shape (ltail)
        Indicates which tail the supplied probabilities represent. For j
        = (mod(i-1, ltail)) +1 -1, for i = 1,2,...,max(ltail, lp, la,
        lb):

        `tail`[j] = 'L'

            The lower tail probability, i.e., p_i = (G_i <=
            g_p_i:alpha_i,beta_i).

        `tail`[j] = 'U'

            The upper tail probability, i.e., p_i = (G_i >=
            g_p_i:alpha_i,beta_i).

    p : float, array-like, shape (lp)
        p_i, the probability of the required gamma distribution as
        defined by `tail`.

    a : float, array-like, shape (la)
        alpha_i, the first parameter of the required gamma distribution.

    b : float, array-like, shape (lb)
        beta_i, the second parameter of the required gamma distribution.

    tol : float, optional
        The relative accuracy required by you in the results. If
        ``inv_cdf_gamma_vector`` is entered with `tol` greater than or
        equal to 1.0 or less than 10*machine precision (see
        ``machine.precision``), the value of 10*machine precision is
        used instead.

    Returns
    -------
    g : float, ndarray, shape (max(ltail, lp, la, lb))
        g_p_i, the deviates for the gamma distribution.

    ivalid : int, ndarray, shape (max(ltail, lp, la, lb))
        `ivalid`[i-1] indicates any errors with the input arguments,
        with

        `ivalid`[i-1] = 0

            No error.

        `ivalid`[i-1] = 1

            On entry, invalid value supplied in `tail` when calculating
            g_p_i.

        `ivalid`[i-1] = 2

            On entry, invalid value for p_i.

        `ivalid`[i-1] = 3

            On entry, alpha_i <= 0.0, or, alpha_i > 10^6, or, beta_i <=
            0.0.

        `ivalid`[i-1] = 4

            p_i is too close to 0.0 or 1.0 to enable the result to be
            calculated.

        `ivalid`[i-1] = 5

            The solution has failed to converge. The result may be a
            reasonable approximation.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, array size = *<value>*.

            Constraint: ltail > 0.

        (`errno` 3)
            On entry, array size = *<value>*.

            Constraint: lp > 0.

        (`errno` 4)
            On entry, array size = *<value>*.

            Constraint: la > 0.

        (`errno` 5)
            On entry, array size = *<value>*.

            Constraint: lb > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one value of `tail`, `p`, `a`, or `b` was
            invalid.

            Check `ivalid` for more information.

    Notes
    -----
    The deviate, g_p_i, associated with the lower tail probability, p_i,
    of the gamma distribution with shape parameter alpha_i and scale
    parameter beta_i, is defined as the solution to

        (G_i <= g_p_i:alpha_i,beta_i) = p_i =
        1/(beta_i^(alpha_i)Gamma(alpha_i))int_0^g_p_ie_i^(-G_i/beta_i)G_i^(alpha_i-1)dG_i,
        0 <= g_p_i < infinity ; alpha_i,beta_i > 0.

    The method used is described by Best and Roberts (1975) making use
    of the relationship between the gamma distribution and the
    chi^2-distribution.

    Let y_i = 2(g_p_i)/(beta_i).
    The required y_i is found from the Taylor series expansion

        y_i = y_0+sum_r C_r(y_0)/r!((E_i)/phi(y_0))^r,

    where y_0 is a starting approximation

        C_1(u_i) = 1,

        C_{r+1}(u_i) = (r Psi+d/du_i)C_r(u_i),

        Psi_i = 1/2-(alpha_i-1)/(u_i),

        E_i = p_i-int_0^y_0 phi_i(u_i)du_i,

        phi_i(u_i) =
        1/(2^(alpha_i)Gamma(alpha_i))e_i^(-u_i/2)u_i^(alpha_i-1).

    For most values of p_i and alpha_i the starting value

        y_01 = 2 alpha_i(z_i*sqrt(1/9 alpha_i)+1-1/9 alpha_i)^3

    is used, where z_i is the deviate associated with a lower tail
    probability of p_i for the standard Normal distribution.

    For p_i close to zero,

        y_02 = (p_ialpha_i 2^(alpha_i)Gamma(alpha_i))^(1/alpha_i)

    is used.

    For large p_i values, when y_01 > 4.4 alpha_i+6.0,

        y_03 = -2*[ln(1-p_i)-(alpha_i-1)ln(1/2y_01)+ln(Gamma(alpha_i))]

    is found to be a better starting value than y_01.

    For small alpha_i (alpha_i <= 0.16), p_i is expressed in terms of an
    approximation to the exponential integral and y_04 is found by
    Newton--Raphson iterations.

    Seven terms of the Taylor series are used to refine the starting
    approximation, repeating the process if necessary until the required
    accuracy is obtained.

    The input arrays to this function are designed to allow maximum
    flexibility in the supply of vector arguments by re-using elements
    of any arrays that are shorter than the total number of evaluations
    required.
    See [the G01 Introduction] for further information.

    References
    ----------
    Best, D J and Roberts, D E, 1975, `Algorithm AS 91. The percentage
    points of the chi^2 distribution`, Appl. Statist. (24), 385--388
    """

    fun_name = 'naginterfaces.library.stat.inv_cdf_gamma_vector'

    _tail = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'tail'},
        tail,
        copy=False,
    )
    _ltail_py = _b_utils._EngineCharArrayType.get_shape(_tail, {'fun_name': fun_name, 'entity': 'tail'}, exp_rank=1)[0]
    _ltail = _b_utils._EngineIntScalarType(
        _ltail_py, {'fun_name': fun_name, 'entity': '_ltail'},
    ).to_ctypes()
    _p = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lp_py = _b_utils._EngineFloat64ArrayType.get_shape(_p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=1)[0]
    _lp = _b_utils._EngineIntScalarType(
        _lp_py, {'fun_name': fun_name, 'entity': '_lp'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _la_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _la = _b_utils._EngineIntScalarType(
        _la_py, {'fun_name': fun_name, 'entity': '_la'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lb_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=1)[0]
    _lb = _b_utils._EngineIntScalarType(
        _lb_py, {'fun_name': fun_name, 'entity': '_lb'},
    ).to_ctypes()
    _g = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _la_py, _lb_py),),
    )
    _ivalid = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(_ltail_py, _lp_py, _la_py, _lb_py),),
    )
    try:
        _b_stat.inv_cdf_gamma_vector(
            _ltail,
            _tail,
            _lp,
            _p,
            _la,
            _a,
            _lb,
            _b,
            tol,
            _g,
            _ivalid,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _g_py = _g
    _ivalid_py = _ivalid
    _return_data = InvCdfGammaVectorReturnData(
        _g_py,
        _ivalid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

MovingAverageReturnData = _collections.namedtuple(
    'MovingAverageReturnData',
    ['pn', 'rmean', 'rsd'],
)

def moving_average(m, x, iwt=0, wt=None, pn=0, wantsd=False, comm=None):
    u"""
    Computes the mean and standard deviation using a rolling window.

    ``moving_average`` calculates the mean and, optionally, the standard
    deviation using a rolling window for an arbitrary sized data stream.

    For full information please refer to the NAG Library document for
    g01wa

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01waf.html

    Parameters
    ----------
    m : int
        m, the length of the rolling window.

        If `pn` != 0, `m` must be unchanged since the last call to
        ``moving_average``.

    x : float, array-like, shape (nb)
        The current block of observations, corresponding to x_i, for i =
        k+1,...,k+b, where k is the number of observations processed so
        far and b is the size of the current block of data.

    iwt : int, optional
        The type of weighting to use.

        `iwt` = 0

            No weights are used.

        `iwt` = 1

            Each observation has its own weight.

        `iwt` = 2

            Each position in the window has its own weight.

        `iwt` = 3

            Each position in the window has a weight equal to its
            position number.

        If `pn` != 0, `iwt` must be unchanged since the last call to
        ``moving_average``.

    wt : None or float, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `iwt`=1: nb; if `iwt`=2: `m`; otherwise: 0.

        The user-supplied weights.

        If `iwt` = 1, `wt`[i-1] = nu_{i+k}, for i = 1,2,...,b.

        If `iwt` = 2, `wt`[j-1] = w_j, for j = 1,2,...,m.

    pn : int, optional
        k, the number of observations processed so far. On the first
        call to ``moving_average``, or when starting to summarise a new
        dataset, `pn` must be set to 0.

        If `pn` != 0, it must be the same value as returned by the last
        call to ``moving_average``.

    wantsd : bool, optional
        If the standard deviations are required then `wantsd` should be
        set to True.

    comm : None or dict, communication object, optional, modified in
    place
        Communication structure.

        If None all the data must be supplied in one go, otherwise need
        not be set.

    Returns
    -------
    pn : int
        k+b, the updated number of observations processed so far.

    rmean : float, ndarray, shape (max(0, nb+min(0, `pn`-`m`+1)))
        mu_l, the (weighted) moving averages, for l = 1,2,...,b+min(0,
        k-m+1). Therefore, mu_l is the mean of the data in the window
        that ends on `x`[l+m-min(k, m-1)-2].

        If, on entry, `pn` >= `m`-1, i.e., at least one windows worth of
        data has been previously processed, then `rmean`[l-1] is the
        summary corresponding to the window that ends on `x`[l-1].

        On the other hand, if, on entry, `pn` = 0, i.e., no data has
        been previously processed, then `rmean`[l-1] is the summary
        corresponding to the window that ends on `x`[`m`+l-2] (or,
        equivalently, starts on `x`[l-1]).

    rsd : None or float, ndarray, shape (:)
        If `wantsd` = True then sigma_l, the (weighted) standard
        deviation. The ordering of `rsd` is the same as the ordering of
        `rmean`.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 12)
            On entry, `m` = *<value>*.

            On entry at previous call, `m` = *<value>*.

            Constraint: if `pn` > 0, `m` must be unchanged since
            previous call.

        (`errno` 21)
            On entry, nb = *<value>*.

            Constraint: nb >= 0.

        (`errno` 22)
            On entry, nb = *<value>*, `m` = *<value>*.

            Constraint: if `comm` is None, nb >= `m`.

        (`errno` 41)
            On entry, `iwt` = *<value>*.

            Constraint: `iwt` = 0, 1, 2 or 3.

        (`errno` 42)
            On entry, `iwt` = *<value>*.

            On entry at previous call, `iwt` = *<value>*.

            Constraint: if `pn` > 0, `iwt` must be unchanged since
            previous call.

        (`errno` 51)
            On entry, `wt`[*<value>*] = *<value>*.

            Constraint: `wt`[i-1] >= 0.

        (`errno` 52)
            On entry, `wt`[0] = *<value>*.

            Constraint: if `iwt` = 2, `wt`[0] > 0.

        (`errno` 55)
            On entry, sum of weights supplied in `wt` is *<value>*.

            Constraint: if `iwt` = 2, the sum of the weights > 0.

        (`errno` 61)
            On entry, `pn` = *<value>*.

            Constraint: `pn` >= 0.

        (`errno` 62)
            On entry, `pn` = *<value>*.

            On exit from previous call, `pn` = *<value>*.

            Constraint: if `pn` > 0, `pn` must be unchanged since
            previous call.

        (`errno` 101)
            `comm`['rcomm'] has been corrupted between calls.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 53)
            On entry, at least one window had all zero weights.

        (`errno` 54)
            On entry, unable to calculate at least one standard
            deviation due to the weights supplied.

    Notes
    -----
    Given a sample of n observations, denoted by x = {x_i:i = 1,2,...,n}
    and a set of weights, w = {w_j:j = 1,2,...,m}, ``moving_average``
    calculates the mean and, optionally, the standard deviation, in a
    rolling window of length m.

    For the ith window the mean is defined as

        mu_i = sum_j = 1^mw_j*x_{i+j-1}/W

    and the standard deviation as

        sigma_i = sqrt(sum_j = 1^mw_j*(x_{i+j-1}-mu_i)^2/(W-sum_j =
        1^mw_j^2/W))

    with W = sum_j = 1^mw_j.

    Four different types of weighting are possible:

    (i) **No weights (w_j = 1)**

        When no weights are required both the mean and standard
        deviations can be calculated in an iterative manner, with

            [table omitted]

        where the initial values mu_1 and sigma_1 are obtained using the
        one pass algorithm of West (1979).

    (#) **Each observation has its own weight**

        In this case, rather than supplying a vector of m weights a
        vector of n weights is supplied instead, v = {v_j:j = 1,2,...,n}
        and w_j = v_{i+j-1} in [equation] and [equation].

        If the standard deviations are not required then the mean is
        calculated using the iterative formula:

            [table omitted]

        where W_1 = sum_i = 1^mv_i and mu_1 = W_1^-1*sum_i = 1^mv_i*x_i.

        If both the mean and standard deviation are required then the
        one pass algorithm of West (1979) is used in each window.

    (#) **Each position in the window has its own weight**

        This is the case as described in [equation] and [equation],
        where the weight given to each observation differs depending on
        which summary is being produced.
        When these types of weights are specified both the mean and
        standard deviation are calculated by applying the one pass
        algorithm of West (1979) multiple times.

    (#) **Each position in the window has a weight equal to its position
        number (w_j = j)**

        This is a special case of (iii).

        If the standard deviations are not required then the mean is
        calculated using the iterative formula:

            [table omitted]

        where S_1 = sum_i = 1^mx_i and mu_1 = 2*(m^2+m)^-1S_1.

        If both the mean and standard deviation are required then the
        one pass algorithm of West is applied multiple times.

    For large datasets, or where all the data is not available at the
    same time, x (and if each observation has its own weight, v) can be
    split into arbitrary sized blocks and ``moving_average`` called
    multiple times.

    References
    ----------
    Chan, T F, Golub, G H and Leveque, R J, 1982, `Updating Formulae and
    a Pairwise Algorithm for Computing Sample Variances`, Compstat,
    Physica-Verlag

    West, D H D, 1979, `Updating mean and variance estimates: An
    improved method`, Comm. ACM (22), 532--555

    See Also
    --------
    :meth:`naginterfaces.library.examples.stat.moving_average_ex.main`
    """

    fun_name = 'naginterfaces.library.stat.moving_average'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nb_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nb = _b_utils._EngineIntScalarType(
        _nb_py, {'fun_name': fun_name, 'entity': '_nb'},
    ).to_ctypes()
    _wt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'wt'},
        wt,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _pn = _b_utils._EngineIntScalarType(
        pn, {'fun_name': fun_name, 'entity': 'pn'},
    ).to_ctypes()
    _pn_py = _b_utils._EngineIntScalarType.to_py(_pn)
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _rmean = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(0, _nb_py+min(0, _pn_py-_m_py+1)),),
    )
    _wantsd = _b_utils._EngineLogicalScalarType(
        wantsd, {'fun_name': fun_name, 'entity': 'wantsd'},
    ).to_ctypes()
    _wantsd_py = _b_utils._EngineLogicalScalarType.to_py(_wantsd)
    if _wantsd_py:
        if _wantsd_py:
            _rsd_dim1 = max(0, _nb_py+min(0, _pn_py-_m_py+1))
        else:
            _rsd_dim1 = 0
        exp_shape_rsd = (_rsd_dim1,)
        _rsd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_rsd,
        )
    else:
        _rsd = None
    if _pn_py != 0:
        if comm is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when pn != 0',
            )
    try:
        _return_scalar = _b_stat.moving_average(
            _m,
            _nb,
            _x,
            iwt,
            _wt,
            _pn,
            _rmean,
            _rsd,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _pn_py = _return_scalar
    _rmean_py = _rmean[:max(0, _nb_py+min(0, _pn_py-_m_py+1))]
    if _wantsd_py:
        _rsd_dim1 = max(0, _nb_py+min(0, _pn_py-_m_py+1))
    else:
        _rsd_dim1 = 0
    if _wantsd_py:
        _rsd_py = _rsd[:_rsd_dim1]
    else:
        _rsd_py = None
    _return_data = MovingAverageReturnData(
        _pn_py,
        _rmean_py,
        _rsd_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

InitVavilovReturnData = _collections.namedtuple(
    'InitVavilovReturnData',
    ['xl', 'xu', 'comm'],
)

def init_vavilov(rkappa, beta2, mode):
    u"""
    Initialization function for ``pdf_vavilov`` and ``prob_vavilov``.

    ``init_vavilov`` is used to initialize functions ``pdf_vavilov`` and
    ``prob_vavilov``.

    It is intended to be used before a call to ``pdf_vavilov`` or
    ``prob_vavilov``.

    For full information please refer to the NAG Library document for
    g01zu

    https://www.nag.com/numeric/nl/nagdoc_30/flhtml/g01/g01zuf.html

    Parameters
    ----------
    rkappa : float
        The argument kappa of the function.

    beta2 : float
        The argument beta^2 of the function.

    mode : int
        If `mode` = 0, ``pdf_vavilov`` is to be called after the call to
        ``init_vavilov``. Otherwise, ``prob_vavilov`` is to be called.

    Returns
    -------
    xl : float
        x_l, a threshold value below which phi_V(lambda;kappa,beta^2)
        will be set to zero by ``pdf_vavilov`` and
        Phi_V(lambda;kappa,beta^2) will be set to zero by
        ``prob_vavilov`` if lambda < x_l.

    xu : float
        x_u, a threshold value above which phi_V(lambda;kappa,beta^2)
        will be set to zero by ``pdf_vavilov`` and
        Phi_V(lambda;kappa,beta^2) will be set to unity by
        ``prob_vavilov`` if lambda > x_u.

    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `beta2` = *<value>*.

            Constraint: `beta2` <= 1.0.

        (`errno` 1)
            On entry, `beta2` = *<value>*.

            Constraint: `beta2` >= 0.0.

        (`errno` 1)
            On entry, `rkappa` = *<value>*.

            Constraint: `rkappa` <= 10.0.

        (`errno` 1)
            On entry, `rkappa` = *<value>*.

            Constraint: `rkappa` >= 0.01.

        (`errno` 2)
            The initialization has been abandoned due to an internal
            error. This error exit is unlikely to occur, but if it does
            change the values of `rkappa` and/or `beta2` and rerun
            ``init_vavilov``.

    Notes
    -----
    ``init_vavilov`` initializes the array `comm`['rcomm'] for use by
    ``pdf_vavilov`` or ``prob_vavilov`` in the evaluation of the Vavilov
    functions phi_V(lambda;kappa,beta^2) and Phi_V(lambda;kappa,beta^2)
    respectively.

    Multiple calls to ``prob_vavilov`` or ``pdf_vavilov`` can be made
    following a single call to ``init_vavilov``, provided that `rkappa`
    or `beta2` do not change, and that either all calls are to
    ``prob_vavilov`` or all calls are to ``pdf_vavilov``.
    If you wish to call both ``prob_vavilov`` and ``pdf_vavilov``, then
    you will need to initialize both separately.

    References
    ----------
    Schorr, B, 1974, `Programs for the Landau and the Vavilov
    distributions and the corresponding random numbers`, Comp. Phys.
    Comm. (7), 215--224
    """

    _return_scalars = _b_stat.init_vavilov(
        rkappa,
        beta2,
        mode,
    )
    _xl_py = _return_scalars.xl
    _xu_py = _return_scalars.xu
    _comm_py = _return_scalars.comm
    _return_data = InitVavilovReturnData(
        _xl_py,
        _xu_py,
        _comm_py,
    )
    return _return_data
